Package: ViralEntropR
Title: A Computational Pipeline for Entropy-Informed Detection of
        Emerging Viral Variants
Version: 0.6.2
Authors@R: c(
              person("Vadim", "Tyuryaev", ,"vadim.tyuryaev@gmail.com", 
                      role = c("aut", "cre"), 
                      comment = c(ORCID = "0009-0008-1361-6265")),
              person("Jane", "Heffernan", , "jmheffer@yorku.ca", 
                      role = c("aut")),
              person("Hanna", "Jankowski", , "hkj@yorku.ca", 
                      role = c("aut"))
             )
Description: Implements an entropy-informed pipeline for detecting
    emerging variants in viral amino acid sequence data, extending
    prior clustering-based approaches including hemagglutinin
    clustering methods (Li et al., 2015)
    <doi:10.1142/9789814667944_0018>. Provides a fully vectorized
    FASTA preprocessing toolkit covering header parsing, two-pass
    date and country extraction, ambiguous-residue filtering, and
    integer encoding under a 25-symbol amino acid alphabet. Computes
    per-site Shannon entropy across user-defined cumulative,
    sliding, or disjoint temporal partitions and clusters per-site
    entropy values using Gaussian mixture models via 'mclust'
    (Scrucca et al., 2016) <doi:10.32614/RJ-2016-021>. Quantifies
    temporal distributional shifts between partitions using the
    Hellinger distance (van der Vaart, 1998)
    <doi:10.1017/CBO9780511802256>, and detects temporal change
    points non-parametrically using energy statistics (Matteson and
    James, 2014) <doi:10.1080/01621459.2013.849605> via 'ecp' or
    wild binary segmentation (Fryzlewicz, 2014)
    <doi:10.1214/14-AOS1245> via 'HDcpDetect'. Per-site amino-acid 
    frequency tables and entropy trajectory plots characterize sequence 
    composition and evolutionary dynamics across time. A configurable 
    multi-variant simulation engine generates synthetic sequence 
    time series with known ground truth for benchmarking detection pipelines. 
    A curated dataset of SARS-CoV-2 Variants of Concern and Variants of
    Interest with associated lineage and surveillance metadata is
    included, along with a bundled National Center for Biotechnology
    Information (NCBI) Spike protein sample and vignettes
    demonstrating the full workflow.
License: MIT + file LICENSE
Language: en-GB
Date: 2026-05-07
URL: https://github.com/vadimtyuryaev/ViralEntropR,
        https://doi.org/10.5281/zenodo.19040165,
        https://vadimtyuryaev.github.io/ViralEntropR/
BugReports: https://github.com/vadimtyuryaev/ViralEntropR/issues
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
Imports: ggplot2 (>= 3.4.0), grDevices, HDcpDetect, ecp, kableExtra,
        lubridate, magrittr, mclust, rlang, stats, stringr, utils, zoo
Suggests: Biostrings, DT, dplyr, here, knitr, readxl, rmarkdown, R.rsp,
        testthat (>= 3.0.0)
Config/testthat/edition: 3
VignetteBuilder: knitr, R.rsp
NeedsCompilation: no
Packaged: 2026-05-27 18:20:05 UTC; vadim
Author: Vadim Tyuryaev [aut, cre] (ORCID:
    <https://orcid.org/0009-0008-1361-6265>),
  Jane Heffernan [aut],
  Hanna Jankowski [aut]
Maintainer: Vadim Tyuryaev <vadim.tyuryaev@gmail.com>
Depends: R (>= 3.5.0)
Repository: CRAN
Date/Publication: 2026-05-30 13:40:21 UTC
Built: R 4.6.0; ; 2026-05-30 17:09:44 UTC; unix

# Generated by roxygen2: do not edit by hand

export(calculate_entropy)
export(calculate_hellinger_matrix)
export(cluster_sites_by_entropy)
export(decode_aa_sequence)
export(detect_changepoints_ecp)
export(detect_changepoints_hdcp)
export(encode_aa_sequence)
export(extract_fasta_countries)
export(extract_fasta_dates)
export(fasta_to_char_matrix)
export(filter_ambiguous_sequences)
export(partition_time_windows)
export(plot_entropy_trajectories)
export(plot_site_class_trajectory)
export(relabel_entropy_classes)
export(simulate_variant_evolution)
export(tabulate_site_evolution)
importFrom(HDcpDetect,binary.segmentation)
importFrom(HDcpDetect,wild.binary.segmentation)
importFrom(ecp,ks.cp3o)
importFrom(ggplot2,aes)
importFrom(ggplot2,element_text)
importFrom(ggplot2,expansion)
importFrom(ggplot2,geom_label)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggsave)
importFrom(ggplot2,guide_legend)
importFrom(ggplot2,guides)
importFrom(ggplot2,labs)
importFrom(ggplot2,scale_color_grey)
importFrom(ggplot2,scale_color_manual)
importFrom(ggplot2,scale_linetype_manual)
importFrom(ggplot2,scale_linewidth_manual)
importFrom(ggplot2,scale_x_continuous)
importFrom(ggplot2,scale_y_continuous)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_bw)
importFrom(ggplot2,unit)
importFrom(grDevices,hcl.colors)
importFrom(kableExtra,column_spec)
importFrom(kableExtra,kable)
importFrom(kableExtra,kable_styling)
importFrom(kableExtra,save_kable)
importFrom(lubridate,"%m+%")
importFrom(lubridate,"%m-%")
importFrom(lubridate,interval)
importFrom(lubridate,period)
importFrom(magrittr,"%>%")
importFrom(mclust,Mclust)
importFrom(mclust,mclustBIC)
importFrom(rlang,.data)
importFrom(stats,aggregate)
importFrom(stats,rbinom)
importFrom(stats,runif)
importFrom(stats,setNames)
importFrom(stringr,str_extract)
importFrom(utils,setTxtProgressBar)
importFrom(utils,tail)
importFrom(utils,txtProgressBar)
importFrom(zoo,as.Date)
importFrom(zoo,as.yearmon)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# ViralEntropR

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/ViralEntropR)](https://CRAN.R-project.org/package=ViralEntropR)
[![R-CMD-check](https://github.com/vadimtyuryaev/ViralEntropR/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/vadimtyuryaev/ViralEntropR/actions/workflows/R-CMD-check.yaml)
[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
![R](https://img.shields.io/badge/R-276DC3?logo=r&logoColor=white)
[![License:
MIT](https://img.shields.io/badge/License-MIT-FFD700.svg)](https://opensource.org/licenses/MIT)
<!-- badges: end -->

**ViralEntropR** is an R package for the efficient preprocessing of
large FASTA archives and computational surveillance of emerging viral
variants. It provides a fully vectorized preprocessing layer, including
header parsing, country and date extraction, ambiguity filtering, and
amino-acid encoding, together with the core computational components for
downstream analysis: per-site Shannon entropy (Shannon 1948),
time-window partitioning, Gaussian-mixture-based entropy clustering
(Everitt et al. 2011; Scrucca et al. 2016), pairwise Hellinger distances
(van der Vaart 1998) between empirical amino-acid distributions, and
non-parametric change-point detection (Fryzlewicz 2014; Matteson and
James 2014). A controlled variant-simulation engine is included for
benchmarking detection performance against known ground truth.

The package was developed alongside a proposed analysis pipeline for
emerging-variant detection, demonstrated in the bundled vignettes:
GMM-driven selection of high-entropy sites, partitioning around medoids
on the Gower distance (Gower 1971; Kaufman and Rousseeuw 1990; Rousseeuw
1987) over the selected sites, t-SNE visualization of cluster geometry
(van der Maaten and Hinton 2008), non-parametric change-point detection
validation, and hierarchical agglomerative clustering (HAC) (Everitt et
al. 2011; Murtagh and Legendre 2014; Sangalli et al. 2010; Tucker et al.
2013) of per-site entropy curves and time-indexed Hellinger-distance
trajectories, treated within a functional-data-analysis framework
(Ramsay and Silverman 2005) as discrete realizations of continuous
functions of time. The pipeline was evaluated on filtered,
post-processed SARS-CoV-2 Spike-protein sequences: 137,132 raw NCBI
records (Sayers et al. 2022) reduced to **109,536** US-filtered
sequences (archived on Zenodo at DOI
[10.5281/zenodo.19040165](https://doi.org/10.5281/zenodo.19040165)), and
16.7 million raw GISAID records (Shu and McCauley 2017) reduced to
**129,371** unique US-filtered sequences. The methods are general enough
to apply to any aligned amino-acid or nucleotide time series.

## Pipeline

                           ┌─────────────────────────┐
                           │       FASTA file        │
                           └─────────────────────────┘
                                        │
                                        ▼
                    ┌───────────────────────────────────────┐
                    │     All sequences of equal width?     │
                    └───────────────────────────────────────┘
                       │                                 │
                      yes                               no
                       │                                 │
                       │                                 ▼
                       │                 ┌───────────────────────────────┐
                       │                 │       Multiple-sequence       │
                       │                 │        alignment, e.g.        │
                       │                 │         msa::msa() /          │
                       │                 │     DECIPHER::AlignSeqs()     │
                       │                 └───────────────────────────────┘
                       │                                 │
                       └────────────────┬────────────────┘
                                        ▼
                        ┌───────────────────────────────┐
                        │     Aligned sequence set      │
                        └───────────────────────────────┘
                                        │
                                        ▼
              ┌───────────────────────────────────────────────────┐
              │                Preprocessing layer                │
              │  ───────────────────────────────────────────────  │
              │           extract_fasta_dates()                   │
              │           extract_fasta_countries()               │
              │           fasta_to_char_matrix()                  │
              │           filter_ambiguous_sequences()            │
              │           encode_aa_sequence()                    │
              └───────────────────────────────────────────────────┘
                                        │
                                        ▼
                    ┌───────────────────────────────────────┐
                    │            Feature matrix:            │
                    │        m sequences  ×  n sites  +     │
                    │               Date  +  Country        │
                    └───────────────────────────────────────┘
                                        │
                                        ▼
              ┌───────────────────────────────────────────────────┐
              │             partition_time_windows()              │
              │                                                   │
              │     cumulative / sliding / disjoint               │
              │     per-window entropies + GMM,                   │
              │     via cluster_sites_by_entropy() internally     │
              └───────────────────────────────────────────────────┘
                                        │
                       ┌────────────────┴────────────────┐
                       ▼                                 ▼
       ┌───────────────────────────────┐ ┌───────────────────────────────┐
       │      Entropy / GMM-class      │ │ calculate_hellinger_matrix()  │
       │         trajectories          │ │                               │
       │                               │ │                               │
       │ relabel_entropy_classes()     │ │     pairwise per-window       │
       │ plot_entropy_trajectories()   │ │     residue-distribution      │
       │ plot_site_class_trajectory()  │ │     distances from the        │
       │ tabulate_site_evolution()     │ │     reference window          │
       └───────────────────────────────┘ └───────────────────────────────┘
                       │                                 │
                       └────────────────┬────────────────┘
                                        ▼
              ┌───────────────────────────────────────────────────┐
              │       Non-parametric change-point detection       │
              │      applied to entropy AND Hellinger series      │
              │  ───────────────────────────────────────────────  │
              │            detect_changepoints_ecp()              │
              │            detect_changepoints_hdcp()             │
              └───────────────────────────────────────────────────┘
                                        │
                       ┌────────────────┴────────────────┐
                       ▼                                 ▼
       ┌───────────────────────────────┐ ┌───────────────────────────────┐
       │   Gower + Silhouette + PAM    │ │   HAC of entropy curves and   │
       │     + t-SNE visualization     │ │    Hellinger trajectories     │
       │                               │ │                               │
       │       cluster::daisy(),       │ │    fdacluster::fdahclust()    │
       │       cluster::pam(),         │ │                               │
       │       Rtsne::Rtsne()          │ │                               │
       │                               │ │                               │
       │  orchestrated in vignettes;   │ │  orchestrated in vignettes;   │
       │  not exported by the package  │ │  not exported by the package  │
       └───────────────────────────────┘ └───────────────────────────────┘
                       │                                 │
                       └────────────────┬────────────────┘
                                        ▼
    ┌───────────────────────────────────────────────────────────────────────┐
    │                           Pipeline outputs                            │
    │  ───────────────────────────────────────────────────────────────────  │
    │    • Post-processed feature matrix                                    │
    │    • Partitioned time windows with GMM-ranked sites per partition     │
    │    • Hellinger distance matrices                                      │
    │    • Clustered sites (candidate variants)                             │
    │    • Change points (variant emergence times)                          │
    │    • Frequency tables + entropy class changes                         │
    │      (sites of high variability / under selection)                    │
    └───────────────────────────────────────────────────────────────────────┘

## Installation

You can install the released version from CRAN:

``` r
install.packages("ViralEntropR")
```

The development version is on GitHub:

``` r
# install.packages("remotes")
remotes::install_github("vadimtyuryaev/ViralEntropR")
```

To build the vignettes locally, install with:

``` r
remotes::install_github("vadimtyuryaev/ViralEntropR", build_vignettes = TRUE)
```

## Quick start

The package ships with a 100-sequence sample of NCBI SARS-CoV-2 Spike
sequences in `inst/extdata/` for runnable examples:

``` r
# This example requires Biostrings (Bioconductor):
# install.packages("BiocManager"); BiocManager::install("Biostrings")

library(ViralEntropR)

# 1. Load the bundled sample (100 NCBI Spike sequences).
sample_path <- system.file("extdata", "sarscov2_sample.fasta.gz",
                           package = "ViralEntropR")
fasta <- Biostrings::readAAStringSet(sample_path)

# 2. Extract dates and countries from FASTA headers.
#    NCBI Virus format: option = 4 (date at end), position = 2 (between pipes).
dates_result     <- extract_fasta_dates(fasta, option = 4)
countries_result <- extract_fasta_countries(fasta, position = 2)
# diagnostic: were any headers unparseable?
dates_result$message       
countries_result$message

# 3. Drop sequences whose date or country could not be parsed.
#    `missing_id` is NA when nothing failed and an integer vector of indices
#    otherwise — strip NAs from the union before use.
ids      <- c(dates_result$missing_id, countries_result$missing_id)
drop_ids <- unique(ids[!is.na(ids)])
keep     <- setdiff(seq_len(length(fasta)), drop_ids)

fasta           <- fasta[keep]
corrected_dates <- dates_result$corrected_dates[keep]
countries       <- countries_result$countries[keep]

# 4. Convert to a character matrix, drop sequences with ambiguous residues
#    (B, J, X, Z), and align metadata to the surviving rows.
char_mat        <- fasta_to_char_matrix(fasta)
filtered        <- filter_ambiguous_sequences(char_mat, option = 2)
keep_idx        <- setdiff(seq_len(nrow(char_mat)), filtered$DeletedSeqId)
corrected_dates <- corrected_dates[keep_idx]
countries       <- countries[keep_idx]

# 5. Integer-encode under the 25-symbol ViralEntropR alphabet.
int_mat <- encode_aa_sequence(filtered$FilteredMatrix)

# 6. Assemble the analysis-ready data frame: sites 1..1273 + Date + Country.
AL_df           <- as.data.frame(int_mat)
colnames(AL_df) <- as.character(seq_len(ncol(int_mat)))
AL_df$Date      <- as.Date(format(corrected_dates, "%Y-%m-01"))
AL_df$Country   <- countries
AL_df           <- AL_df[order(AL_df$Date), ]

# 7. Per-site Shannon entropy + GMM-based site classification.
#    Sites of zero entropy and singletons are excluded. 
ent         <- apply(int_mat, 2, calculate_entropy)
cls         <- cluster_sites_by_entropy(ent, nr = nrow(int_mat))
cls_labeled <- relabel_entropy_classes(cls$DataFrame)
head(cls_labeled, 10)
```

For the full pipeline — including time-window partitioning, Hellinger
distances, change-point detection, and visualization — see the bundled
vignettes.

## Vignettes

Three pre-rendered vignettes walk through the full workflow:

``` r
browseVignettes("ViralEntropR")
```

| Vignette | Topic |
|----|----|
| `preprocessing_pipeline` | Complete preprocessing of raw NCBI SARS-CoV-2 Spike-protein FASTA: two-pass date extraction (`yyyy-mm-dd` then `yyyy-mm`), country extraction from pipe-delimited headers, ambiguity filtering (B/J/X/Z), 25-symbol integer encoding, and assembly of an analysis-ready data frame keyed on `Date` and `Country`. Applied to the full 137,132-sequence NCBI Spike archive on Zenodo. |
| `detecting_variants_simulation` | End-to-end variant detection demonstration on a controlled synthetic benchmark: four variants emerging over 24 months under pairwise competition with stochastic growth; per-partition entropies and Hellinger distances; three complementary change-point methods — `ks.cp3o` (dynamic programming, exact globally optimal), `detect_changepoints_ecp()` (expanding-window for online surveillance), and `e.agglo` (agglomerative hierarchical, K-free) — compared against the known emergence schedule. |
| `clustering_accuracy` | Empirical evaluation of Entropy-GMM-Gower-Silhouette-PAM clustering against a labelled ground truth: wild-type period (May–June 2020) versus Delta-dominant period (July–August 2021); GMM-driven site selection, PAM on the Gower distance with silhouette-based k (optimal number of clusters) selection over k = 2 … 40, 2D t-SNE embedding, precision / recall / F1 at selected k levels, and medoid analysis cross-referenced against the curated mutation catalog of SARS-CoV-2 mutations in `sarscov2_variants`. |

## Analysis

### Completed

Three standalone scripts in `analysis/` (excluded from the installed
package via `.Rbuildignore`; available by cloning the GitHub
repository):

| Script (in `analysis/`) | Topic |
|----|----|
| `Correlation_Analysis/` | Mantel-test study (vegan::mantel, 9,999 permutations) quantifying how faithfully pairwise Gower distances on SARS-CoV-2 Spike sequences are preserved after GMM entropy-based dimensionality reduction, by progressively accumulating GMM entropy classes (k=1,…,N−1, class 1 = highest mean entropy) and comparing the resulting reduced Gower matrix against the full 1,273-site reference; applied independently to NCBI US (109,536 sequences; 8,155 unique sequence are utilized) and GISAID US (129,371 sequences; subsampled to 10,000 unique haplotypes for tractability), with Pearson r at every k. |
| `GISAID_data_preprocessing/` | Reproducible six-stage preprocessing of the GISAID Spike-protein archive: load, date / country extraction, US country and date-window filter, ambiguity filtering, ClustalOmega alignment of short sequences (width \< 1273 aa) merged with full-length sequences, encoding, and final data-frame assembly. |
| `GMM_transformation_study/` | Empirical robustness study of GMM-based site selection under six monotone transformations of per-site entropy against the untransformed baseline; selected sites cross-referenced against VOC SNP catalogues (Alpha / Beta / Gamma / Delta for NCBI; with Omicron added for GISAID data). Conclusion: entropy-based GMM site selection is invariant to the choice of transformation. |
| `Entropy_Partitioning_Study/` | Site-specific Shannon entropy dynamics at variant-defining mutation sites under three temporal partitioning strategies — cumulative (1-month expanding from a fixed origin), sliding 2-month (one-month overlap), and disjoint 2-month — for six variants (Alpha (B.1.1.7), Beta (B.1.351), Gamma (P.1), Delta (B.1.617.2), Omicron (B.1.1.529), D614G). Four output types are produced and each is interpretable: entropy trajectories identify sites whose variability rises or falls together over time, providing the basis for the functional-data-analysis treatment; class-trajectory plots track each site’s GMM class assignment across partitions, with sites under positive selection expected to enter the highest-entropy class earlier and to escalate through classes faster than neutral sites; amino-acid frequency tables record which substitutions appear at each defining SNP position in each partition; and GMM class-assignment tables summarise per-site class labels per partition, with `Nseq` and `Nclust` summary rows giving the per-partition sample size and the number of fitted GMM components. |
| `Sample_Size_Simulation_Study/` | Simulation study quantifying the number of emerging-variant sequences required for the entropy-GMM site-selection step to detect a new variant against a realistic, multi-variant SARS-CoV-2 surveillance background. Four scenarios spanning increasing background complexity (1 to 3 co-circulating variants) and emerging-variant mutational load (non-Omicron versus Omicron-like) are evaluated across a factorial design grid of 8,400 cells with 30 replicates each, yielding 252,000 simulations; results are summarised via censoring-aware Kaplan-Meier analysis, BCa bootstrap confidence intervals, and per-cell detection rates. |

### Forthcoming analyses

Three additional analyses are in progress.

| Script (in `analysis/`) | Topic |
|----|----|
| `FDA_Analysis/` | Hierarchical agglomerative clustering of per-site Shannon entropy curves and pairwise Hellinger distance trajectories within a functional-data-analysis framework, for the five WHO-designated SARS-CoV-2 variants of concern (Alpha (B.1.1.7), Beta (B.1.351), Gamma (P.1), Delta (B.1.617.2), Omicron (B.1.1.529)). Trajectories are treated as discrete realizations of continuous functions of time and clustered using HAC framework. Outputs include animated `.gif` visualizations of the clustered entropy and Hellinger trajectories themselves, showing how cluster membership and curve shape evolve before, at, and after official detection dates for the variants in question. |
| `CP_Detection_Study/` | Systematic evaluation of the two non-parametric change-point detection methods exported by the package — `detect_changepoints_ecp()` and `detect_changepoints_hdcp()` — applied to the post-processed NCBI and GISAID Spike-protein time series. Detected change points are compared against the documented first-detection dates of WHO-designated variants of concern, taken from `sarscov2_variants$Date_First_Detected_US`. Detection accuracy is reported across the three partitioning strategies (cumulative, sliding, disjoint). |

## Data

- **`sarscov2_variants`** — bundled metadata for twelve WHO-designated
  SARS-CoV-2 variants of concern and variants of interest, plus 21
  peer-reviewed literature references with DOIs. Loaded automatically
  with the package; see `?sarscov2_variants`.
- **`inst/extdata/sarscov2_sample.fasta.gz`** — 100 randomly sampled
  NCBI Spike-protein sequences for runnable examples (see the *Quick
  start* above).
- **`data-raw/NCBI_US_unaligned_feature_matrix_1273aa.rds`** — the
  post-processed NCBI feature matrix (US-filtered Spike protein, 109,536
  sequences × 1,273 sites with `Date` and `Country` columns) produced by
  the `preprocessing_pipeline` vignette. Tracked in the repository for
  direct downstream use.
- **Source NCBI Spike-protein archive** — the input FASTA underlying the
  preprocessing pipeline (137,132 sequences, ~181.5 MB uncompressed) is
  archived on Zenodo: <https://doi.org/10.5281/zenodo.19040165>.
- **GISAID data are not redistributed.** The raw GISAID Spike-protein
  archive (`spikeprot0410.fasta`, ~22.6 GB) and the derived
  post-processed feature matrix
  (`GISAID_US_aligned_feature_matrix_1273aa.rds`, 129,371 US-filtered
  sequences) are excluded from the repository and from Zenodo in
  accordance with the [GISAID Database Access
  Agreement](https://gisaid.org/terms-of-use/). Users may reproduce
  these objects independently by registering at
  [GISAID](https://gisaid.org/), downloading the equivalent release, and
  re-running `analysis/GISAID_data_preprocessing/`.

## Installing on R server and computational notes

The vignettes and analysis scripts depend on `kableExtra` for HTML table
rendering. On a fresh Linux R server `install.packages("kableExtra")`
may fail because its dependencies `systemfonts` and `svglite` need
system-level `cairo`, `freetype`, and `fontconfig` development headers.
If installation fails, install those headers first
(`apt install libcairo2-dev libfontconfig1-dev libfreetype-dev libxt-dev`
on Debian / Ubuntu, or the equivalent `-devel` packages on RHEL /
CentOS) and retry; if the server administrator has already installed
`kableExtra` system-wide, no action is needed. `Biostrings` is similarly
required and must be installed from Bioconductor
(`BiocManager::install("Biostrings")`).

Three scripts currently in `analysis/`, as well as scripts in progress
that will appear there, have substantial memory or runtime footprints
and should be run on a server rather than a laptop. Estimated peak RAM
varies between approximately 15 GB and 112 GB depending on the script
and the dataset to which it is applied.

The three pre-rendered vignettes and the `sarscov2_sample.fasta.gz`
quick-start are designed to run on a laptop without any of the above
caveats.

## Citation

If you use **ViralEntropR** in your research, please cite both the
software and the underlying methods:

``` r
citation("ViralEntropR")
```

BibTeX entry:

    @Manual{,
      title = {ViralEntropR: A Computational Pipeline for Entropy-Informed Detection of Emerging Viral Variants},
      author = {Vadim Tyuryaev and Jane Heffernan and Hanna Jankowski},
      note = {R package version 0.6.0},
      url = {https://github.com/vadimtyuryaev/ViralEntropR},
      }

## Acknowledgments

We gratefully acknowledge the authors from the originating laboratories
responsible for obtaining the specimens, and the submitting laboratories
that generated and shared the genetic sequence data via GISAID and NCBI
Virus, on which this research is based.

## License

MIT © Vadim Tyuryaev, Jane Heffernan, Hanna Jankowski. See
[`LICENSE.md`](https://github.com/vadimtyuryaev/ViralEntropR/blob/main/LICENSE.md)
for the full text.

## References

<div id="refs" class="references csl-bib-body hanging-indent"
entry-spacing="0">

<div id="ref-Everitt2011" class="csl-entry">

Everitt, B. S., Landau, S., Leese, M., and Stahl, D. (2011), *Cluster
analysis*, Wiley. <https://doi.org/10.1002/9780470977811>.

</div>

<div id="ref-FerratyVieu2006" class="csl-entry">

Ferraty, F., and Vieu, P. (2006), *Nonparametric functional data
analysis: Theory and practice*, Springer.
<https://doi.org/10.1007/0-387-36620-2>.

</div>

<div id="ref-Fryzlewicz2014" class="csl-entry">

Fryzlewicz, P. (2014), “Wild binary segmentation for multiple
change-point detection,” *The Annals of Statistics*, 42, 2243–2281.
<https://doi.org/10.1214/14-AOS1245>.

</div>

<div id="ref-Gower1971" class="csl-entry">

Gower, J. C. (1971), “A general coefficient of similarity and some of
its properties,” *Biometrics*, 27, 857–871.
<https://doi.org/10.2307/2528823>.

</div>

<div id="ref-Kaufman1990" class="csl-entry">

Kaufman, L., and Rousseeuw, P. J. (1990), *Finding groups in data: An
introduction to cluster analysis*, Wiley.
<https://doi.org/10.1002/9780470316801>.

</div>

<div id="ref-Li2015" class="csl-entry">

Li, X., Jankowski, H., Wang, X., and Heffernan, J. M. (2015), “A method
for clustering hemagglutinin influenza protein sequences,” in *BIOMAT
2014*, World Scientific, pp. 272–285.
<https://doi.org/10.1142/9789814667944_0018>.

</div>

<div id="ref-MattesonJames2014" class="csl-entry">

Matteson, D. S., and James, N. A. (2014), “A nonparametric approach for
multiple change point analysis of multivariate data,” *Journal of the
American Statistical Association*, 109, 334–345.
<https://doi.org/10.1080/01621459.2013.849605>.

</div>

<div id="ref-MurtaghLegendre2014" class="csl-entry">

Murtagh, F., and Legendre, P. (2014), “Ward’s hierarchical agglomerative
clustering method: Which algorithms implement ward’s criterion?”
*Journal of Classification*, 31, 274–295.
<https://doi.org/10.1007/s00357-014-9161-z>.

</div>

<div id="ref-RamsaySilverman2005" class="csl-entry">

Ramsay, J. O., and Silverman, B. W. (2005), *Functional data analysis*,
Springer. <https://doi.org/10.1007/b98888>.

</div>

<div id="ref-Rousseeuw1987" class="csl-entry">

Rousseeuw, P. J. (1987), “Silhouettes: A graphical aid to the
interpretation and validation of cluster analysis,” *Journal of
Computational and Applied Mathematics*, 20, 53–65.
<https://doi.org/10.1016/0377-0427(87)90125-7>.

</div>

<div id="ref-Sangalli2010" class="csl-entry">

Sangalli, L. M., Secchi, P., Vantini, S., and Vitelli, V. (2010),
“K-mean alignment for curve clustering,” *Computational Statistics &
Data Analysis*, 54, 1219–1233.
<https://doi.org/10.1016/j.csda.2009.12.008>.

</div>

<div id="ref-Sayers2022" class="csl-entry">

Sayers, E. W., Bolton, E. E., Brister, J. R., Canese, K., Chan, J.,
Comeau, D. C., Connor, R., Funk, K., Kelly, C., Kim, S., Madej, T.,
Marchler-Bauer, A., Lanczycki, C., Lathrop, S., Lu, Z., Thibaud-Nissen,
F., Murphy, T., Phan, L., Skripchenko, Y., Tse, T., Wang, J., Williams,
R., Trawick, B. W., Pruitt, K. D., and Sherry, S. T. (2022), “Database
resources of the national center for biotechnology information,”
*Nucleic Acids Research*, 50, D20–D26.
<https://doi.org/10.1093/nar/gkab1112>.

</div>

<div id="ref-Scrucca2016" class="csl-entry">

Scrucca, L., Fop, M., Murphy, T. B., and Raftery, A. E. (2016),
“<span class="nocase">mclust</span> 5: Clustering, classification and
density estimation using Gaussian finite mixture models,” *The R
Journal*, 8, 289–317. <https://doi.org/10.32614/RJ-2016-021>.

</div>

<div id="ref-Shannon1948" class="csl-entry">

Shannon, C. E. (1948), “A mathematical theory of communication,” *Bell
System Technical Journal*, 27, 379–423.
<https://doi.org/10.1002/j.1538-7305.1948.tb01338.x>.

</div>

<div id="ref-ShuMcCauley2017" class="csl-entry">

Shu, Y., and McCauley, J. (2017), “GISAID: Global initiative on sharing
all influenza data – from vision to reality,” *Eurosurveillance*, 22,
30494. <https://doi.org/10.2807/1560-7917.ES.2017.22.13.30494>.

</div>

<div id="ref-Sievers2011" class="csl-entry">

Sievers, F., Wilm, A., Dineen, D., Gibson, T. J., Karplus, K., Li, W.,
Lopez, R., McWilliam, H., Remmert, M., Söding, J., Thompson, J. D., and
Higgins, D. G. (2011), “Fast, scalable generation of high-quality
protein multiple sequence alignments using Clustal Omega,” *Molecular
Systems Biology*, 7, 539. <https://doi.org/10.1038/msb.2011.75>.

</div>

<div id="ref-Tucker2013" class="csl-entry">

Tucker, J. D., Wu, W., and Srivastava, A. (2013), “Generative models for
functional data using phase and amplitude separation,” *Computational
Statistics & Data Analysis*, 61, 50–66.
<https://doi.org/10.1016/j.csda.2012.12.001>.

</div>

<div id="ref-Tyuryaev2026" class="csl-entry">

Tyuryaev, V., Jankowski, H., and Heffernan, J. M. (2026), “SARS-CoV-2
surface glycoprotein sequences, NCBI data hub, October 2021
(ViralEntropR archive),” Zenodo.
<https://doi.org/10.5281/zenodo.19040165>.

</div>

<div id="ref-vanderMaaten2008" class="csl-entry">

van der Maaten, L., and Hinton, G. (2008), “[Visualizing data using
t-SNE](http://jmlr.org/papers/v9/vandermaaten08a.html),” *Journal of
Machine Learning Research*, 9, 2579–2605.

</div>

<div id="ref-vanDerVaart1998" class="csl-entry">

van der Vaart, A. W. (1998), *Asymptotic statistics*, Cambridge series
in statistical and probabilistic mathematics, Cambridge University
Press. <https://doi.org/10.1017/CBO9780511802256>.

</div>

</div>

#' @title Extract Countries from FASTA Sequence Names
#' @description Extracts country names from the sequence name strings of an
#'   \code{AAStringSet} object loaded via
#'   \code{\link[Biostrings]{readAAStringSet}}. Handles single-word (e.g.
#'   \code{UK}), hyphenated (e.g. \code{Timor-Leste}), and multi-word (e.g.
#'   \code{United States of America}) country names.
#'
#' @details
#' The function selects one of four regex patterns based on \code{position}
#' and applies it to each sequence name via
#' \code{\link[stringr]{str_extract}}. \strong{Only the first match per
#' header is returned.} If a header contains multiple delimited fields,
#' the country must be in the first such field for the corresponding
#' \code{position} value to extract it correctly. For example, with a
#' GISAID-style header
#' \code{Spike|hCoV-19/USA/OH/.../2021|2021-05-15|EPI_ISL_...|},
#' \code{position = 3} (between slashes) returns \code{USA}, but
#' \code{position = 2} (between pipes) returns \code{hCoV-19/USA/OH/...},
#' not \code{USA}. Inspect representative headers with
#' \code{names(sequence)[1]} before choosing \code{position}.
#'
#' \strong{Encoding.} FASTA files with non-ASCII characters in headers
#' (accented characters, byte-order marks, etc.) can break regex
#' extraction. Setting \code{problematic_characters = TRUE} re-encodes
#' headers to UTF-8 with non-representable bytes escaped, allowing the
#' regex to proceed.
#'
#' @param sequence An \code{AAStringSet} object.
#' @param position Integer (1–4). Location of the country field within the
#'   sequence name string:
#'   \itemize{
#'     \item \strong{1} — text before the first \code{|} (e.g.
#'       \code{SouthKorea|...}).
#'     \item \strong{2} — text between the first and second \code{|}.
#'     \item \strong{3} — text between the first and second \code{/}.
#'     \item \strong{4} — text after the last \code{|}.
#'   }
#' @param problematic_characters Logical. If \code{TRUE}, sequence names are
#'   re-encoded to UTF-8, replacing non-representable bytes with their
#'   escaped form. Useful for FASTA files with non-ASCII characters in
#'   headers. Default is \code{FALSE}.
#'
#' @return A named list with three elements:
#' \item{countries}{Character vector of extracted country strings, one per
#'   sequence. \code{NA} where extraction failed (no match against the
#'   chosen pattern).}
#' \item{message}{A single character string summarising extraction
#'   success.}
#' \item{missing_id}{Integer vector of indices where extraction failed, or
#'   \code{NA} if all extractions succeeded.}
#'
#' @seealso \code{\link{extract_fasta_dates}} for the date-extraction
#'   companion; \code{\link[Biostrings]{readAAStringSet}} for loading the
#'   input \code{AAStringSet}.
#'
#' @importFrom stringr str_extract
#' @export
#'
#' @examples
#' \donttest{
#' path_sample  <- system.file("extdata", "sarscov2_sample.fasta.gz",
#'                              package = "ViralEntropR")
#' fasta_sample <- Biostrings::readAAStringSet(path_sample)
#'
#' # Inspect header structure to confirm field positions before extraction.
#' sample(names(fasta_sample), 1)
#'
#' # Extract countries (position 2 = between first and second pipe).
#' result <- extract_fasta_countries(fasta_sample, position = 2)
#' result$message
#' sort(table(result$countries), decreasing = TRUE)
#' }
extract_fasta_countries <- function(sequence,
                                    position,
                                    problematic_characters = FALSE) {
  
  if (!requireNamespace("Biostrings", quietly = TRUE)) {
    stop("Package 'Biostrings' is required. Install it with:\n",
         "  install.packages('BiocManager')\n",
         "  BiocManager::install('Biostrings')",
         call. = FALSE)
  }
  
  if (!position %in% 1:4) {
    stop("`position` must be an integer between 1 and 4.", call. = FALSE)
  }
  
  # Use names() — the public accessor — instead of internal slot access.
  sqnce <- names(sequence)
  
  if (isTRUE(problematic_characters)) {
    sqnce <- iconv(sqnce, "UTF-8", "UTF-8", sub = "byte")
  }
  
  # Regex patterns. Each pattern matches the FIRST occurrence of the field;
  # see Details for header-format implications.
  pattern <- switch(as.character(position),
                    "1" = "^[^|]+",                # text before the first | (non-empty)
                    "2" = "(?<=\\|)[^|]+(?=\\|)",  # text between pipes |...|
                    "3" = "(?<=\\/)[^\\/]+",       # text between slashes /.../
                    "4" = "(?<=\\|)[^|]+$"         # text after the last |
  )
  
  countries  <- stringr::str_extract(sqnce, pattern)
  is_missing <- is.na(countries)
  
  if (any(is_missing)) {
    msg        <- "Several countries have not been extracted or are missing"
    missing_id <- which(is_missing)
  } else {
    msg        <- "All countries have been extracted"
    missing_id <- NA
  }
  
  list(countries  = countries,
       message    = msg,
       missing_id = missing_id)
}

#' @title Calculate Hellinger Distance Matrix
#' @description Computes the Hellinger distance between the amino acid distribution
#'   of a reference time point (first partition) and all subsequent time points,
#'   for each requested site.
#'
#' @details
#' The Hellinger distance between two discrete distributions \eqn{P} and \eqn{Q} is:
#' \deqn{H(P, Q) = \sqrt{\sum_{i=1}^{k} (\sqrt{p_i} - \sqrt{q_i})^2}}
#' When \code{normalized = TRUE} the result is scaled by \eqn{1/\sqrt{2}}, bounding
#' the distance to \eqn{[0, 1]}. Otherwise the range is \eqn{[0, \sqrt{2}]}.
#'
#' Internally, amino acid counts per site per partition are tabulated using
#' \code{\link{get_site_counts}} (built on \code{\link[base]{tabulate}}).
#' Per-partition proportions and Hellinger distances are then computed by
#' fully vectorised matrix operations — no inner loop over partitions.
#'
#' @param partitions A list of data frames, one per time window. Each data frame
#'   must have numeric-encoded amino acid sequences as columns (integers 1 to
#'   \code{aa_levels}). This is typically the \code{$Partitions} element 
#'   returned by \code{\link{partition_time_windows}}.
#' @param sites Integer vector. Indices of the sites to analyse. Defaults to all
#'   sites (\code{seq_len(ncol(partitions[[1]]))}).
#' @param aa_levels Integer. Alphabet size. Must match the encoding used
#'   when the partitions were created: \code{\link{encode_aa_sequence}}
#'   produces values 1–25 by default (20 standard residues, three
#'   ambiguous codes, \code{*}, \code{-}). Default is \code{25L}.
#' @param normalized Logical. If \code{TRUE}, scales distances by
#'   \eqn{1/\sqrt{2}} to bound the result in \eqn{[0, 1]}. Default is
#'   \code{FALSE}.
#' @param labels Character vector of partition labels. Length must equal
#'   \code{length(partitions)}. The first label names the reference
#'   partition; subsequent labels become column names of the returned
#'   matrix. Defaults to \code{"T1"}, \code{"T2"}, …
#' @param include_freq_tables Logical. If \code{TRUE}, the return value also
#'   includes raw count tables and proportion tables for each site. Default is
#'   \code{FALSE}.
#'
#' @return When \code{include_freq_tables = FALSE} (default): a numeric matrix
#'   with rows corresponding to \code{sites} and columns corresponding to
#'   \code{labels[-1]}, each entry being the Hellinger distance from the
#'   reference partition (\code{labels[1]}) at that site. Row names are
#'   the site indices; column names are taken from \code{labels[-1]}.
#'   With default labels, the reference partition is \code{"T1"} and the
#'   matrix has columns \code{"T2"}, \code{"T3"}, ….
#'
#'   When \code{include_freq_tables = TRUE}: a named list with elements
#'   \code{Sites}, \code{Hellinger_Distances}, \code{Frequency_Tables}, and
#'   \code{Proportions_Tables}.
#'   
#' @seealso
#' \code{\link{partition_time_windows}} for producing temporally
#'   partitioned data, \code{\link{encode_aa_sequence}} for integer
#'   encoding consistent with \code{aa_levels}, and
#'   \code{\link{detect_changepoints_ecp}} or
#'   \code{\link{detect_changepoints_hdcp}} for downstream change-point
#'   detection on the returned matrix.
#'
#' @examples
#' # Toy 3-partition data: site 1 starts homogeneous (all Alanine, 1),
#' # acquires Valine (20) across two later partitions; site 2 stays
#' # constant (all Valine throughout).
#' p1 = data.frame(s1 = c(1, 1, 1, 1, 1),  s2 = c(20, 20, 20, 20, 20))
#' p2 = data.frame(s1 = c(20, 20, 20, 20, 20), s2 = c(20, 20, 20, 20, 20))
#' p3 = data.frame(s1 = c(1, 1, 20, 20, 20), s2 = c(20, 20, 20, 20, 20))
#' parts = list(T1 = p1, T2 = p2, T3 = p3)
#'
#' result = calculate_hellinger_matrix(parts, sites = 1:2)
#' print(result)
#' # Site 1 has nonzero distance in T2 and T3 (composition shift).
#' # Site 2 has zero distance throughout (constant).
#'
#' # With raw frequency tables for inspection.
#' result2 = calculate_hellinger_matrix(parts, sites = 1:2,
#'                                       include_freq_tables = TRUE)
#' print(result2$Frequency_Tables[[1]])
#' @export
calculate_hellinger_matrix = function(partitions,
                                      sites            = seq_len(ncol(partitions[[1]])),
                                      aa_levels        = 25L,
                                      normalized       = FALSE,
                                      labels           = paste0("T", seq_along(partitions)),
                                      include_freq_tables = FALSE) {
  
  # --- 1. Input validation ---------------------------------------------------
  n_partitions = length(partitions)
  if (n_partitions < 2L) stop("At least 2 partitions are required.")
  if (length(labels) != n_partitions)
    stop("`labels` must have the same length as `partitions`.")
  
  norm_factor = if (normalized) (1 / sqrt(2)) else 1
  
  # --- 2. Per-site computation -----------------------------------------------
  # compute_site is fully vectorized over partitions: sweep/colSums operate on
  # the full aa_levels x n_partitions matrix in one call per site.
  compute_site = function(site_idx) {
    counts = get_site_counts(partitions, site_idx, aa_levels)  # aa_levels x n_partitions
    
    # Column-wise proportions; guard against all-zero columns (empty partitions)
    cs           = colSums(counts)
    cs[cs == 0L] = 1L
    props        = sweep(counts, 2L, cs, "/")
    
    p0     = props[, 1L]
    others = props[, -1L, drop = FALSE]
    
    dists = norm_factor * sqrt(colSums((sqrt(others) - sqrt(p0))^2))
    
    list(dists  = dists,
         counts = counts,
         props  = props)
  }
  
  # --- 3. Execute over all sites ---------------------------------------------
  site_results = lapply(sites, compute_site)
  
  # --- 4. Assemble output matrix ---------------------------------------------
  dist_matrix              <- do.call(rbind, lapply(site_results, `[[`, "dists"))
  rownames(dist_matrix)    <- as.character(sites)
  colnames(dist_matrix)    <- labels[-1]          # T2, T3, ...
  storage.mode(dist_matrix) <- "double"           # ensure pure numeric for ecp/ks.cp3o
  
  # --- 5. Return -------------------------------------------------------------
  if (include_freq_tables) {
    list(
      Sites               = sites,
      Hellinger_Distances = dist_matrix,
      Frequency_Tables    = lapply(site_results, `[[`, "counts"),
      Proportions_Tables  = lapply(site_results, `[[`, "props")
    )
  } else {
    dist_matrix
  }
}

#' @title Cluster a Univariate Numeric Vector by Gaussian Mixture Model
#' @description Wraps \code{\link[mclust]{Mclust}} for unsupervised clustering
#'   of a univariate numeric vector, with preprocessing rules and edge-case
#'   handling tailored to per-site Shannon entropy values from viral sequence
#'   data, which is the package's primary use case, but applicable to any
#'   univariate data the user wishes to cluster by GMM.
#'
#' @details
#' In the package's typical use, sites are clustered by their Shannon entropy
#' to identify groups of residue positions with similar variability across
#' a sequence collection. Two preprocessing rules apply when clustering
#' entropies: \code{removez = TRUE} drops invariant sites (entropy = 0), and
#' \code{removesngl = TRUE} drops singleton sites whose entropy corresponds
#' to exactly one differing sequence across \code{nr} rows.
#'
#' Class assignment rules (applied in priority order):
#' \itemize{
#'   \item \strong{No rows remaining after filtering}: empty DataFrame returned
#'     with a zero-length \code{class} column (consistent schema).
#'   \item \strong{Single row remaining}: class \code{1} assigned directly;
#'     Mclust is not called (undefined on 1 observation).
#'   \item \strong{All entropies identical}: class \code{999} for all sites
#'     (sentinel — one undifferentiated group).
#'   \item \strong{Normal Mclust result}: raw class labels \code{1, 2, ..., G}.
#'     These are Mclust's own integer labels, ordered by increasing component
#'     mean (univariate Mclust orders components by mean) — call
#'     \code{relabel_entropy_classes()} on the returned data frame to obtain
#'     application-friendly class labels (highest-entropy class \code{= 1},
#'     lowest-entropy class \code{= G}).
#'     \code{\link{relabel_entropy_classes}} on the returned DataFrame to
#'     standardise so that class 1 = highest-entropy group.
#'   \item \strong{Mclust failure}: empty DataFrame returned (same schema as
#'     the no-rows case), treating the partition as uninformative.
#' }
#'
#' @param entropies Numeric vector to cluster. In the package's primary use
#'   case these are per-site Shannon entropy values, but any univariate
#'   numeric vector is accepted.
#' @param nr Integer. Total number of sequences from which the entropies
#'   were computed. Required only when \code{removesngl = TRUE}.
#' @param nsites Integer. Expected number of sites. If it mismatches
#'   \code{length(entropies)}, the actual length is used with a warning.
#'   Default is \code{length(entropies)}.
#' @param precision Integer. Decimal places for rounding during singleton
#'   threshold comparison and the all-identical uniqueness check. Default
#'   is \code{6}.
#' @param removez Logical. If \code{TRUE}, removes sites with entropy = 0
#'   (invariant sites), using a small tolerance \code{1e-9} to absorb
#'   floating-point near-zeros. Default is \code{TRUE}.
#' @param removesngl Logical. If \code{TRUE}, removes sites whose entropy
#'   equals the singleton value (one differing sequence out of \code{nr}).
#'   Uses tolerance-based comparison. Default is \code{TRUE}.
#' @param transfr A function, or an object of class \code{transform} with a
#'   \code{$transform()} method, applied to entropies before clustering.
#'   Default is \code{NULL} (no transformation).
#' @param verbose Logical. If \code{TRUE}, emits diagnostic warnings for
#'   non-fatal events (empty partitions, Mclust failures, etc.). Default is
#'   \code{FALSE}.
#' @param ... Additional arguments passed to \code{\link[mclust]{Mclust}}.
#'
#' @return A named list with two elements:
#' \item{FitObject}{The raw \code{Mclust} result, or a minimal
#'   \code{list(classification = integer(0L))} when clustering was bypassed
#'   or failed.}
#' \item{DataFrame}{A data frame with columns \code{sites} (original site
#'   indices), \code{entropies} (values after any transformation), and
#'   \code{class} (GMM cluster label). The \code{class} column is
#'   \strong{always} present in every return path, including zero-row
#'   DataFrames. Downstream consumers need only guard on \code{nrow(df) > 0}
#'   before accessing class values. Raw Mclust labels are returned as-is;
#'   call \code{\link{relabel_entropy_classes}} to standardise label ordering.}
#'
#' @seealso
#' \code{\link{calculate_entropy}} for computing per-site entropy values,
#'   \code{\link{relabel_entropy_classes}} for standardising the returned
#'   class labels, and \code{\link{partition_time_windows}}, which calls
#'   this function on each temporal partition.
#'
#' @importFrom mclust Mclust mclustBIC
#' @export
#'
#' @examples
#' # Clear bimodal structure: 5 low-entropy + 5 high-entropy sites.
#' set.seed(42)
#' entropies <- c(rnorm(5, mean = 0.1, sd = 0.01),
#'                rnorm(5, mean = 1.5, sd = 0.1))
#' result <- cluster_sites_by_entropy(entropies, removez = FALSE,
#'                                    removesngl = FALSE)
#' print(result$DataFrame)
#'
#' # Single-row edge case: class = 1 assigned directly.
#' res1 <- cluster_sites_by_entropy(0.35, removesngl = FALSE)
#' print(res1$DataFrame)
#'
#' # All-identical edge case: class = 999 (sentinel, one undifferentiated group).
#' res2 <- cluster_sites_by_entropy(c(0.35, 0.35, 0.35), removesngl = FALSE)
#' print(res2$DataFrame)
cluster_sites_by_entropy <- function(entropies,
                                     nr,
                                     nsites     = length(entropies),
                                     precision  = 6L,
                                     removez    = TRUE,
                                     removesngl = TRUE,
                                     transfr    = NULL,
                                     verbose    = FALSE,
                                     ...) {

  # --- Helper: empty result with consistent schema ---------------------------
  empty_result <- function(df) {
    df$class <- integer(0L)
    list(FitObject = list(classification = integer(0L)),
         DataFrame  = df)
  }

  # --- 1. Validate nsites ----------------------------------------------------
  if (length(entropies) != nsites) {
    if (verbose)
      warning("Length of entropies (", length(entropies),
              ") does not match nsites (", nsites,
              "). Using length(entropies) as nsites.")
    nsites <- length(entropies)
  }

  # --- 2. Build and sort data frame ------------------------------------------
  df <- data.frame(sites = seq_len(nsites), entropies = entropies)
  df <- df[order(df$entropies), , drop = FALSE]

  # --- 3. Remove invariant sites (entropy ~ 0) -------------------------------
  if (removez) {
    df <- df[df$entropies > 1e-9, , drop = FALSE]
  }

  # --- 4. Remove singleton sites ---------------------------------------------
  # Singleton entropy = H when exactly 1 sequence out of nr differs.
  if (removesngl) {
    if (missing(nr))
      stop("`nr` must be provided when `removesngl = TRUE`.")
    p1   <- 1 / nr
    p2   <- (nr - 1) / nr
    sngl <- -(p1 * log2(p1) + p2 * log2(p2))
    tol  <- 10^(-precision)
    df   <- df[abs(df$entropies - sngl) > tol, , drop = FALSE]
  }

  # --- 5. Apply optional transformation --------------------------------------
  if (!is.null(transfr)) {
    if (is.function(transfr)) {
      df$entropies <- transfr(df$entropies)
    } else if (inherits(transfr, "transform")) {
      df$entropies <- transfr$transform(df$entropies)
    } else {
      stop("`transfr` must be a function or an object of class `transform`.")
    }
  }

  # --- 6. Strip non-finite values --------------------------------------------
  # Inf / NaN / NA would otherwise cause Mclust to fail with an opaque error.
  bad <- !is.finite(df$entropies)
  if (any(bad)) {
    if (verbose)
      warning(sum(bad), " non-finite entropy value(s) removed before clustering.")
    df <- df[!bad, , drop = FALSE]
  }

  # --- 7. Edge case: no rows remaining ---------------------------------------
  if (nrow(df) == 0L) {
    if (verbose)
      warning("No sites remaining after filtering; returning empty DataFrame.")
    return(empty_result(df))
  }

  # --- 8. Edge case: single row ----------------------------------------------
  # Mclust is undefined on 1 observation. class = 1 is the only valid answer.
  if (nrow(df) == 1L) {
    df$class <- 1L
    return(list(FitObject = list(classification = 1L),
                DataFrame  = df))
  }

  # --- 9. Edge case: all remaining entropies identical -----------------------
  # Class 999 is the sentinel for an undifferentiated group; preserved as-is
  # by relabel_entropy_classes so the sentinel meaning is never lost.
  if (length(unique(round(df$entropies, precision))) == 1L) {
    df$class <- 999L
    return(list(FitObject = list(classification = rep(999L, nrow(df))),
                DataFrame  = df))
  }

  # --- 10. Cluster with Mclust -----------------------------------------------
  # suppressWarnings: silence Mclust EM convergence warnings during heavy use.
  # capture.output: silence Mclust's internal cat() during BIC search.
  fit <- tryCatch({
    suppressWarnings({
      utils::capture.output(
        fit_inner <- mclust::Mclust(df$entropies, ...)
      )
      fit_inner
    })
  }, error = function(e) {
    if (verbose) warning("Mclust error: ", conditionMessage(e))
    NULL
  })

  # --- 11. Handle genuine Mclust failure -------------------------------------
  if (is.null(fit)) {
    if (verbose)
      warning("Mclust returned NULL; treating partition as uninformative.")
    return(empty_result(df))
  }

  # --- 12. Safety check on classification length ----------------------------
  if (is.null(fit$classification) ||
      length(fit$classification) != nrow(df)) {
    if (verbose)
      warning("Unexpected classification vector; treating partition as uninformative.")
    return(empty_result(df))
  }

  # --- 13. Attach classification — guaranteed present -----------------------
  # Raw Mclust labels (1..G); call relabel_entropy_classes() to standardise.
  df$class <- fit$classification

  list(FitObject = fit, DataFrame = df)
}

#' @title Calculate Shannon Entropy
#' @description Computes the Shannon entropy of a categorical vector.
#'
#' @details
#' Entropy is calculated as \eqn{H(X) = -\sum p(x) \log_b p(x)}, where \eqn{p(x)}
#' is the proportion of observations belonging to category \eqn{x}.
#'
#' @param vctr A vector (character, factor, or integer) representing categorical data.
#' @param base A numeric scalar. The base of the logarithm. Default is 2.
#' @param precision Integer. The number of decimal places to round the result to. Default is 6.
#'
#' @return A numeric scalar representing the entropy. Returns 0 if the vector contains
#' only one unique value or has length 0.
#'
#' @export
#' 
#' @examples
#' seq_vec = c("A", "A", "T", "G", "C", "A")
#' calculate_entropy(seq_vec)
#'
#' # Pure homogeneity
#' calculate_entropy(rep("A", 10))
#' 
calculate_entropy = function(vctr, base = 2, precision = 6) {
  if (length(vctr) < 2) return(0)
  
  # Calculate proportions
  counts = table(vctr)
  probs  = as.vector(counts) / sum(counts)
  
  # Calculate Entropy (handle 0 probabilities implicitly)
  e = -sum(probs * log(probs, base = base))
  
  round(e, precision)
}

#' @title Partition Data into Time Windows
#' @description Splits a data frame of time-stamped sequences into discrete time
#'   windows, computes per-site entropy and Mclust clustering for each window.
#'
#' @details
#' Three windowing strategies are supported via \code{window_type}.
#' Throughout the descriptions below, \code{T} denotes the number of whole
#' months between \code{start_date} and \code{end_date}, and \code{w}
#' denotes \code{window_length}:
#' \itemize{
#'   \item \strong{1 — Cumulative}: Start is fixed; end expands by
#'     \code{w} months each step. Each window includes all prior data
#'     plus one more period. Produces \code{floor(T / w)} chunks.
#'   \item \strong{2 — Sliding}: A window of \code{w} months slides one
#'     month at a time. Produces \code{T - w + 1} chunks.
#'   \item \strong{3 — Disjoint}: Consecutive non-overlapping windows of
#'     \code{w} months. Produces \code{floor(T / w)} chunks. Default.
#' }
#'
#' For example, with \code{T = 12} months and \code{w = 2}: cumulative
#' produces 6 chunks (each progressively larger); sliding produces 11
#' chunks (overlapping, each 2 months wide); disjoint produces 6 chunks
#' (each exactly 2 months, non-overlapping).
#' 
#' \strong{Empty windows.} When a window contains no observations, the
#' corresponding entries in the returned lists carry placeholder values:
#' \code{Entropies} is a zero-vector of length \code{n_sites};
#' \code{Clusters} is a schema-consistent empty result matching
#' \code{cluster_sites_by_entropy}'s empty-input return; \code{Max_Entropy}
#' is \code{NA_integer_}. Downstream consumers can guard on
#' \code{nrow(Partitions[[i]]) > 0} or \code{!is.na(Max_Entropy[i])}.
#'
#' \strong{Column layout requirement.} The function extracts site columns
#' as \code{data[, seq_len(n_sites), drop = FALSE]}. The site columns
#' must therefore occupy positions \code{1} through \code{n_sites}.
#' \code{Date} (and any other metadata columns such as \code{Country})
#' must come after the site columns. A common error is to place
#' \code{Date} first; this will produce incorrect entropy values.
#'
#' @param data Data frame. Must contain a column named \code{Date}
#'   coercible to \code{Date}. Columns \code{1} through \code{n_sites}
#'   must be the numeric site columns; any additional columns (such as
#'   \code{Country} or other per-sequence metadata) may follow.
#' @param n_sites Integer. Number of site columns. Site columns must
#'   occupy positions \code{1} through \code{n_sites} of \code{data};
#'   any other columns (\code{Date}, optionally \code{Country}, etc.)
#'   must come after.
#' @param window_length Integer. Window duration in months. 
#'   Default is \code{1}.
#' @param window_type Integer (1, 2, or 3). Windowing strategy:
#'  \code{1} = Cumulative, \code{2} = Sliding, \code{3} = Disjoint. 
#'  Default is \code{3}.
#' @param start_date Date or character. Window start. Defaults to earliest date
#'   in \code{data}.
#' @param end_date Date or character. Window end (cutoff). Defaults to latest
#'   date in \code{data}.
#' @param date_format Character. Format string for date labels. Default is
#'   \code{"\%b-\%Y"}.
#' @param verbose Logical. If \code{TRUE}, prints a progress bar and completion
#'   summary. Default is \code{FALSE}.
#' @param ... Additional arguments passed to
#'   \code{\link{cluster_sites_by_entropy}}.
#'
#' @return A named list:
#' \item{Partitions}{List of data frame chunks, one per window.}
#' \item{Entropies}{List of numeric entropy vectors, one per window.}
#' \item{Clusters}{List of clustering results from
#'   \code{\link{cluster_sites_by_entropy}}, one per window.}
#' \item{Max_Entropy}{Numeric vector. Maximum cluster class label per window
#'   (equals number of clusters found; \code{NA} if window was empty).}
#' \item{Dates_Labels}{Character vector of window label strings.}
#' \item{N_partitions}{Integer. Total number of windows.}
#' 
#' @seealso \code{\link{calculate_entropy}} for the per-site entropy
#'   computation; \code{\link{cluster_sites_by_entropy}} for the GMM
#'   clustering applied per window; \code{\link{relabel_entropy_classes}}
#'   for standardizing the cluster labels in downstream consumers; and
#'   \code{\link{calculate_hellinger_matrix}} for typical downstream use
#'   of the \code{Partitions} output.
#'
#' @importFrom lubridate %m+% %m-% interval period
#' @importFrom utils txtProgressBar setTxtProgressBar
#'
#' @examples
#' dates <- seq(as.Date("2020-01-01"), as.Date("2020-06-01"), by = "month")
#' df <- data.frame(
#'   s1   = 1L,
#'   s2   = c(rep(1L, 15), rep(2L, 15)),
#'   Date = rep(dates, each = 5)
#' )
#' res <- partition_time_windows(df, n_sites = 2, window_length = 2,
#'                               window_type = 3, verbose = TRUE)
#' print(res$Dates_Labels)
#' print(res$Max_Entropy)
#'
#' @export
partition_time_windows <- function(data,
                                   n_sites,
                                   window_length = 1,
                                   window_type   = 3,
                                   start_date    = NULL,
                                   end_date      = NULL,
                                   date_format   = "%b-%Y",
                                   verbose       = FALSE,
                                   ...) {
  
  # --- 1. Validation ---------------------------------------------------------
  if (!"Date" %in% names(data)) {
    stop("The column with dates is missing or is not properly named.", 
         call. = FALSE)
  }
  
  if (!window_type %in% 1:3)
    stop("`window_type` must be 1 (cumulative), 2 (sliding), or 3 (disjoint).",
         call. = FALSE)
  
  if (n_sites > ncol(data) - 1L)
    stop(sprintf("`n_sites` (%d) exceeds available non-Date columns (%d). ",
                 n_sites, ncol(data) - 1L),
         "Site columns must occupy positions 1 through n_sites.",
         call. = FALSE)
  
  data <- data[order(data$Date), ]
  
  if (is.null(start_date)) start_date <- min(data$Date)
  if (is.null(end_date))   end_date   <- max(data$Date)
  start_date <- as.Date(start_date)
  end_date   <- as.Date(end_date)
  
# Empty-window predicates compare against window boundaries (not raw data
# boundaries) so spurious warnings when start_date or end_date sits within
# one window_length of the data range are avoided.
  first_window_right <- start_date %m+% lubridate::period(window_length, "months")
  
  if (min(data$Date) >= first_window_right)
    warning("`start_date` is earlier than the data by more than one ",
            "window_length; the first window will be empty.", call. = FALSE)
  
  if (window_type != 1L) {
    last_window_left <- end_date %m-% lubridate::period(window_length, "months")
    if (max(data$Date) < last_window_left)
      warning("`end_date` is later than the data by more than one ",
              "window_length; trailing windows will be empty.", call. = FALSE)
    }
  
  start_date_dspl <- format(start_date, format = date_format)
  
  # --- 2. Compute n_chunks per window type -----------------------------------
  total_months <- lubridate::interval(start_date, end_date) %/%
    lubridate::period(1, "months")
  
  if (window_type == 1) {
    n_months <- total_months
    if (n_months %% window_length != 0)
      n_months <- n_months - (n_months %% window_length)
    n_chunks <- n_months / window_length
    
  } else if (window_type == 2) {
    n_chunks <- total_months - window_length + 1
    
  } else {
    n_chunks <- total_months %/% window_length
  }
  
  if (n_chunks < 1L)
    stop(sprintf("Computed n_chunks = %d for window_type %d with window_length = %d ",
                 n_chunks, window_type, window_length),
         sprintf("over a %d-month range (%s to %s). ",
                 total_months, format(start_date), format(end_date)),
         "The date range is too short for the chosen windowing parameters.",
         call. = FALSE)
  
  # --- 3. Per-window processing ----------------------------------------------
  extra_args <- list(...)
  
  # --- 4. Pre-allocate output structures -------------------------------------
  # Writing directly into pre-allocated vectors/lists avoids accumulating a
  # full raw list in memory. Each iteration's temporary objects (chunk, entrp,
  # all_clust) go out of scope immediately after assignment and can be garbage
  # collected, reducing peak RAM by ~40-70% depending on window type.
  Partitions   <- vector("list", n_chunks)
  Entropies    <- vector("list", n_chunks)
  Clusters     <- vector("list", n_chunks)
  Max_Entropy  <- numeric(n_chunks)
  Dates_Labels <- character(n_chunks)
  
  if (verbose) {
    window_type_name <- switch(as.character(window_type),
                               "1" = "cumulative",
                               "2" = "sliding",
                               "3" = "disjoint",
                               "unknown")
    message(sprintf("Partitioning %d %s window%s ...",
                    n_chunks, window_type_name,
                    if (n_chunks == 1L) "" else "s"))
    pb <- utils::txtProgressBar(min = 0, max = n_chunks, style = 3, width = 50)
  }
  
  for (i in seq_len(n_chunks)) {
    
    # Window bounds
    if (window_type == 1) {
      win_start <- start_date
      win_end   <- lubridate::`%m+%`(start_date,
                                     lubridate::period(window_length * i, "months"))
      
    } else if (window_type == 2) {
      win_start <- lubridate::`%m+%`(start_date,
                                     lubridate::period(i - 1, "months"))
      win_end   <- lubridate::`%m+%`(win_start,
                                     lubridate::period(window_length, "months"))
      
    } else {
      win_start <- lubridate::`%m+%`(start_date,
                                     lubridate::period(window_length * (i - 1), "months"))
      win_end   <- lubridate::`%m+%`(win_start,
                                     lubridate::period(window_length, "months"))
    }
    
    chunk <- data[data$Date >= win_start & data$Date < win_end &
                    data$Date <= end_date, , drop = FALSE]
    
    # Date label
    if (window_type == 1) {
      dates_disp <- if (nrow(chunk) > 0)
        format(as.Date(max(chunk$Date)), format = date_format)
      else
        format(win_end, format = date_format)
      label <- paste(start_date_dspl, dates_disp, sep = " - ")
    } else {
      label <- paste(
        format(win_start, format = date_format),
        format(lubridate::`%m+%`(win_end, lubridate::period(-1, "months")),
               format = date_format),
        sep = " - "
      )
    }
    
    # Entropy + clustering
    if (nrow(chunk) > 0) {
      entrp_all <- apply(chunk[, seq_len(n_sites), drop = FALSE], 2,
                         calculate_entropy)
      all_clust <- do.call(cluster_sites_by_entropy,
                           c(list(entrp_all,
                                  nr     = nrow(chunk),
                                  nsites = n_sites),
                             extra_args))
      
      # max_ent = max raw Mclust class label = number of clusters found.
      # Computed before any downstream relabeling.
      # relabel_entropy_classes belongs in the downstream consumer.
      Partitions[[i]]  <- chunk
      Entropies[[i]]   <- entrp_all
      Clusters[[i]]    <- all_clust
      Max_Entropy[i]   <- if (nrow(all_clust$DataFrame) > 0 &&
                              "class" %in% names(all_clust$DataFrame))
        max(as.numeric(all_clust$DataFrame$class), na.rm = TRUE)
      else
        NA_integer_
      Dates_Labels[i]  <- label
      
    } else {
      Partitions[[i]]  <- chunk
      Entropies[[i]]   <- rep(0, n_sites)
      Clusters[[i]] <- list(FitObject = list(classification = integer(0L)),
                            DataFrame = data.frame(sites     = integer(0L),
                                                   entropies = numeric(0L),
                                                   class     = integer(0L)))
      Max_Entropy[i]   <- NA_integer_
      Dates_Labels[i]  <- label
    }
    
    if (verbose) utils::setTxtProgressBar(pb, i)
  }
  
  if (verbose) {
    close(pb)
    message(sprintf("Partitioning complete: %d partition%s generated (%s to %s).",
                    n_chunks,
                    if (n_chunks == 1L) "" else "s",
                    format(start_date, date_format),
                    format(end_date, date_format)))
  }
  
  # --- 5. Return -------------------------------------------------------------
  list(
    Partitions   = Partitions,
    Entropies    = Entropies,
    Clusters     = Clusters,
    Max_Entropy  = Max_Entropy,
    Dates_Labels = Dates_Labels,
    N_partitions = n_chunks
  )
}

#' @title Parse SARS-CoV-2 VOC/VOI Variant Metadata from Excel
#'
#' @description Reads a structured Excel workbook of SARS-CoV-2 Variants of
#'   Concern (VOC) and Variants of Interest (VOI) and returns a named list
#'   containing mutation profiles, nomenclature, temporal detection metadata,
#'   defining SNPs, and a fully citable reference table with interactive
#'   display support.
#'
#' @details
#' Intended to be called once during data preparation via
#' \code{data-raw/sarscov2_variants.R}:
#'
#' \preformatted{
#'   variants_dat  <- readxl::read_excel("SARS_CoV_2_VOC_VOI.xlsx")
#'   variants_list <- get_variants(variants_dat)
#'   saveRDS(variants_list,
#'           file = "inst/extdata/sarscov2_variants.rds")
#' }
#'
#' The saved object can then be loaded anywhere in the package or vignettes:
#'
#' \preformatted{
#'   voc_data <- readRDS(system.file("extdata", "sarscov2_variants.rds",
#'                                   package = "ViralEntropR"))
#' }
#'
#' \strong{Column order of variants in the Excel workbook:}
#' Alpha, Beta, Epsilon, Eta, Iota, Kappa, Delta, Lambda, Gamma, Zeta,
#' Theta, Omicron.
#'
#' \strong{Returned list elements:}
#' \describe{
#'   \item{\code{WHO_Label}}{List. WHO variant label strings (e.g. \code{"Alpha"}).}
#'   \item{\code{Pango_Lineage}}{List. Pango lineage designations.}
#'   \item{\code{GISAID_Clade}}{Character vector. GISAID clade strings.}
#'   \item{\code{Nextstrain_Clade}}{Character vector. Nextstrain clade strings.}
#'   \item{\code{Country_First_Detected}}{List. Country of first documented
#'     detection per variant.}
#'   \item{\code{Date_Earliest_Sample}}{Character vector. Month-Year of the
#'     earliest documented sample per variant.}
#'   \item{\code{Date_First_Detected}}{Character vector. Month-Year of
#'     world-level first detection.}
#'   \item{\code{Date_First_Detected_US}}{Character vector. Month-Year of
#'     first US detection.}
#'   \item{\code{Spike_Mutations}}{List. Spike protein mutation strings read
#'     from the Excel workbook.}
#'   \item{\code{Mutation_Sites}}{List. Integer site positions extracted from
#'     \code{Spike_Mutations}.}
#'   \item{\code{Defining_SNPs}}{List. Canonical defining SNP strings per
#'     variant (\code{NA} where not characterised).}
#'   \item{\code{Defining_SNP_Sites}}{List. Integer positions extracted from
#'     \code{Defining_SNPs}.}
#'   \item{\code{References}}{Named list with three elements:
#'     \code{$data} — data frame of 21 verified references;
#'     \code{$display(variant = NULL)} — interactive \code{\link[DT]{datatable}},
#'     optionally filtered by WHO label;
#'     \code{$cite(variant)} — character vector of formatted citation strings
#'     suitable for manuscript use.}
#' }
#'
#' @param tibble A \code{tibble} or \code{data.frame} produced by
#'   \code{\link[readxl]{read_excel}} from \code{SARS_CoV_2_VOC_VOI.xlsx}.
#'   Column 1 must be the \code{Variant} field column; columns 2 onward are
#'   one column per variant in the order listed above.
#' @param check Logical. If \code{TRUE}, validates that all internal vectors
#'   have equal length and stops with an informative message on mismatch.
#'   Default \code{FALSE}.
#'
#' @return A named list with 13 elements; see Details for full descriptions.
#'
#' @keywords internal
get_variants = function(tibble, check = FALSE) {
  
  if (!requireNamespace("readxl", quietly = TRUE))
    stop("Package 'readxl' is required. Install with: install.packages('readxl')",
         call. = FALSE)
  if (!requireNamespace("DT", quietly = TRUE))
    stop("Package 'DT' is required. Install with: install.packages('DT')",
         call. = FALSE)
  
  n  <- dim(tibble)[1]
  m  <- dim(tibble)[2]
  mm <- m - 1L
  
  # ── Temporal metadata ────────────────────────────────────────────────────────
  # Order follows variant column order in the Excel sheet (see header above).
  
  # Alphabetical order: Alpha, Beta, Delta, Epsilon, Eta, Gamma, Iota, Kappa,
  #                     Lambda, Omicron, Theta, Zeta
  
  earliest_documented_samples <- c(
    "Sep-2020",  # Alpha   -- ref 2
    "May-2020",  # Beta    -- ref 2
    "Oct-2020",  # Delta   -- ref 2
    "Mar-2020",  # Epsilon -- ref 2
    "Nov-2020",  # Eta     -- ref 3
    "Nov-2020",  # Gamma   -- ref 2
    "Nov-2020",  # Iota    -- ref 2
    "Oct-2020",  # Kappa   -- ref 2
    "Aug-2020",  # Lambda  -- ref 5
    "Nov-2021",  # Omicron -- ref 6
    "Jan-2021",  # Theta   -- ref 2
    "Apr-2020"   # Zeta    -- ref 2
  )
  
  detection_date <- c(
    "Dec-2020",  # Alpha   -- ref 3
    "Oct-2020",  # Beta    -- ref 3
    "Dec-2020",  # Delta   -- ref 3
    "Jul-2020",  # Epsilon -- ref 9
    "Nov-2020",  # Eta     -- ref 3
    "Dec-2020",  # Gamma   -- ref 3
    "Nov-2020",  # Iota    -- ref 3
    "Oct-2020",  # Kappa   -- ref 4
    "Dec-2020",  # Lambda  -- ref 11
    "Nov-2021",  # Omicron -- ref 3
    "Feb-2021",  # Theta   -- ref 3
    "Apr-2020"   # Zeta    -- ref 3
  )
  
  US_detection_date <- c(
    "Dec-2020",  # Alpha   -- ref 7
    "Jan-2021",  # Beta    -- ref 8
    "Mar-2021",  # Delta   -- ref 3
    "Jul-2020",  # Epsilon -- ref 9  ---> in 1 of 1247 samples, then October 2020
    "Nov-2020",  # Eta     -- ref 3
    "Jan-2021",  # Gamma   -- ref 12
    "Nov-2020",  # Iota    -- ref 3
    "Feb-2021",  # Kappa   -- ref 10
    "Feb-2021",  # Lambda  -- ref 11
    "Dec-2021",  # Omicron -- ref 13
    "Jan-2021",  # Theta   -- ref 15
    "Oct-2020"   # Zeta    -- ref 14
  )
  
  # -- Defining SNPs ------------------------------------------------------------
  
  defining_snps <- list(
    c("N501Y", "A570D", "P681H", "T716I", "S982A", "D1118H"),                   # Alpha   -- ref 4
    c("D80A", "D215G", "K417N", "N501Y", "E484K", "A701V"),                     # Beta    -- ref 4
    c("T19R", "L452R", "T478K", "P681R", "D950N"),                              # Delta   -- ref 4
    NA,                                                                          # Epsilon -- not in ref 4
    c("Q52R", "E484K", "Q677H", "F888L"),                                        # Eta     -- ref 4
    c("L18F", "T20N", "P26S", "D138Y", "R190S", "K417T",
      "E484K", "N501Y", "H655Y", "T1027I"),                                      # Gamma   -- ref 4
    NA,                                                                          # Iota    -- not in ref 4
    c("E154K", "L452R", "E484Q", "P681R"),                                       # Kappa   -- ref 4
    NA,                                                                          # Lambda  -- not in ref 4
    c("A76V", "T95I", "L212I", "G339D", "S371L","S373P", "S375F",
      "K417N", "N440K", "G446S", "S477N","T478K", "E484A", "Q493R",
      "G496S", "Q498R", "N501Y","Y505H", "T547K", "D614G", "H655Y",
      "N679K", "P681H","N764K", "D796Y", "N856K", "Q954H", "N969K", "L981F"),   # Omicron -- refs 6, 17, 18
    NA,                                                                          # Theta   -- not in ref 4
    NA                                                                           # Zeta    -- not in ref 4
  )
  
  # -- Nomenclature -------------------------------------------------------------
  # Note: GISAID clades are not unique across variants.
  
  GISAID_clade <- c(
    "GRY",         # Alpha
    "GH/501Y.V2",  # Beta
    "G/478K.V1",   # Delta
    "GH/452R.V1",  # Epsilon
    "G/484K.V3",   # Eta
    "GR/501Y.V3",  # Gamma
    "GH/253G.V1",  # Iota
    "G/452R.V3",   # Kappa
    "GR/452Q.V1",  # Lambda
    "GR/484A",     # Omicron
    "GR/1092K.V1", # Theta
    "GR/484K.V2"   # Zeta
  )
  
  Nextstrain_clade <- c(
    "20I/501Y.V1",  # Alpha
    "20H/501Y.V2",  # Beta
    "21A/S:478K",   # Delta
    "21C",          # Epsilon
    "21D/S:484K",   # Eta
    "20J/501Y.V3",  # Gamma
    "21F",          # Iota
    "21B/S:154K",   # Kappa
    "21G",          # Lambda
    "21K",          # Omicron
    "20B/S:265C",   # Theta
    "20B/S:484K"    # Zeta
  )
  
  # -- WHO label, Pango lineage, country -- read from Excel ---------------------
  
  who_label     <- vector(length = mm, mode = "list")
  pango_lineage <- vector(length = mm, mode = "list")
  country       <- vector(length = mm, mode = "list")
  
  for (i in seq_len(mm)) {
    who_label[[i]]     <- colnames(tibble)[i + 1L]
    pango_lineage[[i]] <- as.character(tibble[1L, i + 1L])
    country[[i]]       <- as.character(tibble[2L, i + 1L])
  }
  
  # -- Mutations -- read from Excel (rows 3 onward) -----------------------------
  
  tibble_temp <- tibble[3:n, ]
  mutations   <- vector(length = mm, mode = "list")
  
  for (i in seq_len(mm)) {
    mutations[[i]] <- c(tibble_temp[!is.na(tibble_temp[, i + 1L]), ]$Variant)
  }
  
  # -- Site extraction ----------------------------------------------------------
  
  get_sites <- function(mutation) {
    as.numeric(gsub("[^0-9]", "", mutation))
  }
  
  sites <- lapply(mutations, get_sites)
  
  # -- Reorder Excel-read vectors to alphabetical order ----------------------
  # Excel column order: Alpha, Beta, Epsilon, Eta, Iota, Kappa,
  #                     Delta, Lambda, Gamma, Zeta, Theta, Omicron
  # Alphabetical order: Alpha, Beta, Delta, Epsilon, Eta, Gamma,
  #                     Iota, Kappa, Lambda, Omicron, Theta, Zeta
  # Reorder index (position in Excel for each alphabetical slot):
  alpha_order <- c("Alpha", "Beta", "Delta", "Epsilon", "Eta", "Gamma",
                   "Iota", "Kappa", "Lambda", "Omicron", "Theta", "Zeta")
  excel_labels <- unlist(who_label)
  reorder_idx  <- match(alpha_order, excel_labels)
  
  who_label     <- who_label[reorder_idx]
  pango_lineage <- pango_lineage[reorder_idx]
  country       <- country[reorder_idx]
  mutations     <- mutations[reorder_idx]
  sites         <- sites[reorder_idx]
  
  defining_snps_sites <- lapply(
    defining_snps,
    function(x) if (length(x) == 1L && is.na(x)) NA else get_sites(x)
  )
  
  # -- Integrity check (optional) -----------------------------------------------
  
  if (isTRUE(check)) {
    vec_list <- list(
      who_label, pango_lineage, GISAID_clade, Nextstrain_clade,
      country, earliest_documented_samples, detection_date,
      US_detection_date, mutations, sites, defining_snps, defining_snps_sites
    )
    if (length(unique(sapply(vec_list, length))) != 1L)
      stop("Vectors have unequal lengths -- check Excel sheet dimensions.", call. = FALSE)
  }
  
  # -- Reference table ----------------------------------------------------------
  #
  # Type vocabulary:
  #   "Peer-reviewed article" -- journal article with DOI
  #   "Government report"     -- CDC MMWR public health release
  #   "Surveillance database" -- outbreak.info live database records (backed by
  #                             Gangavarapu et al. 2023, Nature Methods)
  
  refs_data <- data.frame(
    
    ID = 1:21,
    
    Authors = c(
      "Ghosh N, Nandi S, Saha I",                     #  1
      "Abulsoud AI, et al.",                          #  2
      "Aleem A, Akbar Samad AB, Vaqar S",             #  3
      "Zella D, Giovanetti M, Benedetti F, et al.",   #  4
      "Wink PL, Volpato FCZ, Monteiro FL, et al.",    #  5
      "Chatterjee S, Bhattacharya M, et al.",         #  6
      "Washington NL, Gangavarapu K, et al.",         #  7
      "Long SW, Olsen RJ, Christensen PA, et al.",    #  8
      "Zhang W, Davis BD, Chen SS, et al.",           #  9
      "Gangavarapu K, Abdel Latif A, et al.",         # 10
      "Chen C, Shi Q, Dong XP",                       # 11
      "Firestone MJ, Lorentz AJ, Meyer S, et al.",    # 12
      "CDC COVID-19 Response Team",                   # 13
      "Gangavarapu K, Abdel Latif A, et al.",         # 14
      "Gangavarapu K, Abdel Latif A, et al.",         # 15
      "Gangavarapu K, Abdel Latif A, et al.",         # 16
      "Kannan SR, Spratt AN, Sharma K, et al.",       # 17
      "Dhawan M, Saied AA, Mitra S, et al.",          # 18
      "Tao K, Tzou PL, Nouhin J, et al.",             # 19
      "Brookes AJ",                                   # 20
      "Salleh MZ, Derrick JP, Deris ZZ"              # 21
    ),
    
    Year = c(
      2022L,  #  1 -- Ghosh
      2023L,  #  2 -- Abulsoud
      2023L,  #  3 -- Aleem (last updated May 2023; original PMID: 34033342)
      2021L,  #  4 -- Zella
      2022L,  #  5 -- Wink (epub Sep 2021; published in volume Dec 2022)
      2023L,  #  6 -- Chatterjee
      2021L,  #  7 -- Washington
      2021L,  #  8 -- Long
      2021L,  #  9 -- Zhang
      2023L,  # 10 -- Gangavarapu et al. (outbreak.info, Nature Methods 2023)
      2021L,  # 11 -- Chen/Cao
      2021L,  # 12 -- Firestone
      2021L,  # 13 -- CDC MMWR
      2023L,  # 14 -- Gangavarapu (outbreak.info Zeta)
      2023L,  # 15 -- Gangavarapu (outbreak.info Theta)
      2023L,  # 16 -- Gangavarapu (outbreak.info Lambda)
      2022L,  # 17 -- Kannan (epub Dec 2021; published 2022)
      2022L,  # 18 -- Dhawan
      2021L,  # 19 -- Tao
      1999L,  # 20 -- Brookes
      2021L   # 21 -- Salleh
    ),
    
    Title = c(
      "A review on evolution of emerging SARS-CoV-2 variants based on spike glycoprotein",
      "Mutations in SARS-CoV-2: Insights on structure, variants, vaccines, and biomedical interventions",
      "Emerging Variants of SARS-CoV-2 and Novel Therapeutics Against Coronavirus (COVID-19)",
      "The variants question: What is the problem?",
      "First identification of SARS-CoV-2 lambda (C.37) variant in Southern Brazil",
      "A Detailed Overview of SARS-CoV-2 Omicron: Its Sub-Variants, Mutations and Pathophysiology, Clinical Characteristics, Immunological Landscape, Immune Escape, and Therapies",
      "Emergence and rapid transmission of SARS-CoV-2 B.1.1.7 in the United States",
      "Sequence Analysis of 20,453 Severe Acute Respiratory Syndrome Coronavirus 2 Genomes from the Houston Metropolitan Area Identifies the Emergence and Widespread Distribution of Multiple Isolates of All Major Variants of Concern",
      "Emergence of a Novel SARS-CoV-2 Variant in Southern California",
      "Kappa (B.1.617.1) Variant Report",
      "SARS-CoV-2 Lambda Variant: Spatiotemporal Distribution and Potential Public Health Impact",
      "First Identified Cases of SARS-CoV-2 Variant P.1 in the United States - Minnesota, January 2021",
      "SARS-CoV-2 B.1.1.529 (Omicron) Variant - United States, December 1-8, 2021",
      "Zeta (P.2) Variant Report",
      "Theta (P.3) Variant Report",
      "Lambda (C.37) Variant Report",
      "Omicron SARS-CoV-2 variant: Unique features and their impact on pre-existing antibodies",
      "Omicron variant (B.1.1.529) and its sublineages: What do we know so far amid the emergence of recombinant variants of SARS-CoV-2?",
      "The biological and clinical significance of emerging SARS-CoV-2 variants",
      "The essence of SNPs",
      "Structural Evaluation of the Spike Glycoprotein Variants on SARS-CoV-2 Transmission and Immune Evasion"
    ),
    
    Journal_Source = c(
      "International Immunopharmacology",              #  1
      "Biomedicine & Pharmacotherapy",                 #  2
      "StatPearls [Internet], StatPearls Publishing",  #  3
      "Journal of Medical Virology",                   #  4
      "Infection Control & Hospital Epidemiology",     #  5
      "Viruses",                                       #  6
      "Cell",                                          #  7
      "The American Journal of Pathology",             #  8
      "JAMA",                                          #  9
      "outbreak.info",                                 # 10
      "Zoonoses",                                      # 11
      "MMWR Morb Mortal Wkly Rep",                     # 12
      "MMWR Morb Mortal Wkly Rep",                     # 13
      "outbreak.info",                                 # 14
      "outbreak.info",                                 # 15
      "outbreak.info",                                 # 16
      "Journal of Autoimmunity",                       # 17
      "Biomedicine & Pharmacotherapy",                 # 18
      "Nature Reviews Genetics",                       # 19
      "Gene",                                          # 20
      "International Journal of Molecular Sciences"    # 21
    ),
    
    Volume_Issue_Pages = c(
      "105:108565",            #  1 -- Ghosh
      "157:113977",            #  2 -- Abulsoud
      NA_character_,           #  3 -- StatPearls (no volume/pages)
      "93(12):6479-6485",      #  4 -- Zella
      "43(12):1996-1997",      #  5 -- Wink
      "15(1):167",             #  6 -- Chatterjee
      "184(10):2587-2594.e7",  #  7 -- Washington
      "191(6):983-992",        #  8 -- Long
      "325(13):1324-1326",     #  9 -- Zhang
      NA_character_,           # 10 -- outbreak.info (database record)
      "1(1)",                  # 11 -- Chen/Cao, Zoonoses vol 1 issue 1
      "70(10):346-347",        # 12 -- Firestone MMWR
      "70(50):1731-1734",      # 13 -- CDC MMWR
      NA_character_,           # 14 -- outbreak.info
      NA_character_,           # 15 -- outbreak.info
      NA_character_,           # 16 -- outbreak.info
      "126:102779",            # 17 -- Kannan
      "154:113522",            # 18 -- Dhawan
      "22(12):757-773",        # 19 -- Tao
      "234(2):177-186",        # 20 -- Brookes
      "22(14):7425"            # 21 -- Salleh
    ),
    
    DOI = c(
      "10.1016/j.intimp.2022.108565",   #  1 -- Ghosh
      "10.1016/j.biopha.2022.113977",   #  2 -- Abulsoud
      NA_character_,                    #  3 -- StatPearls (PMID: 34033342; no DOI)
      "10.1002/jmv.27196",              #  4 -- Zella
      "10.1017/ice.2021.390",           #  5 -- Wink
      "10.3390/v15010167",              #  6 -- Chatterjee
      "10.1016/j.cell.2021.03.052",     #  7 -- Washington
      "10.1016/j.ajpath.2021.03.004",   #  8 -- Long
      "10.1001/jama.2021.1612",         #  9 -- Zhang
      "10.1038/s41592-023-01769-3",     # 10 -- Gangavarapu (outbreak.info)
      "10.15212/ZOONOSES-2021-0009",    # 11 -- Chen/Cao
      "10.15585/mmwr.mm7010e1",         # 12 -- Firestone
      "10.15585/mmwr.mm7050e1",         # 13 -- CDC MMWR
      "10.1038/s41592-023-01769-3",     # 14 -- Gangavarapu (outbreak.info Zeta)
      "10.1038/s41592-023-01769-3",     # 15 -- Gangavarapu (outbreak.info Theta)
      "10.1038/s41592-023-01769-3",     # 16 -- Gangavarapu (outbreak.info Lambda)
      "10.1016/j.jaut.2021.102779",     # 17 -- Kannan
      "10.1016/j.biopha.2022.113522",   # 18 -- Dhawan
      "10.1038/s41576-021-00408-x",     # 19 -- Tao
      "10.1016/S0378-1119(99)00219-X",  # 20 -- Brookes
      "10.3390/ijms22147425"            # 21 -- Salleh
    ),
    
    URL = c(
      "https://www.sciencedirect.com/science/article/pii/S1567576922000492",              #  1
      "https://www.sciencedirect.com/science/article/pii/S075333222201366X",              #  2
      "https://pubmed.ncbi.nlm.nih.gov/34033342/",                                        #  3
      "https://onlinelibrary.wiley.com/doi/full/10.1002/jmv.27196",                       #  4
      "https://pmc.ncbi.nlm.nih.gov/articles/PMC8564022/",                                #  5
      "https://www.mdpi.com/1999-4915/15/1/167",                                          #  6
      "https://www.sciencedirect.com/science/article/pii/S0092867421003834",              #  7
      "https://www.sciencedirect.com/science/article/pii/S0002944021001085",              #  8
      "https://jamanetwork.com/journals/jama/fullarticle/2776543",                        #  9
      "https://outbreak.info/situation-reports/kappa",                                    # 10
      "https://www.scienceopen.com/hosted-document?doi=10.15212/ZOONOSES-2021-0009",      # 11
      "https://www.cdc.gov/mmwr/volumes/70/wr/mm7010e1.htm",                              # 12
      "https://www.cdc.gov/mmwr/volumes/70/wr/mm7050e1.htm",                              # 13
      "https://outbreak.info/situation-reports/Zeta",                                     # 14
      "https://outbreak.info/situation-reports/Theta",                                    # 15
      "https://outbreak.info/situation-reports/lambda",                                   # 16
      "https://www.sciencedirect.com/science/article/pii/S0896841121001876",              # 17
      "https://www.sciencedirect.com/science/article/pii/S0753332222009118",              # 18
      "https://www.nature.com/articles/s41576-021-00408-x",                               # 19
      "https://www.sciencedirect.com/science/article/pii/S037811199900219X",               # 20
      "https://www.mdpi.com/1422-0067/22/14/7425"                                          # 21
    ),
    
    Type = c(
      "Peer-reviewed article",  #  1
      "Peer-reviewed article",  #  2
      "Peer-reviewed article",  #  3 
      "Peer-reviewed article",  #  4
      "Peer-reviewed article",  #  5
      "Peer-reviewed article",  #  6
      "Peer-reviewed article",  #  7
      "Peer-reviewed article",  #  8
      "Peer-reviewed article",  #  9
      "Surveillance database",  # 10
      "Peer-reviewed article",  # 11
      "Government report",      # 12
      "Government report",      # 13
      "Surveillance database",  # 14
      "Surveillance database",  # 15
      "Surveillance database",  # 16
      "Peer-reviewed article",  # 17
      "Peer-reviewed article",  # 18
      "Peer-reviewed article",  # 19
      "Peer-reviewed article",  # 20
      "Peer-reviewed article"   # 21
    ),
    
    Variants_Covered = c(
      "Alpha, Beta, Epsilon, Eta, Iota, Kappa, Delta, Lambda, Gamma, Zeta, Theta, Omicron",    #  1
      "Alpha, Beta, Epsilon, Eta, Iota, Kappa, Delta, Lambda, Gamma, Zeta, Theta, Omicron",    #  2
      "Alpha, Beta, Epsilon, Eta, Iota, Kappa, Delta, Lambda, Gamma, Zeta, Theta, Omicron",    #  3
      "Alpha, Beta, Epsilon, Eta, Iota, Kappa, Delta, Lambda, Gamma, Zeta, Theta",             #  4
      "Lambda",                                                                                #  5
      "Omicron",                                                                               #  6
      "Alpha",                                                                                 #  7
      "Alpha, Beta, Epsilon, Gamma, Zeta",                                                     #  8
      "Epsilon",                                                                               #  9
      "Kappa",                                                                                 # 10
      "Lambda",                                                                                # 11
      "Gamma",                                                                                 # 12
      "Omicron",                                                                               # 13
      "Zeta",                                                                                  # 14
      "Theta",                                                                                 # 15
      "Lambda",                                                                                # 16
      "Omicron",                                                                               # 17
      "Omicron",                                                                               # 18
      "Alpha, Beta, Delta, Gamma",                                                             # 19
       NA_character_,                                                                          # 20
      "Alpha, Beta, Delta, Epsilon, Eta, Gamma, Iota, Kappa, Lambda, Theta, Zeta"              # 21
    ),
    
    Data_Field = c(
      "Mutations, Sites, GISAID_clade, Nextstrain_clade, M-Y_Earliest_Documented_Sample",         #  1 -- Ghosh: phylogenetic analysis of 77,681 genomes; GISAID clade assignments; variant emergence timeline
      "General review of SARS-CoV-2 mutations and variant biology",                               #  2 -- Abulsoud: broad review; does not provide the specific dated records used in this list
      "M-Y_Detection, M-Y_Earliest_Documented_Sample (Eta), M-Y_Detection_US (Eta, Iota, Delta)", #  3 -- Aleem StatPearls: archived review; primary source for world detection dates and US detection dates lacking dedicated primary studies
      "Defining_SNPs, Defining_SNPs_Sites, M-Y_Detection (Kappa)",                                #  4 -- Zella: VOC/VOI mutational characterization; also source for Kappa world detection date (Oct-2020)
      "Lambda emergence context",                                                                 #  5 -- Wink: first Lambda identification in Southern Brazil mid-2021; supports Lambda spread narrative
      "Defining_SNPs, Defining_SNPs_Sites, M-Y_Earliest_Documented_Sample (Omicron)",             #  6 -- Chatterjee: detailed Omicron overview; source for Omicron earliest documented sample (Nov-2021); Omicron world detection date now sourced from ref 3
      "M-Y_Detection_US (Alpha)",                                                                 #  7 -- Washington: B.1.1.7 US emergence and logistic growth analysis
      "M-Y_Detection_US (Beta)",                                                                  #  8 -- Long: B.1.351 identified in Houston (replaces SC DHEC press release)
      "M-Y_Detection_US, M-Y_Detection (Epsilon)",                                                #  9 -- Zhang: Epsilon (B.1.427/B.1.429) first identified in Southern California; US origin means US detection = world detection (Jul-2020)
      "M-Y_Detection_US (Kappa)",                                                                 # 10 -- outbreak.info Kappa live report
      "M-Y_Detection, M-Y_Detection_US (Lambda)",                                                 # 11 -- Chen/Cao: spatiotemporal distribution; source for Lambda world detection (Dec-2020) and first documented US presence (Feb-2021)
      "M-Y_Detection_US (Gamma)",                                                                 # 12 -- Firestone: first P.1 cases in Minnesota, January 2021
      "M-Y_Detection_US (Omicron)",                                                               # 13 -- CDC MMWR: Omicron US December 1-8, 2021
      "M-Y_Detection_US (Zeta)",                                                                  # 14 -- outbreak.info Zeta live report
      "M-Y_Detection_US (Theta)",                                                                 # 15 -- outbreak.info Theta live report
      "M-Y_Earliest_Documented_Sample (Lambda)",                                                  # 16 -- outbreak.info Lambda: earliest GISAID record Aug-2020 (Peru); Feb-2021 first appearance in US databases
      "Defining_SNPs, Defining_SNPs_Sites (Omicron)",                                             # 17 -- Kannan: 46 high-prevalence Omicron mutations; structural antibody impact
      "Defining_SNPs, Defining_SNPs_Sites (Omicron)",                                             # 18 -- Dhawan: Omicron sublineages and recombinant variants
      "General evolutionary context",                                                             # 19 -- Tao: biological and clinical significance of SARS-CoV-2 VOCs
      "SNP methodology",                                                                          # 20 -- Brookes: foundational SNP definition and population genetics methodology
      "GISAID_Clade, Nextstrain_Clade"                                                            # 21 -- Salleh: structural evaluation of spike variants; source for GISAID/Nextstrain clade assignments
    ),
    
    stringsAsFactors = FALSE
  )
  
  # Assemble formatted citation string per row for manuscript use.
  # Format: Authors (Year). Title. Journal. Vol(Issue):Pages. doi:DOI
  # Components absent for StatPearls, outbreak.info, etc. are omitted gracefully.
  refs_data$Citation <- mapply(function(auth, yr, ttl, jrnl, vip, doi) {
    parts <- c(
      sprintf("%s (%d).", auth, yr),
      sprintf("%s.", ttl),
      if (!is.na(jrnl)) sprintf("%s.", jrnl),
      if (!is.na(vip))  sprintf("%s.", vip),
      if (!is.na(doi))  sprintf("doi:%s", doi)
    )
    paste(parts, collapse = " ")
  },
  refs_data$Authors, refs_data$Year, refs_data$Title,
  refs_data$Journal_Source, refs_data$Volume_Issue_Pages, refs_data$DOI,
  SIMPLIFY = TRUE)
  
  # -- Reference accessor functions ---------------------------------------------
  
  # display(variant = NULL)
  #   Renders an interactive, filterable DT table.
  #   Supply a WHO label (e.g. "Alpha") to show only references for that variant.
  display_refs <- function(variant = NULL) {
    
    if (!requireNamespace("DT", quietly = TRUE))
      stop("Package 'DT' is required to display the reference table.\n",
           "Install it with: install.packages('DT')",
           call. = FALSE)
    
    df <- refs_data
    
    if (!is.null(variant)) {
      keep <- grepl(variant, df$Variants_Covered, fixed = TRUE)
      if (!any(keep))
        stop(sprintf("No references found for variant '%s'.", variant), 
             call. = FALSE)
      df <- df[keep, ]
    }
    
    df$Title_Link <- sprintf(
      '<a href="%s" target="_blank">%s</a>', df$URL, df$Title
    )
    
    display_cols <- c("ID", "Authors", "Year", "Title_Link",
                      "Journal_Source", "Volume_Issue_Pages",
                      "Type", "Variants_Covered", "Data_Field")
    col_names    <- c("ID", "Authors", "Year", "Title",
                      "Journal / Source", "Vol(Issue):Pages",
                      "Type", "Variants Covered", "Data Field")
    
    caption_text <- if (!is.null(variant))
      sprintf("References for variant: %s", variant)
    else
      "ViralEntropR - SARS-CoV-2 VOC/VOI Reference Table (21 verified references)"
    
    DT::datatable(
      df[, display_cols],
      colnames  = col_names,
      escape    = FALSE,
      rownames  = FALSE,
      filter    = "top",
      caption   = caption_text,
      options   = list(
        paging  = FALSE,       # show all rows -- no pagination (fixes Rmd truncation)
        scrollX = TRUE,
        dom     = "fti",       # filter + table + info; no pagination controls
        columnDefs = list(
          list(width = "240px", targets = 3L),  # Title
          list(width = "180px", targets = 7L),  # Variants Covered
          list(width = "200px", targets = 8L)   # Data Field
        )
      )
    )
  }
  
  # cite(variant)
  #   Returns a character vector of formatted citation strings for a given
  #   WHO variant label -- suitable for copy-pasting into a manuscript.
  #   Example: variants_list$References$cite("Omicron")
  cite_refs <- function(variant) {
    keep <- grepl(variant, refs_data$Variants_Covered, fixed = TRUE)
    if (!any(keep))
      stop(sprintf("No references found for variant '%s'.", variant), call. = FALSE)
    refs_data$Citation[keep]
  }
  
  # -- Assemble and return ------------------------------------------------------
  #
  # All existing list elements are preserved exactly.
  # References is a named list with three elements:
  #   $data    -- the full reference data frame (21 rows x 11 columns)
  #   $display -- function: renders interactive DT, optionally filtered by variant
  #   $cite    -- function: returns formatted citation strings for a variant
  
  list(
    WHO_Label              = who_label,
    Pango_Lineage          = pango_lineage,
    GISAID_Clade           = GISAID_clade,
    Nextstrain_Clade       = Nextstrain_clade,
    Country_First_Detected = country,
    Date_Earliest_Sample   = earliest_documented_samples,
    Date_First_Detected    = detection_date,
    Date_First_Detected_US = US_detection_date,
    Spike_Mutations        = mutations,
    Mutation_Sites         = sites,
    Defining_SNPs          = defining_snps,
    Defining_SNP_Sites     = defining_snps_sites,
    References             = list(
      data    = refs_data,
      display = display_refs,
      cite    = cite_refs
    )
  )
}

#' @title Simulate Viral Variant Evolution
#' @description Generates a synthetic, ground-truth-traceable time series of
#'   viral amino-acid sequences in which a reference strain is progressively
#'   challenged by emerging variants under stochastic growth and pairwise
#'   competition. Designed as a controllable signal generator for benchmarking
#'   the entropy, Hellinger, and change-point components of the
#'   \pkg{ViralEntropR} pipeline; the dynamics are deliberately simplified
#'   rather than a faithful biological model of evolutionary fitness.
#' @export
#'
#' @details
#' Simulates the trajectory of a viral population over discrete monthly time
#' steps. All stochastic components are reproducible under a fixed
#' \code{seed} and are designed to expose a known emergence pattern that
#' downstream detection methods can be evaluated against.
#'
#' \strong{Simulation Phases:}
#' \enumerate{
#'   \item \strong{Reference Phase:} A stable period in which only the
#'     reference strain exists. Reference months are resampled with
#'     replacement to assemble \code{n_ref_months} monthly batches.
#'   \item \strong{Variant Emergence:} New variants are introduced at the
#'     monthly intervals specified by \code{variant_intervals}.
#'   \item \strong{Pairwise Growth:} Each month the newest variant grows
#'     alongside its immediate predecessor. Both grow stochastically: each
#'     variant's next-month count is its previous-month count times
#'     \code{Uniform(mult*(1-variability), mult*(1+variability))}. The two
#'     are not coupled by a shared frequency constraint — only by the hard
#'     monthly quota
#'     \code{ceiling(n_sequences_total / (n_periods - n_ref_months))} that
#'     trims any overshoot. Older, non-paired variants enter the fill pool
#'     with sampling weights \code{2^(i+1)}.
#'   \item \strong{Capped "Deleterious" Sites:} A Bernoulli-random subset of
#'     each variant's mutated positions is flagged with probability
#'     \code{deleterious_rate}. At each post-emergence month the count of
#'     variant rows still carrying the flagged allele is capped at
#'     \code{n_deleterious_limit} by reverting overflow sequences to the
#'     reference allele \emph{at the flagged site only}. This is a
#'     per-site cap, not a genotype-level fitness penalty.
#'   \item \strong{Random Substitution Events:} Each month independently
#'     fires a Bernoulli coin flip with probability
#'     \code{prob_deletion_event}. If it fires, the last
#'     \code{n_rows_to_delete} rows of that month's batch receive a single
#'     non-reference amino-acid substitution at a randomly chosen site
#'     (sites 1 and 2 excluded), and those rows are flagged with
#'     \code{Delet = "Yes"}. Parameter names retain "deletion" / "delete"
#'     for backward compatibility; the operation itself is a substitution.
#' }
#'
#' @param ref_sequences Either a single character string giving the reference
#'   amino-acid sequence, or a data.frame with a \code{Date} column and site
#'   columns named with integer position labels (\code{"1"}, \code{"2"},
#'   ..., \code{as.character(L)}; one column per site). The sequence length
#'   \code{L} must be at least 3, since variant mutations and random
#'   substitution events are sampled from sites \eqn{>2} (sites 1 and 2 are
#'   reserved for reference variability when \code{ref_variability = TRUE}).
#' @param n_ref_months Integer. Duration of the initial reference phase
#'   (months). Default \code{3}.
#' @param start_date Character or Date. Simulation start.
#' @param end_date Character or Date. Simulation end (inclusive month).
#' @param variants_config Integer vector. Number of mutations per variant,
#'   e.g., \code{c(2, 6)}.
#' @param variant_intervals Integer vector. Months between consecutive
#'   variant emergences (length \code{length(variants_config) - 1}).
#' @param n_new_mutations Integer. Number of de-novo sequence \emph{copies}
#'   introduced at a variant's first appearance. Despite the name, this
#'   controls the count of seeded sequences, not the number of mutations
#'   carried by the variant (which is set per-variant by
#'   \code{variants_config}). Default \code{1}.
#' @param mutation_rate Numeric scalar or vector. Monthly growth multiplier
#'   per variant. Default \code{2}.
#' @param mutation_rate_variability Numeric in \code{[0, 1)}. Fractional
#'   spread around \code{mutation_rate}: growth is drawn from
#'   \code{Uniform(mult*(1-v), mult*(1+v))}. Set to \code{0} for
#'   deterministic growth. Default \code{0.25}.
#' @param deleterious_rate Numeric in \code{[0, 1]}. Per-mutation probability
#'   of being flagged as "deleterious". Default \code{0}.
#' @param n_deleterious_limit Integer. Per-site cap on the number of variant
#'   rows allowed to retain a flagged ("deleterious") allele in a given
#'   month; overflow is reverted to the reference allele at the flagged
#'   site only. See \strong{Capped "Deleterious" Sites} in Details.
#'   Default \code{1}.
#' @param n_sequences_total Integer. Total sequences generated over the full
#'   mutation phase (spread evenly per month). Default \code{140}.
#' @param ref_variability Logical. If \code{TRUE}, introduces low-level
#'   variability at sites 1 and 2 of the reference pool. Requires
#'   \code{n_ref_sequences >= 6 * n_ref_months} for the variability rows
#'   to be correctly captured by the multi-variant fill pool; smaller
#'   values trigger a warning. Default \code{FALSE}.
#' @param n_ref_sequences Integer. When \code{ref_sequences} is a single
#'   string, the number of reference rows to create. Default \code{100}.
#' @param prob_deletion_event Numeric in \code{[0, 1]}. Per-month probability
#'   of triggering a substitution event affecting \code{n_rows_to_delete}
#'   rows. (Parameter name retained for backward compatibility; the
#'   operation is substitution, not literal deletion. See Details.)
#'   Default \code{0}.
#' @param n_rows_to_delete Integer. Number of rows affected when a
#'   substitution event fires (see \code{prob_deletion_event}). Default
#'   \code{0}.
#' @param seed Integer or \code{NULL}. Random seed for reproducibility.
#'   Default \code{NULL}.
#'
#' @return An object of class \code{"viralSim"}, a named list:
#' \item{Simulation_Output}{Data frame of all sequences, one column per site
#'   (named \code{"1"}, ..., \code{as.character(L)}) plus \code{Variant}
#'   (per-row strain label), \code{Phase} (\code{"R"} or \code{"RV"} for
#'   reference rows depending on whether the row matches the wild-type
#'   sequence; \code{"M1"}, \code{"M2"}, ... in mutation-phase months —
#'   note that \code{"Mn"} is a \emph{month label} and applies to every
#'   row in that month regardless of strain), \code{Date}, \code{Period}
#'   (1-indexed month), and \code{Delet} (\code{"Yes"} for rows affected
#'   by substitution events, \code{"No"} otherwise).}
#' \item{Variant_Details}{List of per-variant metadata. Each element has
#'   \code{em} (emergence-month index), \code{pos} (mutated site indices),
#'   \code{flags} (logical vector of "deleterious" flags per mutated site),
#'   \code{vseq} (full mutated reference sequence as character vector),
#'   \code{mult} (growth multiplier), \code{last} (count in the most recent
#'   month), \code{cum} (cumulative count), and \code{repl} (logical flag
#'   set when the variant has been displaced by its successor).}
#' \item{Simulation_Dates}{Date vector of all simulated months.}
#' \item{Baseline_Ref_Sequence}{Character. The wild-type reference string.}
#' \item{Delet_Records}{Named list of substitution events keyed by date
#'   (\code{"YYYY-MM-DD"}). Each entry holds \code{period}, \code{date},
#'   \code{site}, \code{old_aa}, \code{new_aa}, and \code{rows} (absolute
#'   row indices in \code{Simulation_Output} that were modified).}
#' \item{Pool}{Fill-pool data frame from the final multi-variant competition
#'   period (includes \code{._weight} column); \code{NULL} if at most one
#'   variant was ever active.}
#'
#' @importFrom stats rbinom runif
#' @importFrom utils tail
#'
#' @seealso \code{\link{partition_time_windows}},
#'   \code{\link{calculate_hellinger_matrix}},
#'   \code{\link{detect_changepoints_ecp}},
#'   \code{\link{detect_changepoints_hdcp}}
#'
#' @examples
#' ref_seq <- "MKTIIALSYIFCLVFADYKDDDDK"
#'
#' sim <- simulate_variant_evolution(
#'   ref_sequences   = ref_seq,
#'   n_ref_months    = 3,
#'   start_date      = "2021-01-01",
#'   end_date        = "2021-12-01",
#'   variants_config = c(3, 5),
#'   variant_intervals = c(4),
#'   n_sequences_total = 50,
#'   mutation_rate   = 1.5,
#'   seed            = 123
#' )
#' head(sim$Simulation_Output)
simulate_variant_evolution <- function(
    ref_sequences,
    n_ref_months             = 3L,
    start_date,
    end_date,
    variants_config,
    variant_intervals,
    n_new_mutations          = 1L,
    mutation_rate            = 2,
    mutation_rate_variability = 0.25,
    deleterious_rate         = 0,
    n_deleterious_limit      = 1L,
    n_sequences_total        = 140L,
    ref_variability          = FALSE,
    n_ref_sequences          = 100L,
    prob_deletion_event      = 0,
    n_rows_to_delete         = 0L,
    seed                     = NULL
) {
  
  # ── 0. Input validation ─────────────────────────────────────────────────
  
  # ref_sequences: single character string OR data.frame with Date column
  if (!((is.character(ref_sequences) && length(ref_sequences) == 1L) ||
        (is.data.frame(ref_sequences) && "Date" %in% names(ref_sequences))))
    stop("`ref_sequences` must be either a single character string or a ",
         "data.frame with a `Date` column.", call. = FALSE)
  
  if (is.character(ref_sequences) && nchar(ref_sequences) < 3L)
    stop("`ref_sequences` must encode a sequence of length >= 3 ",
         "(sites 1 and 2 are reserved for reference variability).",
         call. = FALSE)
  if (is.data.frame(ref_sequences) &&
      sum(grepl("^[0-9]+$", names(ref_sequences))) < 3L)
    stop("`ref_sequences` data.frame must contain at least 3 site columns ",
         "named with integer position labels (\"1\", \"2\", ...).",
         call. = FALSE)
  
  # Date arguments
  start_date <- tryCatch(as.Date(start_date),
                         error = function(e)
                           stop("`start_date` is not a valid date.",
                                call. = FALSE))
  end_date   <- tryCatch(as.Date(end_date),
                         error = function(e)
                           stop("`end_date` is not a valid date.",
                                call. = FALSE))
  if (anyNA(c(start_date, end_date)))
    stop("`start_date` and `end_date` must both be valid dates.",
         call. = FALSE)
  if (start_date > end_date)
    stop("`start_date` must be on or before `end_date`.", call. = FALSE)
  
  # Integer-count arguments (non-negative scalars)
  is_count <- function(x) is.numeric(x) && length(x) == 1L &&
    !is.na(x) && x >= 0 && x == as.integer(x)
  if (!is_count(n_ref_months) || n_ref_months < 1L)
    stop("`n_ref_months` must be a positive integer.", call. = FALSE)
  for (nm in c("n_new_mutations", "n_deleterious_limit",
               "n_sequences_total", "n_ref_sequences", "n_rows_to_delete")) {
    if (!is_count(get(nm)))
      stop(sprintf("`%s` must be a non-negative integer.", nm),
           call. = FALSE)
  }
  
  # Timeline must extend beyond the reference phase
  n_months_total <- length(seq(start_date, end_date, by = "month"))
  if (n_months_total <= n_ref_months)
    stop("The simulation spans only ", n_months_total, " month(s) but ",
         "`n_ref_months` is ", n_ref_months, "; the timeline must include ",
         "at least one period beyond the reference phase.", call. = FALSE)
  
  # variants_config: positive integer vector, length >= 1
  if (!is.numeric(variants_config) || length(variants_config) < 1L ||
      anyNA(variants_config) || any(variants_config < 1) ||
      any(variants_config != as.integer(variants_config)))
    stop("`variants_config` must be a positive-integer vector of length >= 1.",
         call. = FALSE)
  
  # variant_intervals: positive integers when more than one variant requested
  nv_arg <- length(variants_config)
  if (nv_arg > 1L) {
    if (!is.numeric(variant_intervals) || length(variant_intervals) < 1L ||
        anyNA(variant_intervals) || any(variant_intervals < 1) ||
        any(variant_intervals != as.integer(variant_intervals)))
      stop("`variant_intervals` must be a positive-integer vector of length ",
           ">= 1 when `variants_config` has more than one variant.",
           call. = FALSE)
  }
  
  # mutation_rate: positive numeric, scalar or matched to variants_config
  if (!is.numeric(mutation_rate) || anyNA(mutation_rate) ||
      any(mutation_rate <= 0) ||
      !(length(mutation_rate) == 1L || length(mutation_rate) == nv_arg))
    stop("`mutation_rate` must be a positive numeric scalar or a vector of ",
         "length `length(variants_config)`.", call. = FALSE)
  
  # Probability-like arguments
  is_prob <- function(x, strict_upper = FALSE)
    is.numeric(x) && length(x) == 1L && !is.na(x) && x >= 0 &&
    (if (strict_upper) x < 1 else x <= 1)
  if (!is_prob(mutation_rate_variability, strict_upper = TRUE))
    stop("`mutation_rate_variability` must be a numeric scalar in [0, 1).",
         call. = FALSE)
  if (!is_prob(deleterious_rate))
    stop("`deleterious_rate` must be a numeric scalar in [0, 1].",
         call. = FALSE)
  if (!is_prob(prob_deletion_event))
    stop("`prob_deletion_event` must be a numeric scalar in [0, 1].",
         call. = FALSE)
  
  # ref_variability: single logical
  if (!is.logical(ref_variability) || length(ref_variability) != 1L ||
      is.na(ref_variability))
    stop("`ref_variability` must be a single logical value.", call. = FALSE)
  
  # Multi-variant fill-pool integrity: the hardcoded row indices c(1, 2, 5)
  # used in the fill-pool construction (Section 7c) assume each ref-phase
  # month carries at least 6 rows when `ref_variability = TRUE`, so that
  # idx 1 = wild-type, idx 2-4 = site-1 variability, idx 5-6 = site-2
  # variability. Below this density the variability rows can be pulled from
  # the wrong month chunk and the fill pool is mis-composed.
  if (ref_variability && n_ref_sequences < 6L * n_ref_months)
    warning("`ref_variability = TRUE` assumes >= 6 reference rows per month ",
            "(have ", n_ref_sequences, " across ", n_ref_months,
            " months). Site-1/site-2 variability rows may not be correctly ",
            "captured in the multi-variant fill pool.", call. = FALSE)
  
  # seed: NULL or single numeric value
  if (!is.null(seed) &&
      !(is.numeric(seed) && length(seed) == 1L && !is.na(seed)))
    stop("`seed` must be NULL or a single numeric value.", call. = FALSE)
  
  if (!is.null(seed)) set.seed(seed)
  
  aa <- c("A","R","N","D","C","E","Q","G","H","I",
          "L","K","M","F","P","S","T","W","Y","V")
  
  # ── 1. Internal helpers ──────────────────────────────────────────────────
  
  # del_cap: revert excess deleterious-mutation sequences back to reference
  del_cap <- function(mat, vidx, vdet, limit, ref_chars) {
    if (!is.matrix(mat)) {
      v <- as.vector(mat)
      mat <- matrix(v, nrow = 1L, ncol = length(v))
    }
    vd <- vdet[[vidx]]
    if (!any(vd$flags)) return(mat)
    for (p in vd$pos[vd$flags]) {
      allele <- vd$vseq[p]
      cnt    <- sum(mat[, p] == allele)
      if (cnt > limit) {
        cands <- which(mat[, p] == allele)
        mat[sample(cands, cnt - limit), p] <- ref_chars[p]
      }
    }
    mat
  }
  
  # ── 2. Build reference data frame ────────────────────────────────────────
  
  if (is.character(ref_sequences) && length(ref_sequences) == 1L) {
    base_seq <- ref_sequences
    L        <- nchar(base_seq)
    ref_df   <- as.data.frame(
      matrix(strsplit(base_seq, "")[[1]], nrow = n_ref_sequences, ncol = L,
             byrow = TRUE),
      stringsAsFactors = FALSE
    )
    colnames(ref_df) <- as.character(seq_len(L))
    ref_df$Date <- rep(
      seq(start_date, by = "month", length.out = n_ref_months),
      length.out = n_ref_sequences
    )
  } else {
    ref_df   <- ref_sequences
    seq_num  <- grep("^[0-9]+$", names(ref_df))
    base_seq <- paste(ref_df[1L, seq_num], collapse = "")
    L        <- length(seq_num)
  }
  
  seq_cols <- as.character(seq_len(L))
  
  # Add Month for reference-phase construction (kept until ref_phase is built)
  ref_df$Month <- format(as.Date(ref_df$Date), "%Y-%m")
  
  # ── 3. Reference variability (2 RNG draws if TRUE) ───────────────────────
  
  if (ref_variability) {
    sub1 <- sample(setdiff(aa, substr(base_seq, 1L, 1L)), 1L)
    sub2 <- sample(setdiff(aa, substr(base_seq, 2L, 2L)), 1L)
    for (m in unique(ref_df$Month)) {
      idx <- sort(which(ref_df$Month == m))
      if (length(idx) >= 4L) ref_df[idx[2L:4L], seq_cols[1L]] <- sub1
      if (length(idx) >= 6L) ref_df[idx[5L:6L], seq_cols[2L]] <- sub2
    }
  }
  
  # ── 4. Simulation timeline ───────────────────────────────────────────────
  
  # seq.Date inclusive of both endpoints gives the full set of monthly time
  # points. Note that lubridate::interval() %/% period() returns the
  # *distance* between two dates (e.g. 23 for Jan-2020 -> Dec-2021), which
  # is one short of the 24 monthly time points actually traversed.
  sim_dates <- seq(start_date, end_date, by = "month")
  n_periods <- length(sim_dates)
  
  template_cols <- c(seq_cols, "Variant", "Phase", "Date",
                     "Period", "Delet", "._weight")
  
  ensure_cols <- function(d) {
    miss <- setdiff(template_cols, names(d))
    d[miss] <- NA
    d[, template_cols, drop = FALSE]
  }
  
  # ── 5. Build reference phase via random month sampling (1 RNG draw) ──────
  
  choice    <- sample(unique(ref_df$Month), n_ref_months, replace = TRUE)
  ref_phase <- do.call(rbind, lapply(seq_len(n_ref_months), function(i) {
    chunk          <- ref_df[ref_df$Month == choice[i], ]
    chunk$Month    <- NULL
    chunk$Date     <- sim_dates[i]
    chunk$Period   <- i
    chunk$Variant  <- "Reference"
    chunk$Phase    <- ifelse(
      apply(chunk[, seq_cols, drop = FALSE], 1L,
            function(r) paste(r, collapse = "")) == base_seq,
      "R", "RV"
    )
    chunk$Delet    <- "No"
    ensure_cols(chunk)
  }))
  ref_phase[["._weight"]] <- 1
  
  output <- ref_phase
  
  # ── 6. Define variants (RNG: pos + AA per variant + rbinom flags) ─────────
  
  nv    <- length(variants_config)
  gaps  <- rep(variant_intervals, length.out = max(0L, nv - 1L))
  em    <- numeric(nv)
  em[1] <- n_ref_months + 1L
  if (nv > 1L) for (j in 2L:nv) em[j] <- em[j - 1L] + gaps[j - 1L]
  
  valid           <- which(em <= n_periods)
  em              <- em[valid]
  variants_config <- variants_config[valid]
  nv_valid        <- length(em)
  rates <- if (length(mutation_rate) == 1L) rep(mutation_rate, nv) else mutation_rate
  
  ref_seq_chars <- as.character(ref_phase[1L, seq_cols])
  
  variant_details <- vector("list", nv_valid)
  vrow_pool       <- vector("list", nv_valid)  # pre-built rows for fill pool
  
  for (i in seq_len(nv_valid)) {
    nm      <- variants_config[[i]]
    pos     <- sample(setdiff(seq_len(L), 1L:2L), nm)
    new_aas <- sapply(pos, function(p) sample(setdiff(aa, ref_seq_chars[p]), 1L))
    vseq    <- ref_seq_chars; vseq[pos] <- new_aas
    flags   <- stats::rbinom(nm, 1L, deleterious_rate) == 1L
    variant_details[[i]] <- list(
      em   = em[i], pos = pos, flags = flags, vseq = vseq,
      mult = rates[i], last = 0L, cum = 0L, repl = FALSE
    )
    # Pre-built pool row (used when this variant is neither cur nor cmp)
    vr                   <- as.data.frame(
      matrix(vseq, 1L, dimnames = list(NULL, seq_cols)),
      stringsAsFactors = FALSE
    )
    vr$Variant           <- paste0("Variant_", i)
    vr$Phase             <- paste0("M", i)
    vr$Date              <- as.Date(NA)
    vr$Period            <- NA_integer_
    vr$Delet             <- "No"
    vr[["._weight"]]     <- 2^(i + 1L)
    vrow_pool[[i]]       <- ensure_cols(vr)
  }
  
  n_fixed_per_month <- ceiling(n_sequences_total /
                                 (n_periods - n_ref_months))
  delet_records     <- list()
  pool              <- NULL   # updated each competition period; returned for inspection
  
  # ── 7. Mutation phase loop ───────────────────────────────────────────────
  
  for (t in (n_ref_months + 1L):n_periods) {
    
    today     <- sim_dates[t]
    row_start <- nrow(output) + 1L
    
    # Identify most-recently-emerged non-replaced variant
    active_idx <- 0L
    for (v in seq_len(nv_valid)) {
      if (t >= variant_details[[v]]$em && !variant_details[[v]]$repl)
        active_idx <- v
    }
    
    # ── 7a. Reference-only (pre-variant) ──────────────────────────────────
    if (active_idx == 0L) {
      # Dead code under standard parameters (loop starts at em[1] == n_ref_months+1)
      # but kept for correctness if parameters ever create a gap
      samp <- sample(nrow(ref_phase), n_fixed_per_month, TRUE,
                     ref_phase[["._weight"]])
      df              <- ref_phase[samp, template_cols]
      df$Date         <- today
      df$Period       <- t
      
      # ── 7b. Single variant ────────────────────────────────────────────────
    } else if (active_idx == 1L) {
      
      vd    <- variant_details[[1L]]
      n_cur <- if (t == vd$em) n_new_mutations else
        max(1L, ceiling(vd$last *
                          stats::runif(1L,
                                       vd$mult * (1 - mutation_rate_variability),
                                       vd$mult * (1 + mutation_rate_variability))))
      variant_details[[1L]]$last <- n_cur
      variant_details[[1L]]$cum  <- variant_details[[1L]]$cum + n_cur
      
      m_cur <- matrix(vd$vseq, nrow = n_cur, ncol = L, byrow = TRUE,
                      dimnames = list(NULL, seq_cols))
      left  <- n_fixed_per_month - n_cur
      
      if (left > 0L) {
        samp <- sample(nrow(ref_phase), left, TRUE, ref_phase[["._weight"]])
        mat  <- rbind(m_cur, as.matrix(ref_phase[samp, seq_cols]))
        lbl  <- c(rep("Variant_1", n_cur), ref_phase$Variant[samp])
      } else {
        mat <- m_cur
        lbl <- rep("Variant_1", n_cur)
      }
      
      idx_v      <- seq_len(min(n_cur, nrow(mat)))
      mat[idx_v, ] <- del_cap(mat[idx_v, , drop = FALSE], 1L,
                              variant_details, n_deleterious_limit,
                              ref_seq_chars)
      
      df         <- as.data.frame(mat, stringsAsFactors = FALSE)
      df$Variant <- lbl; df$Phase <- "M1"
      df$Date    <- today; df$Period <- t; df$Delet <- "No"
      
      # ── 7c. Multi-variant competition ─────────────────────────────────────
    } else {
      
      cur    <- active_idx
      cmp    <- cur - 1L
      vd_cur <- variant_details[[cur]]
      vd_cmp <- variant_details[[cmp]]
      
      # Stochastic growth for both competing variants
      n_cur <- if (t == vd_cur$em) n_new_mutations else
        max(1L, ceiling(vd_cur$last *
                          stats::runif(1L,
                                       vd_cur$mult * (1 - mutation_rate_variability),
                                       vd_cur$mult * (1 + mutation_rate_variability))))
      variant_details[[cur]]$last <- n_cur
      variant_details[[cur]]$cum  <- variant_details[[cur]]$cum + n_cur
      
      n_cmp <- max(1L, ceiling(vd_cmp$last *
                                 stats::runif(1L,
                                              vd_cmp$mult * (1 - mutation_rate_variability),
                                              vd_cmp$mult * (1 + mutation_rate_variability))))
      variant_details[[cmp]]$last <- n_cmp
      variant_details[[cmp]]$cum  <- variant_details[[cmp]]$cum + n_cmp
      
      m_cur <- matrix(vd_cur$vseq, nrow = n_cur, ncol = L, byrow = TRUE,
                      dimnames = list(NULL, seq_cols))
      m_cmp <- matrix(vd_cmp$vseq, nrow = n_cmp, ncol = L, byrow = TRUE,
                      dimnames = list(NULL, seq_cols))
      mat   <- rbind(m_cur, m_cmp)
      lbl   <- c(rep(paste0("Variant_", cur), n_cur),
                 rep(paste0("Variant_", cmp), n_cmp))
      
      # Fill remainder from reference + older (non-competing) variants
      left <- n_fixed_per_month - nrow(mat)
      if (left > 0L) {
        pool <- if (ref_variability) ref_phase[c(1L, 2L, 5L), ]
        else                 ref_phase[1L, , drop = FALSE]
        if (cur >= 3L) {
          for (j in seq_len(cur - 2L)) {
            vr          <- vrow_pool[[j]]
            vr$Date     <- today
            vr$Period   <- t
            vr$Phase    <- paste0("M", cur)
            pool        <- rbind(pool, vr)
          }
        }
        samp <- sample(nrow(pool), left, TRUE, pool[["._weight"]])
        mat  <- rbind(mat, as.matrix(pool[samp, seq_cols]))
        lbl  <- c(lbl, pool$Variant[samp])
      }
      
      # Trim to quota if cur+cmp overshoot
      if (nrow(mat) > n_fixed_per_month) {
        mat <- mat[seq_len(n_fixed_per_month), ]
        lbl <- lbl[seq_len(n_fixed_per_month)]
      }
      
      # Apply deleterious caps to cur and cmp rows
      idx_c <- seq_len(min(n_cur, nrow(mat)))
      mat[idx_c, ] <- del_cap(mat[idx_c, , drop = FALSE], cur,
                              variant_details, n_deleterious_limit,
                              ref_seq_chars)
      cs <- min(n_cur + 1L, nrow(mat))
      ce <- min(n_cur + n_cmp, nrow(mat))
      if (cs <= ce)
        mat[cs:ce, ] <- del_cap(mat[cs:ce, , drop = FALSE], cmp,
                                variant_details, n_deleterious_limit,
                                ref_seq_chars)
      
      df         <- as.data.frame(mat, stringsAsFactors = FALSE)
      df$Variant <- lbl; df$Phase <- paste0("M", cur)
      df$Date    <- today; df$Period <- t; df$Delet <- "No"
      
      # Mark predecessor as replaced once it drops out completely
      if (!any(lbl == paste0("Variant_", cmp)))
        variant_details[[cmp]]$repl <- TRUE
    }
    
    # ── 7d. Coin-flip substitution event ──────────────────────────────────
    if (stats::runif(1L) < prob_deletion_event) {
      site_idx          <- sample(setdiff(seq_len(L), 1L:2L), 1L)
      old_aa            <- df[1L, as.character(site_idx)]
      new_aa            <- sample(setdiff(aa, old_aa), 1L)
      n_rows            <- min(nrow(df), n_rows_to_delete)
      if (n_rows > 0L) {
        tr                              <- tail(seq_len(nrow(df)), n_rows)
        df[tr, as.character(site_idx)]  <- new_aa
        df$Delet[tr]                    <- "Yes"
        delet_records[[as.character(today)]] <- list(
          period = t, date = today, site = site_idx,
          old_aa = old_aa, new_aa = new_aa,
          rows   = row_start + tr - 1L
        )
      }
    }
    
    output <- rbind(output, ensure_cols(df))
  }
  
  # ── 8. Finalise output ───────────────────────────────────────────────────
  
  output[["._weight"]] <- NULL
  rownames(output)     <- NULL
  names(output)[seq_len(L)] <- as.character(seq_len(L))
  
  structure(
    list(
      Simulation_Output     = output,
      Variant_Details       = variant_details,
      Simulation_Dates      = sim_dates,
      Baseline_Ref_Sequence = base_seq,
      Delet_Records         = delet_records,
      Pool                  = pool
    ),
    class = "viralSim"
  )
}

#' @title Build Distance/Entropy Matrix
#' @description Converts the output of \code{\link{calculate_hellinger_matrix}}
#'   or the legacy \code{get_hellinger_dist} / entropy list into a numeric
#'   \code{sites × time_steps} matrix suitable for change point detection.
#'
#' @details
#' In the current pipeline \code{\link{calculate_hellinger_matrix}} already
#' returns a \code{sites × time_steps} matrix, so calling this function is
#' usually unnecessary — transposing that matrix directly gives \code{dat_t}
#' for \code{\link[ecp]{e.agglo}} or \code{\link[ecp]{ks.cp3o}}:
#'
#' \preformatted{
#'   hell_mat = calculate_hellinger_matrix(partitions, sites = seq_len(n_sites))
#'   dat_t    = t(hell_mat)   # time_steps × sites — ready for ECP
#' }
#'
#' This function is retained for \strong{backward compatibility} with code that
#' was written against the old \code{get_hellinger_dist} list format, and as a
#' convenience wrapper that also accepts entropy lists.
#'
#' \strong{Input formats accepted:}
#' \enumerate{
#'   \item A numeric matrix (e.g. from \code{calculate_hellinger_matrix}) —
#'     returned as-is with site rownames normalised.
#'   \item The legacy named list from \code{get_hellinger_dist} containing
#'     \code{$Sites} and \code{$Hellinger_Distances} (list of vectors).
#'   \item An entropy list containing \code{$Sites} and \code{$Entropies}
#'     (list of numeric vectors), used when no Hellinger distances are present.
#' }
#'
#' @param data_input Either:
#'   \itemize{
#'     \item A numeric matrix (\code{sites × time_steps}) as returned by
#'       \code{\link{calculate_hellinger_matrix}}.
#'     \item A named list with elements \code{Sites} and one of
#'       \code{Hellinger_Distances} or \code{Entropies}, as returned by the
#'       legacy \code{get_hellinger_dist} function.
#'   }
#'
#' @return A numeric matrix with:
#'   \item{Rows}{Sites (1 to \code{max(sites)}, sparse if sites are
#'     non-contiguous).}
#'   \item{Columns}{Time steps (partitions T2, T3, … or entropy periods).}
#'   Row names are character site indices. Column names are preserved from
#'   the input where available.
#'
#' @seealso \code{\link{calculate_hellinger_matrix}},
#'   \code{\link{detect_changepoints_ecp}}
#'
#' @keywords internal
build_distance_matrix = function(data_input) {
  
  # --- 1. Matrix input: already in the right format -------------------------
  # Returned by calculate_hellinger_matrix — just normalise rownames and return
  if (is.matrix(data_input)) {
    if (is.null(rownames(data_input))) {
      rownames(data_input) = as.character(seq_len(nrow(data_input)))
    }
    return(data_input)
  }
  
  # --- 2. Legacy list input -------------------------------------------------
  if (!is.list(data_input)) {
    stop("`data_input` must be a matrix or a named list (see ?build_distance_matrix).")
  }
  
  # Determine data source: Hellinger distances take priority over Entropies
  has_hellinger = length(data_input$Hellinger_Distances) > 0
  dat_temp      = if (has_hellinger) data_input$Hellinger_Distances
  else               data_input$Entropies
  
  if (is.null(dat_temp) || length(dat_temp) == 0) {
    stop("Input list must contain non-empty `Hellinger_Distances` or `Entropies`.")
  }
  
  sites    = sort(unique(as.numeric(as.character(data_input$Sites))))
  n_cols   = length(dat_temp[[1L]])            # time steps per site
  data_mat = matrix(0, nrow = max(sites), ncol = n_cols)
  rownames(data_mat) = as.character(seq_len(max(sites)))
  
  # Preserve column names from the first element if available
  if (!is.null(names(dat_temp[[1L]]))) {
    colnames(data_mat) = names(dat_temp[[1L]])
  }
  
  # Fill rows by site index (sparse-safe: zeros remain for absent sites)
  for (i in seq_along(sites)) {
    loc              = sites[i]
    match_idx        = which(data_input$Sites == loc)
    data_mat[loc, ]  = dat_temp[[match_idx]]
  }
  
  data_mat
}

#' @title ViralEntropR: A Computational Pipeline for Entropy-Informed Detection of Emerging Viral Variants
#'
#' @description
#' A computational pipeline for detecting emerging variants in viral amino
#' acid sequence data, combining per-site Shannon entropy, Gaussian mixture
#' model site selection, Gower-distance Partitioning Around Medoids
#' clustering, Hellinger-distance quantification of distributional shifts,
#' and multivariate non-parametric change-point detection.
#'
#' @section Pipeline overview:
#' The package supports a four-stage workflow:
#' \itemize{
#'   \item \strong{Preprocessing.} Parse FASTA headers, filter ambiguous
#'     residues, and convert between integer and character representations
#'     of amino acid sequences under a 25-symbol alphabet. See
#'     \code{\link{extract_fasta_dates}},
#'     \code{\link{extract_fasta_countries}},
#'     \code{\link{fasta_to_char_matrix}},
#'     \code{\link{filter_ambiguous_sequences}},
#'     \code{\link{encode_aa_sequence}}, and
#'     \code{\link{decode_aa_sequence}}.
#'   \item \strong{Site selection.} Compute per-site Shannon entropy across
#'     temporal partitions and cluster sites by entropy via Gaussian
#'     mixture models. See \code{\link{calculate_entropy}},
#'     \code{\link{partition_time_windows}},
#'     \code{\link{cluster_sites_by_entropy}}, and
#'     \code{\link{relabel_entropy_classes}}.
#'   \item \strong{Distributional analysis.} Quantify residue-composition
#'     shifts between time windows using the Hellinger distance. See
#'     \code{\link{calculate_hellinger_matrix}}.
#'   \item \strong{Change-point detection.} Identify temporal change points
#'     non-parametrically using energy statistics or wild binary
#'     segmentation. See \code{\link{detect_changepoints_ecp}} and
#'     \code{\link{detect_changepoints_hdcp}}.
#' }
#'
#' @section Visualisation and tabulation:
#' \itemize{
#'   \item \code{\link{plot_entropy_trajectories}} — customizable multi-site 
#'     Shannon entropy trajectories, summarizing evolutionary dynamics 
#'     across time.
#'   \item \code{\link{plot_site_class_trajectory}} — single-site entropy
#'     trajectory with class-change markers for inspecting individual
#'     residues of interest.
#'   \item \code{\link{tabulate_site_evolution}} — per-site amino acid
#'     count and proportion tables across partitions, optionally
#'     rendered as styled HTML.
#' }
#'
#' @section Simulation:
#' \code{\link{simulate_variant_evolution}} provides a configurable
#' multi-variant simulation engine with user-specified emergence
#' schedules, growth rates, mutation-rate variability, and
#' deleterious-mutation injection. Generates synthetic time-series data
#' with known ground truth for benchmarking detection pipelines.
#'
#' @section Bundled data:
#' \itemize{
#'   \item \code{\link{sarscov2_variants}} — curated metadata for twelve
#'     SARS-CoV-2 Variants of Concern and Variants of Interest, including
#'     WHO labels, Pango lineages, GISAID and Nextstrain clades, dates
#'     and countries of first detection, defining Spike-protein
#'     mutations and SNP sites, and 21 peer-reviewed references with DOIs.
#'   \item \code{\link{sarscov2_sample}} — a random sample of 100 NCBI Spike 
#'     protein sequences for end-to-end testing without external downloads.
#' }
#' 
#' @section External data:
#' The complete preprocessed NCBI Spike protein dataset (137,132
#' sequences, ~181.5 MB uncompressed FASTA) underlying the package's
#' real-data pre-processing vignette is archived on Zenodo:
#' \doi{10.5281/zenodo.19040165}. The dataset can be read directly with
#' \code{\link[Biostrings]{readAAStringSet}} and processed end-to-end
#' using the preprocessing toolkit; see
#' \code{vignette("preprocessing_pipeline", "ViralEntropR")} for the
#' full workflow.
#'
#' @section Vignettes:
#' Three pre-rendered vignettes walk through the full workflow on real
#' and simulated data: \code{vignette("preprocessing_pipeline",
#' "ViralEntropR")}, \code{vignette("detecting_variants_simulation",
#' "ViralEntropR")}, and \code{vignette("clustering_accuracy",
#' "ViralEntropR")}.
#'
#' @keywords internal
"_PACKAGE"

#' @title Encode Amino Acid Sequences
#' @description Converts a character matrix of amino acid codes to its
#'   integer-encoded representation under the package's 25-symbol alphabet.
#'   Inverse of \code{\link{decode_aa_sequence}}.
#'
#' @details
#' Encoding is a single vectorised lookup against a fixed named map. The
#' input matrix is flattened, normalised to uppercase, indexed against the
#' alphabet's name vector in one operation, and reshaped — there is no
#' per-row or per-element loop.
#'
#' The 25-symbol alphabet covers the twenty standard residues
#' (\code{A=1, R=2, N=3, D=4, C=5, Q=6, E=7, G=8, H=9, I=10, L=11, K=12,
#' M=13, F=14, P=15, S=16, T=17, W=18, Y=19, V=20}), the three IUPAC
#' ambiguous codes \code{B=21, Z=22, X=23}, the stop codon \code{*=24},
#' and the alignment gap \code{-=25}.
#'
#' Any character outside this alphabet — including \code{NA}, the empty
#' string, lowercase letters not in the standard set after uppercasing
#' (e.g. \code{J} for leucine/isoleucine), and non-letter symbols — maps
#' to the sentinel \code{0}. This sentinel is recognised by downstream
#' functions: \code{\link{filter_ambiguous_sequences}} treats \code{0}
#' as ambiguous (alongside \code{B}, \code{X}, \code{Z}) and removes
#' affected sequences, and \code{\link{decode_aa_sequence}} round-trips
#' it back to the string \code{"0"}.
#'
#' Row and column names of the input matrix are preserved on the output.
#'
#' @param matrix_input A character matrix of amino acid codes, typically
#'   produced by \code{\link{fasta_to_char_matrix}}. A non-matrix input is
#'   coerced via \code{\link[base]{as.matrix}}. Lowercase input is
#'   accepted; characters are uppercased before lookup.
#'
#' @return A numeric matrix of the same dimensions as \code{matrix_input},
#'   with the same \code{dimnames}. Each cell is either an integer in
#'   \code{1:25} from the alphabet table above, or \code{0} for any input
#'   not in the alphabet (including \code{NA} and the empty string).
#'
#' @seealso \code{\link{decode_aa_sequence}} for the inverse operation;
#'   \code{\link{fasta_to_char_matrix}} for the FASTA-to-character-matrix
#'   step that typically precedes encoding;
#'   \code{\link{filter_ambiguous_sequences}} for downstream removal of
#'   sequences containing ambiguous or unrecognised residues.
#'
#' @export
#'
#' @examples
#' # 1. Encode a simple matrix of sequences.
#' seq_mat = matrix(c("A", "R", "N", "D"), nrow = 2, byrow = TRUE)
#' encoded = encode_aa_sequence(seq_mat)
#' print(encoded)
#'
#' # 2. Gaps and unknown characters.
#' # '-' maps to 25; '?' is not in the alphabet and maps to the sentinel 0.
#' gapped_mat = matrix(c("A", "-", "G", "?"), nrow = 1)
#' encode_aa_sequence(gapped_mat)
#'
#' # 3. Lowercase input is accepted (uppercased before lookup).
#' encode_aa_sequence(matrix(c("a", "r", "n", "d"), nrow = 2))
#'
#' # 4. NA and empty string both map to 0.
#' encode_aa_sequence(matrix(c("A", NA, "G", ""), nrow = 2))
encode_aa_sequence = function(matrix_input) {
  # Internal mapping - 25 standard/ambiguous codes
  aa_map = c(
    "A" = 1,   "R" = 2,   "N" = 3,   "D" = 4,   "C" = 5,
    "Q" = 6,   "E" = 7,   "G" = 8,   "H" = 9,   "I" = 10,
    "L" = 11,  "K" = 12,  "M" = 13,  "F" = 14,  "P" = 15,
    "S" = 16,  "T" = 17,  "W" = 18,  "Y" = 19,  "V" = 20,
    "B" = 21,  "Z" = 22,  "X" = 23,  "*" = 24,  "-" = 25
  )
  
  if (!is.matrix(matrix_input)) matrix_input = as.matrix(matrix_input)
  
  # Normalize input to uppercase
  input_vec = toupper(as.vector(matrix_input))
  
  # Map characters to numbers
  # Values not in the map (including your specified unknown_char) become NA, then 0
  encoded_vec = aa_map[input_vec]
  encoded_vec[is.na(encoded_vec)] = 0
  
  matrix(encoded_vec, 
         nrow = nrow(matrix_input), 
         ncol = ncol(matrix_input),
         dimnames = dimnames(matrix_input))
}

#' @title Remove Sequences Containing Ambiguous Residues
#' @description Removes rows (sequences) that contain at least one ambiguous
#'   amino acid residue (B, J, X, or Z) — and, under integer-encoded input,
#'   any unrecognised character — from a sequence matrix. Accepts both
#'   integer-encoded matrices and character matrices.
#'
#' @details
#' \strong{What is removed.} Sequences are flagged for removal if any of
#' their residue positions contain one of the four IUPAC ambiguous codes:
#' \itemize{
#'   \item \code{B} — Aspartate / Asparagine.
#'   \item \code{J} — Leucine / Isoleucine.
#'   \item \code{X} — any residue.
#'   \item \code{Z} — Glutamate / Glutamine.
#' }
#'
#' \strong{What is NOT removed.} Standard alignment gaps (\code{-},
#' integer code \code{25}) are retained — gaps represent known absences
#' rather than uncertain identities and are typically positionally
#' meaningful in aligned data. Sequences containing only canonical 20
#' amino acids and gaps are kept.
#'
#' \strong{How input mode is handled.}
#' \itemize{
#'   \item \code{option = 1} (integer-encoded input from
#'     \code{\link{encode_aa_sequence}}): rows are removed if any cell
#'     equals \code{0} (unrecognised — including J, NA, empty, lowercase
#'     mismatches, byte-order marks, and other characters that fell
#'     outside the encoding alphabet), \code{21} (B), \code{22} (Z), or
#'     \code{23} (X). The \code{0} sentinel acts as a catch-all for
#'     anything not in the 25-symbol alphabet.
#'   \item \code{option = 2} (character input from
#'     \code{\link{fasta_to_char_matrix}}): rows are removed if any cell
#'     is exactly \code{"B"}, \code{"J"}, \code{"X"}, or \code{"Z"}.
#'     Unrecognised characters in character input (e.g. lowercase letters,
#'     \code{NA}, empty strings) are NOT caught at this stage; encode
#'     first if you need that catch-all behaviour.
#' }
#'
#' \strong{Performance.} Detection is fully vectorised: a single logical
#' matrix comparison followed by \code{\link[base]{rowSums}} counts
#' ambiguous residues per sequence in one C-level call, replacing the
#' original row-by-row loop for a substantial speed improvement on
#' large matrices (100k+ rows).
#'
#' @param NumMatrix A matrix. Rows are sequences, columns are sites. Either
#'   integer-encoded (\code{option = 1}) or character (\code{option = 2}).
#'   Despite the name, character matrices are also accepted under
#'   \code{option = 2}.
#' @param option Integer. \code{1} (default) for integer-encoded matrices
#'   produced by \code{\link{encode_aa_sequence}}; \code{2} for character
#'   matrices produced by \code{\link{fasta_to_char_matrix}}.
#'
#' @return A named list:
#' \item{OriginalDim}{Character string reporting the number of input
#'   sequences.}
#' \item{NewDim}{Character string reporting the number of sequences
#'   remaining after filtering.}
#' \item{NumberAmbiguous}{Character string reporting the number of
#'   sequences that contained at least one ambiguous residue.}
#' \item{RangeAmbiguous}{Character string reporting the min and max count
#'   of ambiguous residues per removed sequence, or
#'   \code{"No ambiguous sequences found"} when none were removed.}
#' \item{DeletedSeqId}{Integer vector of row indices that were removed.
#'   Empty integer vector if nothing was removed.}
#' \item{FilteredMatrix}{The filtered matrix with ambiguous rows removed,
#'   preserving the original column structure and storage mode.}
#'
#' @seealso \code{\link{encode_aa_sequence}} and
#'   \code{\link{fasta_to_char_matrix}} for producing the typical input;
#'   \code{\link{decode_aa_sequence}} for inspecting the surviving
#'   sequences in character form.
#'
#' @export
#'
#' @examples
#' # Synthetic example: 50 sequences, 10 sites, drawn from canonical residues.
#' set.seed(1)
#' m <- matrix(sample(1:20, 500, replace = TRUE), nrow = 50, ncol = 10)
#' # Inject ambiguous codes into 3 specific rows: 21 (B), 23 (X), 0 (unrecognised).
#' m[c(3, 17, 42), sample(1:10, 3)] <- c(21, 23, 0)
#' result <- filter_ambiguous_sequences(m, option = 1)
#' cat(result$NumberAmbiguous, "\n")
#' cat(result$RangeAmbiguous, "\n")
#' dim(result$FilteredMatrix)
#'
#' # Character-mode example.
#' chr <- matrix(c("M", "K", "T", "I", "I", "X", "K", "T", "I", "I"),
#'                nrow = 2, byrow = TRUE)
#' filter_ambiguous_sequences(chr, option = 2)$DeletedSeqId
filter_ambiguous_sequences <- function(NumMatrix, option = 1) {

  if (!option %in% 1:2) stop("`option` must be 1 (numeric) or 2 (character).", 
                             call. = FALSE)
  
  if (!is.matrix(NumMatrix)) stop("`NumMatrix` must be a matrix.", 
                                  call. = FALSE)

  # Vectorized detection: build a logical matrix of ambiguous positions,
  # then rowSums counts ambiguous residues per sequence in one C-level call.
  # This replaces the original row-by-row for loop — O(n*m) but fully in C.
  if (option == 1) {
    ambig_mat <- NumMatrix == 21L | NumMatrix == 22L |
                 NumMatrix == 23L | NumMatrix == 0L
  } else {
    ambig_mat <- NumMatrix == "B" | NumMatrix == "J" |
                 NumMatrix == "X" | NumMatrix == "Z"
  }

  ambig_per_row <- rowSums(ambig_mat)   # integer vector, length = nrow
  id_delete     <- which(ambig_per_row > 0L)
  l_w           <- ambig_per_row[id_delete]

  filtered <- if (length(id_delete) > 0L) NumMatrix[-id_delete, , drop = FALSE]
              else NumMatrix

  range_str <- if (length(l_w) > 0L)
    sprintf(
      "Number of ambiguous protein characters per sequence varies between %d and %d",
      min(l_w), max(l_w)
    )
  else
    "No ambiguous sequences found"

  list(
    OriginalDim     = sprintf("Number of original sequences is %d",  nrow(NumMatrix)),
    NewDim          = sprintf("Number of filtered sequences is %d",  nrow(filtered)),
    NumberAmbiguous = sprintf(
      "Number of sequences containing at least one of B, X, Z or J characters is %d",
      length(id_delete)
    ),
    RangeAmbiguous  = range_str,
    DeletedSeqId    = id_delete,
    FilteredMatrix  = filtered
  )
}

#' @title Detect Temporal Change Points (ECP)
#' @description Runs Energy Change Point detection on time-series matrices
#'   (typically Hellinger distance matrices) over one or more time windows.
#' @export
#'
#' @details
#' The function applies \code{\link[ecp]{ks.cp3o}} repeatedly to slices of
#' \code{data_matrix}, advancing the slice forward by one row at each step.
#' A total of \code{n_timesteps + 1} detections are performed: one on the
#' initial window \code{[min_window, max_window]}, then \code{n_timesteps}
#' additional detections on subsequent windows.
#'
#' Two window-advancement modes are supported:
#' \itemize{
#'   \item \strong{Expanding} (\code{rolling_window = FALSE}, default): the
#'     start index is fixed at \code{min_window}; the end index grows by 1
#'     each step. Each iteration uses one more row than the previous.
#'     Natural for online surveillance, where change-point detection is
#'     re-run as new data accumulates.
#'   \item \strong{Rolling} (\code{rolling_window = TRUE}): both start and
#'     end indices advance by 1 each step, keeping the window length
#'     constant. Natural for retrospective windowed analysis, where local
#'     change-point structure is examined in fixed-size segments.
#' }
#'
#' @param data_matrix A numeric matrix (Time x Features). Rows are time
#'   points, columns are features (e.g. Hellinger distances per site).
#' @param min_window Integer. Starting row index of the initial window.
#' @param max_window Integer. Ending row index of the initial window. Must
#'   not exceed \code{nrow(data_matrix)}.
#' @param n_timesteps Integer >= 0. Number of \emph{additional} detections
#'   to run after the initial window. The function performs
#'   \code{n_timesteps + 1} detections in total. Set to \code{0} for a
#'   single detection on the initial window.
#' @param rolling_window Logical. If \code{TRUE}, both window endpoints
#'   advance per step (sliding window of constant length). If \code{FALSE}
#'   (default), only the end advances (expanding window).
#' @param dynamic_k Logical. If \code{TRUE}, sets \code{K} to
#'   \code{nrow(subset) - 2L} at each step, so the algorithm considers
#'   the maximum number of change points permitted by the window length
#'   (subject to \code{minsize}). Useful for exploratory analysis where
#'   the true number of change points is unknown a priori. Most idiomatic
#'   when combined with \code{n_timesteps > 0}, where the maximum
#'   admissible K legitimately scales with window size, but also valid
#'   for single-window exploratory use. If \code{FALSE} (default), the
#'   value of \code{K} supplied via \code{...} is honoured at every step;
#'   if no \code{K} is supplied, the \code{ks.cp3o} default of 1 is used.
#' @param ... Additional arguments passed to \code{\link[ecp]{ks.cp3o}}.
#'   Common choices include \code{K} (maximum number of change points,
#'   default 1) and \code{minsize} (minimum segment size, default 30).
#'   Note that \code{K} is overridden when \code{dynamic_k = TRUE}; a
#'   warning is emitted in that case.
#'
#' @return A named list:
#' \item{Points_List}{List of integer vectors \code{c(start, end)} giving
#'   the row index window used at each step.}
#' \item{ECP_list}{List of full \code{ks.cp3o} result objects, one per
#'   step. Inspect \code{$cpLoc} for the optimal change-point locations
#'   at each candidate count, and \code{$gofM} for the goodness-of-fit
#'   curve, to apply post-hoc filtering.}
#' \item{ECP_est_list}{List of change-point estimate vectors (the
#'   algorithm's own selection, equivalent to \code{$estimates}).}
#'
#' @section Warnings and errors:
#' \describe{
#'   \item{\code{stop()}}{Triggered if \code{max_window} exceeds
#'     \code{nrow(data_matrix)}: the initial window cannot be
#'     constructed and no detection can run.}
#'   \item{\code{warning()}}{Triggered (and execution continues) if
#'     (a) \code{max_window + n_timesteps} exceeds
#'     \code{nrow(data_matrix)}, in which case the offending later
#'     iterations are skipped and \code{NULL} entries appear in the
#'     result lists; (b) \code{dynamic_k = TRUE} and \code{K} is also
#'     supplied via \code{...}, in which case the user-supplied
#'     \code{K} is silently overridden by \code{nrow(subset) - 2L}.}
#' }
#'
#' @importFrom ecp ks.cp3o
#'
#' @examples
#' set.seed(123)
#' baseline = matrix(rnorm(50, mean = 0, sd = 0.1), nrow = 10, ncol = 5)
#' variant  = matrix(rnorm(50, mean = 3, sd = 0.1), nrow = 10, ncol = 5)
#' data_mat = rbind(baseline, variant)
#'
#' # Single-window detection: one true change point at row 11.
#' res = detect_changepoints_ecp(
#'   data_matrix = data_mat,
#'   min_window  = 1,
#'   max_window  = 20,
#'   n_timesteps = 0,
#'   minsize     = 5
#' )
#' print(res$ECP_est_list[[1]])
detect_changepoints_ecp = function(data_matrix,
                                   min_window,
                                   max_window,
                                   n_timesteps,
                                   rolling_window = FALSE,
                                   dynamic_k      = FALSE,
                                   ...) {
  
  # --- 1. Validation ---------------------------------------------------------
  if (!is.matrix(data_matrix))
    stop("`data_matrix` must be a matrix.", call. = FALSE)
  
  if (!is.numeric(n_timesteps) || length(n_timesteps) != 1L || n_timesteps < 0)
    stop("`n_timesteps` must be a single non-negative integer.",
         call. = FALSE)
  
  n_rows = nrow(data_matrix)
  
  # Hard guard: if the initial window itself is impossible, no detection
  # can run. Returning a list of NULLs would be silently degraded output;
  # an explicit error forces the caller to fix the call.
  if (max_window > n_rows)
    stop("`max_window` (", max_window, ") exceeds nrow(data_matrix) (",
         n_rows, "). The initial window [", min_window, ", ",
         max_window, "] cannot be constructed.",
         call. = FALSE)
  
  # Soft guard: if later iterations would go past the matrix, warn with
  # concrete numbers so the caller can decide whether to fix or accept
  # the degraded output. Skipped iterations return NULL entries.
  needed = max_window + n_timesteps
  if (n_rows < needed) {
    n_skipped = needed - n_rows
    warning(sprintf(
      "`data_matrix` has %d rows but max_window (%d) + n_timesteps (%d) requires %d. %d step(s) will be skipped and return NULL.",
      n_rows, max_window, n_timesteps, needed, n_skipped),
      call. = FALSE)
  }
  
  # Capture ... once so it does not need to be re-evaluated on every iteration.
  extra_args = list(...)
  
  # Guard against silent K override: if the caller supplied K via ... and
  # also set dynamic_k = TRUE, the dynamic-K branch overrides the
  # user-supplied K. Warn once at function entry rather than once per window.
  if (isTRUE(dynamic_k) && "K" %in% names(extra_args)) {
    warning(
      "Argument `K` supplied via `...` is ignored because ",
      "`dynamic_k = TRUE`; K is set to nrow(subset) - 2L for each window. ",
      "Set `dynamic_k = FALSE` to honour the user-supplied K.",
      call. = FALSE
    )
  }
  
  # --- 2. Per-step worker ----------------------------------------------------
  # dynamic_k mirrors the original adj_K formula: nrow(subset) - 2.
  
  run_step = function(i) {
    start_idx = if (rolling_window) (min_window + i) else min_window
    end_idx   = max_window + i
    
    if (end_idx > n_rows) {
      warning("Step ", i, ": end_idx (", end_idx,
              ") exceeds matrix rows (", n_rows, "); skipping.",
              call. = FALSE)
      return(list(window   = c(start_idx, end_idx),
                  estimate = NULL,
                  full     = NULL))
    }
    
    subset_mat = data_matrix[start_idx:end_idx, , drop = FALSE]
    
    args = c(list(Z = subset_mat), extra_args)
    if (dynamic_k) args$K = nrow(subset_mat) - 2L
    
    ecp_res = do.call(ecp::ks.cp3o, args)
    
    list(window   = c(start_idx, end_idx),
         estimate = ecp_res$estimates,
         full     = ecp_res)
  }
  
  # --- 3. Execute ------------------------------------------------------------
  results_list = lapply(0:n_timesteps, run_step)
  
  # --- 4. Return — names match original generate_ECP_list_modified -----------
  list(
    Points_List  = lapply(results_list, `[[`, "window"),
    ECP_list     = lapply(results_list, `[[`, "full"),
    ECP_est_list = lapply(results_list, `[[`, "estimate")
  )
}

#' @title Plot Shannon Entropy Trajectories
#'
#' @description Plots per-site Shannon entropy as continuous trajectories across
#'   time partitions for a selected set of sequence sites, using the output of
#'   \code{\link{partition_time_windows}}.
#'
#' @details
#' For each partition the function extracts the GMM clustering result from
#' \code{part_data$Clusters} and assembles a long-format data frame spanning
#' all selected sites across all partitions.  Sites absent from a given
#' partition (removed by zero-entropy or singleton filtering, or because the
#' partition window was empty) are silently omitted from that partition's
#' trajectory and do not interrupt adjacent observations.
#'
#' \strong{Class relabeling.}  This function does not perform any relabeling of
#' GMM class labels.  If class 1 must denote the highest-entropy group
#' throughout the returned \code{$Data_Frame} (e.g. before passing it to
#' \code{\link{plot_site_class_trajectory}}), the user should call
#' \code{\link{relabel_entropy_classes}} on each partition's
#' \code{Clusters[[i]]$DataFrame} and update \code{Max_Entropy[i]} to
#' \code{1L} prior to calling this function.
#'
#' \strong{Colour scheme.}  Site colours are specified through
#' \code{site_colors}, a named character vector whose names are site indices
#' (as character strings) and whose values are valid R colour strings.  Any
#' site not listed in \code{site_colors} receives an automatically assigned
#' colour from the HCL \code{"Dark 2"} qualitative palette.  The final colour
#' mapping is returned as \code{$Colors} so the same scheme can be passed to
#' subsequent calls for cross-plot consistency.
#'
#' \strong{Group-stratified trajectories (\code{by_group = TRUE}).}  When
#' biological groupings must be distinguished visually (e.g. defining SNP
#' sites vs. other mutation sites), \code{groups_list} partitions
#' \code{sites} into explicitly named groups.  Any site not assigned to an
#' explicit group is automatically collected into a remainder group appended
#' as the final element of \code{groups_list}.  Line type and line width are
#' mapped to group membership via \code{line_type_groups} and
#' \code{line_size_groups}, both of which must have length equal to the
#' total number of groups (explicit plus the automatic remainder).  At most
#' six groups are supported.
#'
#' \strong{\code{max_class} column.}  The returned \code{$Data_Frame} carries
#' a \code{max_class} column recording the label of the highest-entropy GMM
#' component for each partition, taken directly from
#' \code{part_data$Max_Entropy[i]}.  This column is consumed by
#' \code{\link{plot_site_class_trajectory}} (red labels) and by downstream
#' class-assignment tables.
#'
#' @param part_data Named list.  Output of \code{\link{partition_time_windows}},
#'   optionally with per-partition \code{Clusters[[i]]$DataFrame} already
#'   relabeled by the user via \code{\link{relabel_entropy_classes}}.  Must
#'   contain elements \code{Clusters}, \code{Max_Entropy}, \code{Dates_Labels},
#'   and \code{N_partitions}.
#' @param sites Integer vector.  Site indices to include.  Defaults to the
#'   union of all sites observed across all partitions (i.e. every site that
#'   has non-zero, non-singleton entropy in at least one partition window).
#' @param labels Character vector of length \code{N_partitions}.  Partition
#'   labels used on the x-axis.  Defaults to \code{part_data$Dates_Labels}.
#' @param site_colors Named character vector.  Names are site indices as
#'   character strings (e.g. \code{"681"}); values are colour strings
#'   (e.g. \code{"#FB8072"}).  Sites absent from \code{site_colors} receive
#'   automatically assigned colours.  Default is \code{NULL} (all colours
#'   auto-assigned).
#' @param by_group Logical.  If \code{TRUE}, maps line type and line width to
#'   site groups defined by \code{groups_list}.  Default is \code{FALSE}.
#' @param groups_list List of integer vectors.  Each element specifies the
#'   site indices belonging to one explicit group.  Sites in \code{sites} not
#'   covered by any explicit group are automatically assigned to a remainder
#'   group appended as the final element.  Total group count (explicit plus
#'   remainder) must not exceed 6.  Required when \code{by_group = TRUE}.
#' @param line_type_groups Character vector.  One line-type string per group
#'   (in order, including the automatic remainder group).  Must have length
#'   equal to the total number of groups.  Defaults to \code{"solid"} for the
#'   first group and \code{"dashed"} for all remaining groups.
#' @param line_size_groups Numeric vector.  One line-width value per group
#'   (in order, including the automatic remainder group).  Must have length
#'   equal to the total number of groups.  Defaults to \code{2} for the first
#'   group and \code{1} for all remaining groups.
#' @param transformation Object of class \code{"transform"} or \code{"trans"}
#'   as returned by \code{\link[scales]{trans_new}}, or \code{NULL} (identity,
#'   no transformation).  Applied to the y-axis via
#'   \code{\link[ggplot2]{scale_y_continuous}}.  Default is \code{NULL}.
#' @param line_size Numeric.  Line width used when \code{by_group = FALSE}.
#'   Default is \code{1.5}.
#' @param legend Logical.  If \code{TRUE} (default), the site colour legend is
#'   displayed.
#' @param legend_text_size Numeric.  Font size of legend text in points.
#'   Default is \code{12}.
#' @param x_angle Numeric.  Rotation angle of x-axis tick labels in degrees.
#'   Default is \code{45}.
#' @param grayscale Logical.  If \code{TRUE}, overrides \code{site_colors} and
#'   renders all trajectories in greyscale.  Default is \code{FALSE}.
#' @param plot_title Character.  Plot title string.  Default is
#'   \code{"Shannon Entropy Trajectories"}.
#'
#' @return A named list with five elements:
#' \item{Data_Frame}{Long-format data frame with columns \code{sites}
#'   (factor), \code{entropies} (numeric), \code{class} (factor),
#'   \code{max_class} (integer), \code{period} (integer), and
#'   \code{coverage} (character, partition label).  Suitable for direct
#'   input to \code{\link{plot_site_class_trajectory}}.}
#' \item{Plot}{A \code{ggplot} object.  Augment with additional layers
#'   (e.g. \code{\link[ggplot2]{geom_vline}} for VOC emergence events) before
#'   printing or saving with \code{\link[ggplot2]{ggsave}}.}
#' \item{Colors}{Named character vector mapping each plotted site index
#'   (character) to its assigned colour string.  Pass as \code{site_colors}
#'   to subsequent calls to \code{plot_entropy_trajectories} for a consistent
#'   colour scheme across figures.}
#' \item{XBreaks}{Integer vector of partition period indices.  Pass as
#'   \code{xbreaks} to \code{\link{plot_site_class_trajectory}}.}
#' \item{XLabels}{Character vector of partition labels aligned with
#'   \code{XBreaks}.  Pass as \code{xlabels} to
#'   \code{\link{plot_site_class_trajectory}}.}
#'
#' @seealso
#' \code{\link{partition_time_windows}},
#' \code{\link{relabel_entropy_classes}},
#' \code{\link{plot_site_class_trajectory}}
#'
#' @importFrom ggplot2 ggplot aes geom_line scale_x_continuous
#'   scale_y_continuous scale_color_manual scale_color_grey
#'   scale_linetype_manual scale_linewidth_manual labs theme_bw theme
#'   element_text unit guides guide_legend
#' @importFrom grDevices hcl.colors
#' @importFrom rlang .data
#' @export
#'
#' @examples
#' # Three-period synthetic dataset with distinct trajectory shapes:
#' # Site 1: up-then-down (peak in P2).
#' # Site 2: monotone up.
#' # Per-partition Shannon entropy (bits):
#' #   P1 (Jan)  s1: 0.469  s2: 0.469
#' #   P2 (Feb)  s1: 1.522  s2: 0.881
#' #   P3 (Mar)  s1: 0.722  s2: 1.522
#' df <- data.frame(
#'   s1 = c(
#'     c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L),         # P1: 9:1
#'     c(1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 3L, 3L),         # P2: 4:4:2
#'     c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L)          # P3: 8:2
#'   ),
#'   s2 = c(
#'     c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L),         # P1: 9:1
#'     c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L),         # P2: 7:3
#'     c(1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 3L, 3L)          # P3: 4:4:2
#'   ),
#'   Date = rep(
#'     seq(as.Date("2020-01-01"), by = "month", length.out = 3L),
#'     each = 10L
#'   )
#' )
#'
#' part_data <- partition_time_windows(
#'   data          = df,
#'   n_sites       = 2L,
#'   window_length = 1L,
#'   window_type   = 3L,
#'   start_date    = "2020-01-01",
#'   end_date      = "2020-04-01",
#'   removez       = FALSE,
#'   removesngl    = FALSE
#' )
#'
#' result <- plot_entropy_trajectories(part_data = part_data)
#' print(result$Plot)
plot_entropy_trajectories <- function(part_data,
                                      sites            = NULL,
                                      labels           = NULL,
                                      site_colors      = NULL,
                                      by_group         = FALSE,
                                      groups_list      = NULL,
                                      line_type_groups = NULL,
                                      line_size_groups = NULL,
                                      transformation   = NULL,
                                      line_size        = 1.5,
                                      legend           = TRUE,
                                      legend_text_size = 12,
                                      x_angle          = 45,
                                      grayscale        = FALSE,
                                      plot_title       = "Shannon Entropy Trajectories") {

  # ---------------------------------------------------------------------------
  # 1. Validate part_data
  # ---------------------------------------------------------------------------
  required_fields <- c("Clusters", "Max_Entropy", "Dates_Labels", "N_partitions")
  missing_fields  <- setdiff(required_fields, names(part_data))
  if (length(missing_fields) > 0L)
    stop("`part_data` is missing required fields: ",
         paste(missing_fields, collapse = ", "),
         ". Ensure `part_data` is the direct output of partition_time_windows().", 
         call. = FALSE)

  n_part <- part_data$N_partitions

  if (n_part < 1L)
    stop("`part_data$N_partitions` is 0. There are no partitions to plot.", 
         call. = FALSE)

  # ---------------------------------------------------------------------------
  # 2. Partition labels
  # ---------------------------------------------------------------------------
  if (is.null(labels)) {
    labels <- part_data$Dates_Labels
  } else {
    labels <- as.character(labels)
    if (length(labels) != n_part)
      stop("`labels` must have length equal to `N_partitions` (", n_part, ").", 
           call. = FALSE)
  }

  # ---------------------------------------------------------------------------
  # 3. Validate transformation
  # ---------------------------------------------------------------------------
  if (!is.null(transformation)) {
    if (!inherits(transformation, c("transform", "trans")))
      stop("`transformation` must be an object of class 'transform' produced by ",
           "scales::trans_new(), or NULL for the identity (no transformation).", 
           call. = FALSE)
  }

  # ---------------------------------------------------------------------------
  # 4. Validate by_group arguments before the main loop
  # ---------------------------------------------------------------------------
  if (isTRUE(by_group)) {
    if (is.null(groups_list) || !is.list(groups_list) || length(groups_list) == 0L)
      stop("`groups_list` must be a non-empty list of integer vectors ",
           "when `by_group = TRUE`.", call. = FALSE)
  }

  # ---------------------------------------------------------------------------
  # 5. Assemble long-format data frame across all partitions
  # ---------------------------------------------------------------------------
  frames <- vector("list", n_part)

  for (i in seq_len(n_part)) {

    cl_i <- part_data$Clusters[[i]]
    df_i <- cl_i$DataFrame

    # Empty partition — skip; its absence creates a gap in the trajectory.
    if (is.null(df_i) || nrow(df_i) == 0L) next

    # max_class is taken directly from the pre-computed Max_Entropy field.
    # If the user pre-relabeled the partition DataFrames, they are responsible
    # for also updating Max_Entropy[i] to 1L before calling this function.
    me_val      <- part_data$Max_Entropy[i]
    max_class_i <- if (is.na(me_val)) NA_integer_ else as.integer(me_val)

    df_i$max_class <- max_class_i
    df_i$period    <- i
    df_i$coverage  <- labels[i]

    frames[[i]] <- df_i
  }

  frames <- Filter(Negate(is.null), frames)

  if (length(frames) == 0L)
    stop("No data are available for plotting: all partitions are empty.", 
         call. = FALSE)

  df_all           <- do.call(rbind, frames)
  rownames(df_all) <- NULL

  # ---------------------------------------------------------------------------
  # 6. Subset to requested sites
  # ---------------------------------------------------------------------------
  all_present <- sort(unique(as.integer(df_all$sites)))

  if (is.null(sites)) {
    sites <- all_present
  } else {
    sites  <- sort(as.integer(sites))
    absent <- setdiff(sites, all_present)
    if (length(absent) > 0L)
      warning(length(absent), " requested site(s) not found in any partition ",
              "(zero entropy, singleton removal, or empty window): ",
              paste(absent, collapse = ", "), ". These sites are omitted from the plot.")
    sites <- intersect(sites, all_present)
  }

  if (length(sites) == 0L)
    stop("No requested sites are present in any partition.", call. = FALSE)

  df_all <- df_all[df_all$sites %in% sites, , drop = FALSE]

  # Coerce for ggplot2: sites as ordered factor (preserves numeric sort order
  # in the legend), class and period as appropriate types.
  df_all$sites  <- factor(df_all$sites, levels = as.character(sort(sites)))
  df_all$class  <- as.factor(df_all$class)
  df_all$period <- as.integer(df_all$period)

  existing_sites <- levels(df_all$sites)     # character, sorted numerically
  n_sites_exist  <- length(existing_sites)

  # ---------------------------------------------------------------------------
  # 7. Assign site colours
  # ---------------------------------------------------------------------------
  # Auto-generate a qualitative palette; override any user-specified entries.
  auto_palette        <- grDevices::hcl.colors(n_sites_exist, palette = "Dark 2")
  names(auto_palette) <- existing_sites

  if (!is.null(site_colors)) {
    if (!is.character(site_colors) || is.null(names(site_colors)))
      stop("`site_colors` must be a named character vector with site indices as ",
           "names (e.g. c(\"681\" = \"#FB8072\", \"501\" = \"#80B1D3\")).", 
           call. = FALSE)
    sc_keys <- as.character(names(site_colors))
    matches <- intersect(sc_keys, existing_sites)
    if (length(matches) > 0L)
      auto_palette[matches] <- site_colors[sc_keys %in% matches]
  }

  final_palette <- auto_palette    # named character: site_char -> colour

  # ---------------------------------------------------------------------------
  # 8. x-axis breaks and labels
  # ---------------------------------------------------------------------------
  m_min    <- min(df_all$period)
  m_max    <- max(df_all$period)
  x_breaks <- seq.int(m_min, m_max)
  x_labels <- labels[x_breaks]

  # ---------------------------------------------------------------------------
  # 9. by_group: assign group membership and resolve line aesthetics
  # ---------------------------------------------------------------------------
  if (isTRUE(by_group)) {

    # Coerce group site indices to character for factor-level matching.
    groups_chr <- lapply(groups_list, function(g) as.character(as.integer(g)))

    # Auto-create remainder group for sites not covered by any explicit group.
    assigned  <- unlist(groups_chr, use.names = FALSE)
    remainder <- setdiff(existing_sites, assigned)
    if (length(remainder) > 0L)
      groups_chr[[length(groups_chr) + 1L]] <- remainder

    n_groups <- length(groups_chr)

    if (n_groups > 6L)
      stop("Total number of groups (explicit + automatic remainder) is ",
           n_groups, ". Maximum allowed is 6.", call. = FALSE)

    # Default line types: solid for group 1, dashed for all others.
    if (is.null(line_type_groups)) {
      line_type_groups <- c("solid", rep("dashed", n_groups - 1L))
    } else {
      if (length(line_type_groups) != n_groups)
        stop("`line_type_groups` must have length ", n_groups,
             " (number of explicit groups plus the automatic remainder group).", 
             call. = FALSE)
    }

    # Default line widths: 2 for group 1, 1 for all others.
    if (is.null(line_size_groups)) {
      line_size_groups <- c(2, rep(1, n_groups - 1L))
    } else {
      if (length(line_size_groups) != n_groups)
        stop("`line_size_groups` must have length ", n_groups,
             " (number of explicit groups plus the automatic remainder group).", 
             call. = FALSE)
    }

    # Assign group integer to each row, matched on the sites factor level.
    df_all$group <- NA_integer_
    for (g in seq_len(n_groups)) {
      mask               <- as.character(df_all$sites) %in% groups_chr[[g]]
      df_all$group[mask] <- g
    }
    df_all$group <- as.factor(df_all$group)

  }

  # ---------------------------------------------------------------------------
  # 10. y-axis label
  # ---------------------------------------------------------------------------
  y_label <- if (is.null(transformation)) {
    "Shannon Entropy"
  } else {
    paste0(transformation$name, "(Shannon Entropy)")
  }

  # ---------------------------------------------------------------------------
  # 11. Build ggplot
  # ---------------------------------------------------------------------------

  # --- 11a. Base plot and geom_line layer ------------------------------------
  # The .data pronoun from rlang is used throughout aes() to make column
  # references unambiguous to R CMD check's static analyser.
  if (isTRUE(by_group)) {
    p <- ggplot2::ggplot(
      df_all,
      ggplot2::aes(x         = .data$period,
                   y         = .data$entropies,
                   colour    = .data$sites,
                   linetype  = .data$group,
                   linewidth = .data$group)
    ) +
      ggplot2::geom_line() +
      ggplot2::scale_linetype_manual(values = line_type_groups,
                                     name   = "Group") +
      ggplot2::scale_linewidth_manual(values = line_size_groups,
                                      name   = "Group")
  } else {
    p <- ggplot2::ggplot(
      df_all,
      ggplot2::aes(x      = .data$period,
                   y      = .data$entropies,
                   colour = .data$sites)
    ) +
      ggplot2::geom_line(linewidth = line_size)
  }

  # --- 11b. Colour scale: grayscale overrides site_colors -------------------
  if (isTRUE(grayscale)) {
    p <- p + ggplot2::scale_color_grey(name = "Site")
  } else {
    p <- p + ggplot2::scale_color_manual(values = final_palette, name = "Site")
  }

  # --- 11c. Axes and y transformation ----------------------------------------
  p <- p + ggplot2::scale_x_continuous(breaks = x_breaks, labels = x_labels,
                                       expand = ggplot2::expansion(add = 0.65))

  if (!is.null(transformation)) {
    p <- p + ggplot2::scale_y_continuous(trans = transformation,
                                         name  = y_label)
  } else {
    p <- p + ggplot2::scale_y_continuous(name = y_label)
  }

  # --- 11d. Labels, theme, guides --------------------------------------------
  p <- p +
    ggplot2::labs(title = plot_title,
                  x     = "Partition") +
    ggplot2::theme_bw() +
    ggplot2::theme(
      plot.title       = ggplot2::element_text(hjust = 0.5),
      axis.text.x      = ggplot2::element_text(angle = x_angle, hjust = 1),
      legend.text      = ggplot2::element_text(size  = legend_text_size),
      legend.key.width = ggplot2::unit(2, "cm")
    ) +
    ggplot2::guides(
      colour = ggplot2::guide_legend(override.aes = list(linewidth = 3))
    )

  # Legend suppression applied last to override any theme defaults.
  if (!isTRUE(legend))
    p <- p + ggplot2::theme(legend.position = "none")

  # ---------------------------------------------------------------------------
  # 12. Return
  # ---------------------------------------------------------------------------
  list(
    Data_Frame = df_all,
    Plot       = p,
    Colors     = final_palette,
    XBreaks    = x_breaks,
    XLabels    = x_labels
  )
}

#' @title Relabel Entropy Classes
#' @description Relabels GMM cluster labels so that class 1 is always the
#'   highest-entropy group, class 2 the second highest, and so on.
#'
#' @details
#' For univariate input (the package's typical use case), \code{\link[mclust]{Mclust}}
#' orders cluster components by increasing mean. Applied to per-site Shannon
#' entropy values, this places the highest-entropy group at the highest
#' label number (label \code{G} for \code{G} fitted components). This
#' function flips that convention so that label \code{1} always denotes
#' the highest-entropy group, which is more natural for downstream
#' filtering ("class 1 = top") and visual labeling. Cluster identities
#' and means are unchanged; only the integer labels are remapped.
#'
#' \strong{Sentinel preservation.} The class label \code{999L} marks
#' undifferentiated groups (all entropies equal in
#' \code{\link{cluster_sites_by_entropy}}) and is never relabeled. If a
#' data frame contains a mixture of real GMM labels and one or more
#' \code{999} entries, only the real labels are ranked and overwritten;
#' the \code{999} rows pass through unchanged.
#'
#' \strong{No-op return paths.} The input is returned unchanged (with at
#' most a \code{num_classes} column added) in four situations:
#' \itemize{
#'   \item Missing required columns (\code{class} or \code{entropies}) —
#'     warns and returns input.
#'   \item Zero rows.
#'   \item All rows already carry the sentinel \code{999L}.
#'   \item Only one class label is present (relabeling is a no-op).
#' }
#'
#' @param df A data frame containing clustering results. Must have columns
#'   \code{class} (integer cluster labels from
#'   \code{\link{cluster_sites_by_entropy}}) and \code{entropies}
#'   (numeric Shannon entropy values per row).
#'
#' @return The data frame with a relabeled \code{class} column (integer)
#'   and an added \code{num_classes} column reporting the count of
#'   distinct labels present after relabeling.
#'
#' @seealso \code{\link{cluster_sites_by_entropy}} for the upstream
#'   clustering step; \code{\link{plot_entropy_trajectories}} and
#'   \code{\link{plot_site_class_trajectory}} for downstream consumers
#'   that rely on the relabeled class convention;
#'   \code{\link{partition_time_windows}} which calls
#'   \code{cluster_sites_by_entropy} per window.
#'
#' @importFrom stats aggregate setNames
#' @export
#'
#' @examples
#' # Standard case: three classes, ranked by mean entropy.
#' df <- data.frame(
#'   sites     = 1:6,
#'   entropies = c(0.1, 0.1, 0.5, 0.5, 1.2, 1.3),
#'   class     = c(1L, 1L, 2L, 2L, 3L, 3L)
#' )
#' relabel_entropy_classes(df)
#' # Class 3 (highest mean entropy 1.25) -> 1; class 2 (0.5) -> 2;
#' # class 1 (0.1) -> 3.
#'
#' # Sentinel preservation: 999 rows pass through unchanged even when
#' # mixed with real classes.
#' df_mixed <- data.frame(
#'   sites     = 1:4,
#'   entropies = c(0.5, 1.2, 0.3, 0.4),
#'   class     = c(1L, 2L, 999L, 1L)
#' )
#' relabel_entropy_classes(df_mixed)
#' # Class 2 (highest) -> 1; class 1 -> 2; class 999 stays 999.
relabel_entropy_classes <- function(df) {
  
  # --- Guards: cases where relabeling is undefined or a no-op ---------------
  if (!all(c("class", "entropies") %in% names(df))) {
    warning("relabel_entropy_classes: 'class' or 'entropies' column missing ",
            "- returning input unchanged.", call. = FALSE)
    return(df)
  }
  
  if (nrow(df) == 0L)
    return(df)
  
  # Sentinel: all entropies identical, one undifferentiated group.
  # Relabeling would overwrite 999 -> 1 and destroy the sentinel.
  if (all(df$class == 999L)) {
    df$num_classes <- 1L
    return(df)
  }
  
  # Single class present — relabeling one group is a no-op.
  if (length(unique(df$class)) == 1L) {
    df$num_classes <- 1L
    return(df)
  }
  
  # --- Relabel: rank classes by mean entropy descending ---------------------
  # Sentinel rows (999) are excluded from ranking and pass through
  # unchanged. Only the real GMM labels are reordered.
  is_sentinel <- df$class == 999L
  
  if (any(is_sentinel) && length(unique(df$class[!is_sentinel])) <= 1L) {
    # Mixed sentinel + at most one real class: no real ranking to perform.
    df$num_classes <- length(unique(df$class))
    return(df)
  }
  
  df_real          <- if (any(is_sentinel)) df[!is_sentinel, , drop = FALSE] else df
  class_stats      <- aggregate(entropies ~ class, data = df_real, FUN = mean)
  class_stats$rank <- rank(-class_stats$entropies, ties.method = "first")
  mapping          <- setNames(class_stats$rank, class_stats$class)
  
  # Coerce to integer (not factor) for a consistent class column type across
  # all return paths. as.numeric(factor) returns level codes which happen to
  # equal the label values here, but plain integer is unambiguous and safe
  # for all downstream comparisons (==, %in%, is.integer).
  if (any(is_sentinel)) {
    df$class[!is_sentinel] <- as.integer(mapping[as.character(df$class[!is_sentinel])])
    # Sentinel rows keep their 999 — no assignment needed.
  } else {
    df$class <- as.integer(mapping[as.character(df$class)])
  }
  
  df$num_classes <- length(unique(df$class))
  
  df
}

#' @title Detect Temporal Change Points (HDcpDetect)
#' @description Runs high-dimensional change point detection on time-series
#'   matrices (typically Hellinger distance matrices) over one or more time
#'   windows, using either binary segmentation or wild binary segmentation.
#' @export
#'
#' @details
#' The function applies \code{\link[HDcpDetect]{binary.segmentation}} or
#' \code{\link[HDcpDetect]{wild.binary.segmentation}} repeatedly to slices of
#' \code{data_matrix}, advancing the slice forward by one row at each step.
#' A total of \code{n_timesteps + 1} detections are performed: one on the
#' initial window \code{[min_window, max_window]}, then \code{n_timesteps}
#' additional detections on subsequent windows.
#'
#' Two window-advancement modes are supported, mirroring
#' \code{\link{detect_changepoints_ecp}}:
#' \itemize{
#'   \item \strong{Expanding} (\code{rolling_window = FALSE}, default): the
#'     start index is fixed at \code{min_window}; the end index grows by 1
#'     each step. Each iteration uses one more row than the previous.
#'     Natural for online surveillance, where change-point detection is
#'     re-run as new data accumulates.
#'   \item \strong{Rolling} (\code{rolling_window = TRUE}): both start and
#'     end indices advance by 1 each step, keeping the window length
#'     constant. Natural for retrospective windowed analysis.
#' }
#'
#' \strong{Choice of segmentation method.} Binary segmentation
#' (\code{wild = FALSE}, default) is the classical recursive method: it
#' finds the most likely change point, splits the series, and recurses.
#' Wild binary segmentation (\code{wild = TRUE}) is the variant of
#' Fryzlewicz (2014) that draws random subintervals to improve detection
#' of multiple closely-spaced change points; it accepts an additional
#' \code{M} argument controlling the number of random intervals.
#'
#' @param data_matrix A numeric matrix (Time x Features). Rows are time
#'   points, columns are features (e.g. Hellinger distances per site).
#' @param min_window Integer. Starting row index of the initial window.
#' @param max_window Integer. Ending row index of the initial window. Must
#'   not exceed \code{nrow(data_matrix)}.
#' @param n_timesteps Integer >= 0. Number of \emph{additional} detections
#'   to run after the initial window. The function performs
#'   \code{n_timesteps + 1} detections in total. Set to \code{0} for a
#'   single detection on the initial window.
#' @param rolling_window Logical. If \code{TRUE}, both window endpoints
#'   advance per step (sliding window of constant length). If \code{FALSE}
#'   (default), only the end advances (expanding window).
#' @param wild Logical. If \code{TRUE}, uses
#'   \code{\link[HDcpDetect]{wild.binary.segmentation}}; if \code{FALSE}
#'   (default), uses \code{\link[HDcpDetect]{binary.segmentation}}. See
#'   Details for the methodological distinction.
#' @param ... Additional arguments passed to the chosen segmentation
#'   function. Note that \code{M} (number of random intervals) is only
#'   accepted by \code{wild.binary.segmentation}; passing it with
#'   \code{wild = FALSE} produces an "unused argument" error.
#'
#' @return A named list:
#' \item{Points_List}{List of integer vectors \code{c(start, end)} giving
#'   the row index window used at each step.}
#' \item{HDcp_list}{List of full segmentation result objects, one per
#'   step. Steps where \code{end_idx} exceeds the matrix rows return
#'   \code{NULL} with a warning.}
#'
#' @section Warnings and errors:
#' \describe{
#'   \item{\code{stop()}}{Triggered if \code{max_window} exceeds
#'     \code{nrow(data_matrix)}: the initial window cannot be
#'     constructed and no detection can run.}
#'   \item{\code{warning()}}{Triggered (and execution continues) if
#'     \code{max_window + n_timesteps} exceeds \code{nrow(data_matrix)},
#'     in which case the offending later iterations are skipped and
#'     \code{NULL} entries appear in \code{HDcp_list}.}
#' }
#'
#' @seealso \code{\link{detect_changepoints_ecp}} for the energy-statistic
#'   alternative; \code{\link{calculate_hellinger_matrix}} for the typical
#'   upstream input.
#'
#' @importFrom HDcpDetect binary.segmentation wild.binary.segmentation
#'
#' @examples
#' set.seed(42)
#' baseline = matrix(rnorm(50, mean = 0, sd = 0.1), nrow = 10, ncol = 5)
#' variant  = matrix(rnorm(50, mean = 3, sd = 0.1), nrow = 10, ncol = 5)
#' data_mat = rbind(baseline, variant)
#'
#' # Single-window detection with the default binary segmentation.
#' res = detect_changepoints_hdcp(
#'   data_matrix = data_mat,
#'   min_window  = 1,
#'   max_window  = 20,
#'   n_timesteps = 0
#' )
#' print(res$HDcp_list[[1]])
#'
#' # Wild binary segmentation with a custom number of random intervals.
#' res_wild = detect_changepoints_hdcp(
#'   data_matrix = data_mat,
#'   min_window  = 1,
#'   max_window  = 20,
#'   n_timesteps = 0,
#'   wild        = TRUE,
#'   M           = 100
#' )
#' print(res_wild$HDcp_list[[1]])
detect_changepoints_hdcp = function(data_matrix,
                                    min_window,
                                    max_window,
                                    n_timesteps,
                                    rolling_window = FALSE,
                                    wild           = FALSE,
                                    ...) {
  
  # --- 1. Validation ---------------------------------------------------------
  if (!is.matrix(data_matrix))
    stop("`data_matrix` must be a matrix.", call. = FALSE)
  
  if (!is.numeric(n_timesteps) || length(n_timesteps) != 1L || n_timesteps < 0)
    stop("`n_timesteps` must be a single non-negative integer.",
         call. = FALSE)
  
  n_rows = nrow(data_matrix)
  
  # Hard guard: if the initial window itself is impossible, no detection
  # can run. Returning a list of NULLs would be silently degraded output;
  # an explicit error forces the caller to fix the call.
  if (max_window > n_rows)
    stop("`max_window` (", max_window, ") exceeds nrow(data_matrix) (",
         n_rows, "). The initial window [", min_window, ", ",
         max_window, "] cannot be constructed.",
         call. = FALSE)
  
  # Soft guard: if later iterations would go past the matrix, warn with
  # concrete numbers so the caller can decide whether to fix or accept
  # the degraded output. Skipped iterations return NULL entries.
  needed = max_window + n_timesteps
  if (n_rows < needed) {
    n_skipped = needed - n_rows
    warning(sprintf(
      "`data_matrix` has %d rows but max_window (%d) + n_timesteps (%d) requires %d. %d step(s) will be skipped and return NULL.",
      n_rows, max_window, n_timesteps, needed, n_skipped),
      call. = FALSE)
  }
  
  # Capture ... once so it does not need to be re-evaluated on every iteration.
  extra_args = list(...)
  
  # --- 2. Per-step worker ----------------------------------------------------
  
  run_step = function(i) {
    start_idx = if (rolling_window) (min_window + i) else min_window
    end_idx   = max_window + i
    
    if (end_idx > n_rows) {
      warning("Step ", i, ": end_idx (", end_idx,
              ") exceeds matrix rows (", n_rows, "); skipping.",
              call. = FALSE)
      return(list(window = c(start_idx, end_idx),
                  hdcp   = NULL))
    }
    
    subset_mat = data_matrix[start_idx:end_idx, , drop = FALSE]
    
    hdcp_res = tryCatch({
      utils::capture.output({
        res_inner = if (wild) {
          do.call(HDcpDetect::wild.binary.segmentation,
                  c(list(subset_mat), extra_args))
        } else {
          do.call(HDcpDetect::binary.segmentation,
                  c(list(subset_mat), extra_args))
        }
      })
      res_inner
    }, error = function(e) {
      warning("Step ", i, ": HDcpDetect call failed: ",
              conditionMessage(e), "; returning NULL.",
              call. = FALSE)
      NULL
    })
    
    list(window = c(start_idx, end_idx),
         hdcp   = hdcp_res)
  }
  
  # --- 3. Execute ------------------------------------------------------------
  results_list = lapply(0:n_timesteps, run_step)
  
  # --- 4. Return -------------------------------------------------------------
  list(
    Points_List = lapply(results_list, `[[`, "window"),
    HDcp_list   = lapply(results_list, `[[`, "hdcp")
  )
}

#' @title Decode Amino Acid Sequences
#' @description Converts an integer-encoded matrix of amino acids back to its
#'   character representation under the package's 25-symbol alphabet. Inverse
#'   of \code{\link{encode_aa_sequence}}.
#'
#' @details
#' Decoding is a single vectorised lookup against the fixed alphabet
#' (\code{A, R, N, D, C, Q, E, G, H, I, L, K, M, F, P, S, T, W, Y, V} for
#' the twenty standard residues, then \code{B, Z, X, *, -} for ambiguous
#' codes, stop codons, and gaps). The input matrix is flattened, indexed
#' against the alphabet vector in one operation, and reshaped — there is
#' no per-row or per-element loop.
#'
#' Values outside the valid range \code{1:25} (including \code{0}, which
#' \code{\link{encode_aa_sequence}} produces for unrecognised characters)
#' are returned as the sentinel string \code{"0"}. This preserves
#' round-trip consistency with the encoder: encoding then decoding any
#' character originally outside the alphabet yields \code{"0"} rather
#' than throwing an error. \code{NA} values are also mapped to
#' \code{"0"}.
#'
#' Row and column names of the input matrix are preserved on the output.
#'
#' @param matrix_input Numeric matrix of integer-encoded amino acids,
#'   typically the output of \code{\link{encode_aa_sequence}}. Values
#'   outside \code{1:25} (including \code{0} and \code{NA}) decode to
#'   the sentinel string \code{"0"}. A non-matrix input is coerced via
#'   \code{\link[base]{as.matrix}}.
#'
#' @return A character matrix of the same dimensions as \code{matrix_input},
#'   with the same \code{dimnames}. Each cell is either a one-character
#'   amino acid code from the 25-symbol alphabet, or the sentinel
#'   \code{"0"} for out-of-range and missing values.
#'
#' @seealso \code{\link{encode_aa_sequence}} for the inverse operation;
#'   \code{\link{fasta_to_char_matrix}} for the FASTA-to-character-matrix
#'   step that typically precedes encoding.
#'
#' @export
#'
#' @examples
#' # 1. Decode a numeric matrix.
#' num_mat = matrix(c(1, 2, 25, 10), nrow = 2, byrow = TRUE)
#' decoded = decode_aa_sequence(num_mat)
#' print(decoded)
#' # 1 -> "A", 2 -> "R", 25 -> "-", 10 -> "I"
#'
#' # 2. Round-trip consistency check (excluding unknowns).
#' orig = matrix(c("A", "C", "W", "G"), nrow = 2)
#' enc  = encode_aa_sequence(orig)
#' dec  = decode_aa_sequence(enc)
#' all.equal(orig, dec)
#'
#' # 3. Out-of-range and NA values both decode to the sentinel "0".
#' decode_aa_sequence(matrix(c(0, NA, 30, 5), nrow = 2))
decode_aa_sequence = function(matrix_input) {
  aa_chars = c("A","R","N","D","C","Q","E","G","H","I",
               "L","K","M","F","P","S","T","W","Y","V",
               "B","Z","X","*","-")
  
  if (!is.matrix(matrix_input)) matrix_input = as.matrix(matrix_input)
  
  # Mask out-of-range and NA values so indexing preserves length:
  # 0 would collapse a position; NA preserves it (and we map NA -> "0" below).
  input_vec  = as.vector(matrix_input)
  safe_idx   = input_vec
  safe_idx[is.na(safe_idx) | safe_idx < 1L | safe_idx > 25L] = NA_integer_
  
  decoded_vec = aa_chars[safe_idx]
  decoded_vec[is.na(decoded_vec)] = "0"
  
  matrix(decoded_vec, nrow = nrow(matrix_input), ncol = ncol(matrix_input),
         dimnames = dimnames(matrix_input))
}

#' SARS-CoV-2 Surface Glycoprotein Sequences -- NCBI Demo Sample
#'
#' @description
#' A compressed FASTA file containing a random sample of size 100 of SARS-CoV-2
#' surface glycoprotein (Spike protein) amino acid sequences, drawn from a
#' dataset of 137,132 complete sequences downloaded from the NCBI SARS-CoV-2
#' Data Hub on October 12, 2021. Intended to demonstrate the full
#' \code{ViralEntropR} pipeline on real-world surveillance data.
#'
#' @details
#' The file is accessed at runtime via:
#' \preformatted{
#' path  <- system.file("extdata", "sarscov2_sample.fasta.gz",
#'                      package = "ViralEntropR")
#' fasta <- Biostrings::readAAStringSet(path)
#' }
#'
#' \strong{Sample file contents:}
#' \itemize{
#'   \item \strong{Sample size}: 100 random sequences
#'   \item \strong{Protein}: Surface glycoprotein (Spike, S protein)
#'   \item \strong{Organism}: \emph{Severe acute respiratory syndrome
#'     coronavirus 2} (SARS-CoV-2), NCBI Taxonomy ID: 2697049
#'   \item \strong{Completeness}: Complete sequences only
#'   \item \strong{Format}: Gzip-compressed FASTA (\code{.fasta.gz}),
#'     readable directly by \code{\link[Biostrings]{readAAStringSet}}
#' }
#'
#' \strong{Full dataset:}
#' The complete 137,132-sequence dataset (~181.5 MB, uncompressed FASTA) is
#' archived on Zenodo (DOI: \doi{10.5281/zenodo.19040165});
#' \code{\link[Biostrings]{readAAStringSet}} reads it directly.
#'
#' \strong{Header format:}
#' Sequence headers follow the NCBI Virus export format. Dates and countries
#' can be extracted directly using the package helper functions:
#' \preformatted{
#' dates     <- extract_fasta_dates(fasta, option = 1)
#' countries <- extract_fasta_countries(fasta, position = 2)
#' }
#'
#' \strong{Subsampling:}
#' To keep the package within CRAN size limits, the bundled sample was
#' generated by random sampling of 100 sequences. The full provenance and
#' reproducible sampling script are in \code{data-raw/sarscov2_ncbi.R} in
#' the package source.
#'
#' @section Provenance:
#' Downloaded from the NCBI SARS-CoV-2 Data Hub
#' (\url{https://www.ncbi.nlm.nih.gov/labs/virus/vssi/}) on
#' \strong{October 12, 2021} with the following filters:
#' \enumerate{
#'   \item Virus: Severe acute respiratory syndrome coronavirus 2
#'     (SARS-CoV-2), taxid: 2697049
#'   \item Nucleotide completeness: complete
#'   \item Protein: surface glycoprotein
#' }
#' Result: \strong{n = 137,132 sequences}, file size ~181.5 MB (uncompressed).
#'
#' @section License:
#' NCBI sequence data is a US Government work and is in the public domain
#' within the United States. Data from international contributors is subject
#' to the INSDC open-access policy (Karsch-Mizrachi et al., 2025;
#' \doi{10.1093/nar/gkae1058}). The compiled dataset is released under
#' CC0 1.0 Universal. Individual GenBank accession numbers in the FASTA
#' headers provide full traceability to original submissions.
#'
#' @format A gzip-compressed FASTA file (\code{.fasta.gz}) readable by
#'   \code{\link[Biostrings]{readAAStringSet}}. Each record contains:
#'   \describe{
#'     \item{Header}{NCBI Virus format: accession, country, and collection date 
#'     metadata separated by \code{|}.}
#'     \item{Sequence}{Amino acid sequence using the standard IUPAC one-letter
#'       code. Ambiguous residues (B, X, Z) and gaps (-) may be present and
#'       can be removed with \code{\link{filter_ambiguous_sequences}}.}
#'   }
#'
#' @name sarscov2_sample
#'
#' @source \url{https://www.ncbi.nlm.nih.gov/labs/virus/vssi/}
#'
#' @references
#' Clark K, Karsch-Mizrachi I, Lipman DJ, Ostell J, Sayers EW (2016).
#' GenBank.
#' \emph{Nucleic Acids Research}, 44(D1), D67--D72.
#' \doi{10.1093/nar/gkv1276}
#'
#' Sayers EW, Bolton EE, Brister JR, et al. (2022).
#' Database resources of the National Center for Biotechnology Information.
#' \emph{Nucleic Acids Research}, 50(D1), D20--D26.
#' \doi{10.1093/nar/gkab1112}
#'
#' NCBI Virus \[Internet\]. Bethesda (MD): National Library of Medicine (US),
#' National Center for Biotechnology Information; \[2020\] --
#' \[cited 2021 Oct 12\]. Available from:
#' \url{https://www.ncbi.nlm.nih.gov/labs/virus/vssi/}
#'
#' @examples
#' \donttest{
#' path  <- system.file("extdata", "sarscov2_sample.fasta.gz",
#'                      package = "ViralEntropR")
#' fasta <- Biostrings::readAAStringSet(path)
#'
#' # Number of sequences in the demo sample
#' length(fasta)
#'
#' # Inspect first 3 headers
#' names(fasta)[1:3]
#'
#' # Extract collection dates
#' dates <- extract_fasta_dates(fasta, option = 1)
#' head(dates$corrected_dates)
#'
#' # Extract countries
#' countries <- extract_fasta_countries(fasta, position = 2)
#' table(countries$countries)
#'
#' # Convert to character matrix for pipeline entry
#' char_mat <- fasta_to_char_matrix(fasta)
#' dim(char_mat)
#'
#' # The full 137,132-sequence dataset (~181.5 MB) is available on Zenodo:
#' # https://zenodo.org/records/19040165
#' # Download sequences.fasta manually and load with:
#' # fasta_full <- Biostrings::readAAStringSet("path/to/sequences.fasta")
#' }
NULL


#' SARS-CoV-2 VOC/VOI Curated Variant Metadata
#'
#' @description
#' A pre-built named list containing curated biological and epidemiological
#' metadata for 12 SARS-CoV-2 Variants of Concern (VOC) and Variants of
#' Interest (VOI): Alpha, Beta, Delta, Epsilon, Eta, Gamma, Iota, Kappa, Lambda, 
#' Omicron, Theta, Zeta. Includes mutation profiles, nomenclature,
#' temporal detection records, defining SNPs, and a fully citable reference
#' table with 21 verified references.
#'
#' @details
#' The object is available automatically after \code{library(ViralEntropR)}
#' via lazy loading. It is produced by \code{\link{get_variants}} from the
#' curated Excel workbook \code{SARS_CoV_2_VOC_VOI.xlsx} (not bundled with
#' the package; see \code{data-raw/sarscov2_variants.R} for the reproducible
#' build script).
#'
#' \strong{List elements:}
#' \describe{
#'   \item{\code{WHO_Label}}{List of 12. WHO variant label strings
#'     (e.g. \code{"Alpha"}, \code{"Omicron"}).}
#'   \item{\code{Pango_Lineage}}{List of 12. Pango lineage designations
#'     (e.g. \code{"B.1.1.7"}).}
#'   \item{\code{GISAID_Clade}}{Character(12). GISAID clade strings
#'     (e.g. \code{"GRY"}, \code{"GR/484A"}).}
#'   \item{\code{Nextstrain_Clade}}{Character(12). Nextstrain clade strings
#'     (e.g. \code{"20I/501Y.V1"}).}
#'   \item{\code{Country_First_Detected}}{List of 12. Country of first
#'     documented detection per variant.}
#'   \item{\code{Date_Earliest_Sample}}{Character(12). Month-Year of earliest
#'     documented sample (e.g. \code{"Sep-2020"}).}
#'   \item{\code{Date_First_Detected}}{Character(12). Month-Year of
#'     world-level first detection.}
#'   \item{\code{Date_First_Detected_US}}{Character(12). Month-Year of first
#'     US detection.}
#'   \item{\code{Spike_Mutations}}{List of 12. Character vectors of Spike
#'     protein mutations per variant, read from the Excel workbook.}
#'   \item{\code{Mutation_Sites}}{List of 12. Integer vectors of amino acid
#'     positions extracted from \code{Spike_Mutations}.}
#'   \item{\code{Defining_SNPs}}{List of 12. Canonical defining SNP strings
#'     (\code{NA} for variants not characterised in the reference set).}
#'   \item{\code{Defining_SNP_Sites}}{List of 12. Integer positions extracted
#'     from \code{Defining_SNPs}.}
#'   \item{\code{References}}{Named list with three elements:
#'     \code{$data} -- data frame of 21 verified references (columns: ID,
#'     Authors, Year, Title, Journal_Source, Volume_Issue_Pages, DOI, URL,
#'     Type, Variants_Covered, Data_Field, Citation);
#'     \code{$display(variant = NULL)} -- renders an interactive
#'     \code{\link[DT]{datatable}}, optionally filtered by WHO label
#'     (requires the \pkg{DT} package);
#'     \code{$cite(variant)} -- returns formatted citation strings for a
#'     given WHO label, suitable for manuscript use.}
#' }
#'
#' @section References:
#' The 21 curated references cover peer-reviewed articles, CDC MMWR government
#' reports, and outbreak.info surveillance database records. Full provenance
#' is available via \code{sarscov2_variants$References$data} or the
#' interactive display.
#'
#' @format A named list of 13 elements (12 metadata fields plus a structured 
#' References object). See Details for per-element descriptions.
#'
#' @docType data
#'
#' @usage data(sarscov2_variants)
#'
#' @source
#' Compiled from 21 peer-reviewed and surveillance sources; see
#' \code{sarscov2_variants$References$data} for the full reference table
#' with DOIs and URLs. Built from \code{SARS_CoV_2_VOC_VOI.xlsx} via
#' \code{\link{get_variants}}.
#'
#' @examples
#' # Object is available immediately after library(ViralEntropR)
#'
#' # --- Basic access ---------------------------------------------------------
#'
#' # All 12 WHO labels
#' unlist(sarscov2_variants$WHO_Label)
#'
#' # Pango lineage for Alpha
#' sarscov2_variants$Pango_Lineage[[
#'   which(unlist(sarscov2_variants$WHO_Label) == "Alpha")
#' ]]
#'
#' # World detection dates for all variants
#' data.frame(
#'   Variant  = unlist(sarscov2_variants$WHO_Label),
#'   Detected = sarscov2_variants$Date_First_Detected,
#'   Country  = unlist(sarscov2_variants$Country_First_Detected)
#' )
#'
#' # --- Mutation sites -------------------------------------------------------
#'
#' # Defining SNPs and sites for Delta
#' idx <- which(unlist(sarscov2_variants$WHO_Label) == "Delta")
#' sarscov2_variants$Defining_SNPs[[idx]]
#' sarscov2_variants$Defining_SNP_Sites[[idx]]
#'
#' # All Spike mutation sites for Omicron
#' sarscov2_variants$Mutation_Sites[[
#'   which(unlist(sarscov2_variants$WHO_Label) == "Omicron")
#' ]]
#'
#' # --- Nomenclature ---------------------------------------------------------
#'
#' data.frame(
#'   WHO        = unlist(sarscov2_variants$WHO_Label),
#'   PANGO      = unlist(sarscov2_variants$Pango_Lineage),
#'   GISAID     = sarscov2_variants$GISAID_Clade,
#'   Nextstrain = sarscov2_variants$Nextstrain_Clade
#' )
#'
#' # --- References -----------------------------------------------------------
#'
#' # Full reference data frame (21 rows)
#' sarscov2_variants$References$data[,
#'   c("ID", "Authors", "Year", "Journal_Source")
#' ]
#'
#' # Formatted citation strings for Gamma
#' sarscov2_variants$References$cite("Gamma")
#'
#' if (interactive() && requireNamespace("DT", quietly = TRUE)) {
#'   # Interactive DT table of all references
#'   sarscov2_variants$References$display()
#'
#'   # Filtered to Omicron references only
#'   sarscov2_variants$References$display(variant = "Omicron")
#' }
"sarscov2_variants"

#' @title Extract Dates from FASTA Sequence Names
#' @description Extracts date strings from the sequence name strings of an
#'   \code{AAStringSet} object loaded via
#'   \code{\link[Biostrings]{readAAStringSet}}. Several built-in date patterns
#'   are provided for the common date conventions used in NCBI and GISAID
#'   exports; a fully custom regex can also be supplied.
#'
#' @details
#' Date strings of the form \code{yyyy-mm-dd} are matched between pipe
#' characters (\code{|...|}) by default. Day value \code{00} (a common
#' GISAID convention indicating unknown collection day) is accepted in the
#' raw string and corrected to \code{01} before coercion to \code{Date}.
#' Both raw and corrected versions are returned, so the caller can decide
#' how to treat unknown-day records downstream.
#'
#' \strong{Choosing a built-in pattern.} The four options correspond to
#' the four most common date conventions in viral sequence repositories:
#' \itemize{
#'   \item \code{option = 1}: \code{yyyy-mm-dd} between pipes — GISAID
#'     export format, where the date is followed by additional
#'     pipe-delimited fields.
#'   \item \code{option = 2}: \code{yyyy-dd-mm} between pipes — some
#'     European data sources reverse day and month.
#'   \item \code{option = 3}: \code{yyyy-mm} between pipes — month-level
#'     resolution, useful when the source omits or hides the day. Pair
#'     with \code{date_format = "\%Y-\%m"}.
#'   \item \code{option = 4}: \code{yyyy-mm-dd} at end of header — NCBI
#'     Virus export format, where the collection date is the final
#'     field with no trailing delimiter. This is the format of the
#'     bundled \code{sarscov2_sample}.
#' }
#' For datasets where the date does not lie between pipes, supply a
#' \code{custom_pattern} matching whatever surrounding context the headers
#' provide.
#'
#' \strong{Coercion to Date.} When \code{date_format = "\%Y-\%m"} the
#' function uses \code{\link[zoo]{as.yearmon}} for coercion so that
#' year-month strings are handled correctly (base \code{as.Date} cannot
#' parse \code{"2021-05"} alone). For all other formats, base
#' \code{as.Date} with the supplied \code{date_format} is used.
#'
#' \strong{Output alignment.} All six elements of the return list are the
#' same length as the input \code{sequence}. Where extraction fails for a
#' record, the corresponding entries are \code{NA}; \code{missing_id}
#' lists the affected indices.
#'
#' @param sequence An \code{AAStringSet} object.
#' @param option Integer (1, 2, 3, or 4). Selects the built-in pattern when
#'   \code{custom_pattern} is not supplied. See Details.
#' @param date_format Character. \code{strptime}-style format string used
#'   to coerce extracted strings to \code{Date}. Default is
#'   \code{"\%Y-\%m-\%d"}. Use \code{"\%Y-\%m"} together with
#'   \code{option = 3}.
#' @param custom_pattern Character or \code{NULL}. A custom regex passed
#'   directly to \code{\link[stringr]{str_extract}}. When supplied,
#'   \code{option} is ignored. Default is \code{NULL}.
#'
#' @return A named list of six elements, each aligned with the input
#'   \code{sequence}:
#' \item{raw_date_strings}{Character vector of extracted date strings
#'   before any correction. \code{NA} where extraction failed.}
#' \item{corrected_date_strings}{Character vector with \code{-00}
#'   substrings replaced by \code{-01}. \code{NA} where extraction
#'   failed.}
#' \item{raw_dates}{\code{Date} vector coerced from
#'   \code{raw_date_strings}. \code{NA} for unparseable or missing
#'   strings (including any record where day = \code{00}).}
#' \item{corrected_dates}{\code{Date} vector coerced from
#'   \code{corrected_date_strings}. \code{NA} for unparseable or missing
#'   strings.}
#' \item{message}{Character string summarising extraction success.}
#' \item{missing_id}{Integer vector of indices where extraction failed,
#'   or \code{NA} if all extractions succeeded.}
#'
#' @seealso \code{\link{extract_fasta_countries}} for the country-extraction
#'   companion; \code{\link[Biostrings]{readAAStringSet}} for loading the
#'   input \code{AAStringSet}; \code{\link[zoo]{as.yearmon}} for the
#'   year-month coercion path.
#'
#' @importFrom stringr str_extract
#' @importFrom zoo as.Date as.yearmon
#' @export
#'
#' @examples
#' \donttest{
#' path_sample  <- system.file("extdata", "sarscov2_sample.fasta.gz",
#'                              package = "ViralEntropR")
#' fasta_sample <- Biostrings::readAAStringSet(path_sample)
#'
#' # Inspect header structure to confirm date field position.
#' sample(names(fasta_sample), 1)
#' # The bundled sample uses NCBI Virus format: date is at end of header.
#'
#' # Default usage on bundled sample: option = 4 for end-of-header dates.
#' dates <- extract_fasta_dates(fasta_sample, option = 4)
#' dates$message
#' head(dates$corrected_dates)
#' range(dates$corrected_dates, na.rm = TRUE)
#'
#' # Custom regex for non-standard headers:
#' dates_custom <- extract_fasta_dates(
#'   fasta_sample,
#'   custom_pattern = "[0-9]{4}-(0?[1-9]|1[0-2])-(0?[1-9]|[12][0-9]|3[01]|00)"
#' )
#' }
extract_fasta_dates <- function(sequence,
                                option         = 1,
                                date_format    = "%Y-%m-%d",
                                custom_pattern = NULL) {
  
  if (!requireNamespace("Biostrings", quietly = TRUE)) {
    stop("Package 'Biostrings' is required. Install it with:\n",
         "  install.packages('BiocManager')\n",
         "  BiocManager::install('Biostrings')",
         call. = FALSE)
  }
  
  # Use names() — public accessor — instead of internal slot access
  seq_names <- names(sequence)
  
  # Select regex pattern
  if (!is.null(custom_pattern)) {
    pattern <- custom_pattern
  } else {
    pattern <- switch(as.character(option),
                      "1" = "(?<=\\|)[0-9]{4}-(0?[1-9]|1[0-2])-(0?[1-9]|[12][0-9]|3[01]|00)(?=\\|)",
                      "2" = "(?<=\\|)[0-9]{4}-(0?[1-9]|[12][0-9]|3[01]|00)-(0?[1-9]|1[0-2])(?=\\|)",
                      "3" = "(?<=\\|)[0-9]{4}-(0?[1-9]|1[0-2])(?=\\|)",
                      "4" = "(?<=\\|)[0-9]{4}-(0?[1-9]|1[0-2])-(0?[1-9]|[12][0-9]|3[01]|00)$",
                      stop("`option` must be 1, 2, 3, or 4.", call. = FALSE)
    )
  }
  
  date_strings <- stringr::str_extract(seq_names, pattern)
  is_missing   <- is.na(date_strings)
  
  if (!any(is_missing)) {
    msg      <- "All date strings have been successfully extracted"
    miss_id  <- NA
  } else {
    msg      <- "There are date strings that have not been recognized"
    miss_id  <- which(is_missing)
  }
  
  raw_date_strings       <- date_strings
  corrected_date_strings <- gsub("-00", "-01", raw_date_strings)
  
  # Coerce to Date — yearmon path for yyyy-mm format
  case_yyyy_mm <- identical(date_format, "%Y-%m")
  
  if (case_yyyy_mm) {
    raw_dates       <- zoo::as.Date(zoo::as.yearmon(raw_date_strings,       format = "%Y-%m"))
    corrected_dates <- zoo::as.Date(zoo::as.yearmon(corrected_date_strings, format = "%Y-%m"))
  } else {
    raw_dates       <- as.Date(raw_date_strings,       format = date_format)
    corrected_dates <- as.Date(corrected_date_strings, format = date_format)
  }
  
  list(raw_date_strings       = raw_date_strings,
       corrected_date_strings = corrected_date_strings,
       raw_dates              = raw_dates,
       corrected_dates        = corrected_dates,
       message                = msg,
       missing_id             = miss_id)
}

#' @title Tabulate Site Frequency Evolution
#' @description Generates a frequency or proportion table showing the amino
#'   acid distribution at a specific site across multiple time partitions,
#'   optionally styled with \pkg{kableExtra} and saved as standalone HTML.
#'
#' @details
#' Aggregates amino acid counts per partition using
#' \code{\link{get_site_counts}}, optionally converts to relative
#' frequencies, applies kableExtra styling (column width, column
#' highlighting, striped rows), and optionally saves to disk. Row names
#' are decoded amino acid codes via \code{\link{decode_aa_sequence}} when
#' \code{use_letters = TRUE}.
#'
#' \strong{Empty partitions.} Partitions containing no observations
#' contribute all-zero columns. When \code{relative = TRUE}, division by
#' zero is avoided by treating empty-column sums as 1, leaving
#' proportions at zero. Inspect partition sizes via
#' \code{sapply(partitions, nrow)} before interpreting the table.
#'
#' \strong{Note on \code{zeros = FALSE}.} Setting \code{zeros = FALSE}
#' replaces numeric zeros with empty strings (\code{""}) for visual
#' clarity in the kable. This conversion forces the underlying data
#' frame to character storage; numeric operations (\code{sum},
#' \code{mean}, etc.) will not work on the returned \code{table}
#' element. Use \code{zeros = TRUE} (default) if downstream numerical
#' use is intended.
#'
#' @param partitions A list of data frames, typically produced by
#'   \code{\link{partition_time_windows}}. Each data frame must contain
#'   integer-encoded amino acid sequences as columns (values 1 to
#'   \code{alphabet_size}).
#' @param site_index Integer. The column index (site) to analyse.
#' @param labels Character vector. Column labels for each partition.
#'   Defaults to \code{names(partitions)}, or \code{"P1"}, \code{"P2"},
#'   ... if unnamed.
#' @param alphabet_size Integer. Total number of possible amino acid
#'   codes. Must match the encoding used during integer encoding
#'   (\code{\link{encode_aa_sequence}} produces values 1-25 by default).
#'   Default is \code{25L}.
#' @param zeros Logical. If \code{TRUE} (default), fills missing counts
#'   with \code{0}. If \code{FALSE}, replaces zeros with \code{""}; see
#'   Details for the type-coercion implication.
#' @param use_letters Logical. If \code{TRUE} (default), uses decoded
#'   single-character codes from the 25-symbol alphabet (A, R, N, D,
#'   C, ..., V, B, Z, X, *, -) as row names. If \code{FALSE}, uses
#'   numeric codes 1 to \code{alphabet_size}.
#' @param relative Logical. If \code{TRUE}, converts counts to
#'   proportions (column-wise division by partition size). Default is
#'   \code{FALSE}.
#' @param digits Integer. Decimal places for rounding when
#'   \code{relative = TRUE}. Default is \code{2L}.
#' @param col_width Character. CSS width string applied to all columns
#'   (e.g. \code{"100px"}). Default is \code{"100px"}.
#' @param highlight_col Integer or \code{NULL}. 1-based column index
#'   (relative to the data columns, not counting the row-name column)
#'   of a partition to highlight with \code{background}. Out-of-range
#'   values trigger a warning and no highlight is applied. \code{NULL}
#'   (default) means no highlight.
#' @param background Character. CSS background colour for the
#'   highlighted column. Default is \code{"#f0f8ff"} (light blue).
#' @param wrap_length Integer. Character width at which to wrap long
#'   column labels using HTML line breaks. Default is \code{10L}.
#' @param save Logical. If \code{TRUE}, saves the rendered HTML table
#'   to disk via \code{\link[kableExtra]{save_kable}}. Default is
#'   \code{FALSE}.
#' @param save_extension Character. File extension for the saved file
#'   (including leading dot). Default is \code{".html"}.
#' @param save_path Character or \code{NULL}.  Directory in which to save the
#'   file.  ...  Must be supplied when \code{save = TRUE}.  Default is
#'   \code{NULL}.
#' @param return_table Logical. If \code{TRUE} (default), returns a
#'   named list with both the raw data frame and the styled kable
#'   object. If \code{FALSE}, returns only the styled kable.
#'
#' @return If \code{return_table = TRUE}, a named list:
#'   \item{table}{The raw count (or proportion) data frame, with row
#'     names corresponding to amino acid codes and column names
#'     corresponding to partition labels.}
#'   \item{styled}{The \pkg{kableExtra} HTML kable object.}
#'   If \code{return_table = FALSE}, returns only the styled kable
#'   object.
#'
#' @seealso \code{\link{partition_time_windows}} for producing the
#'   typical input list of partitions; \code{\link{get_site_counts}}
#'   for the count-tabulation primitive; \code{\link{decode_aa_sequence}}
#'   for the alphabet code mapping; \code{\link{calculate_hellinger_matrix}}
#'   for the related cross-partition distance calculation on the same
#'   data shape.
#'
#' @importFrom kableExtra kable kable_styling column_spec save_kable
#' @importFrom magrittr %>%
#' @export
#'
#' @examples
#' p1 = data.frame(s1 = c(1L, 1L, 1L, 1L, 2L))
#' p2 = data.frame(s1 = c(1L, 1L, 2L, 2L, 2L))
#' parts = list(T1 = p1, T2 = p2)
#'
#' # Default: counts, letters, no save
#' tbl = tabulate_site_evolution(parts, site_index = 1)
#' tbl$table
#'
#' # Relative frequencies, highlight second partition
#' tbl2 = tabulate_site_evolution(parts, site_index = 1,
#'                                 relative = TRUE, highlight_col = 2)
#' tbl2$table
#'
#' # Numeric codes (skip alphabet decoding)
#' tbl3 = tabulate_site_evolution(parts, site_index = 1, use_letters = FALSE)
#' rownames(tbl3$table)
tabulate_site_evolution = function(partitions,
                                   site_index,
                                   labels         = NULL,
                                   alphabet_size  = 25L,
                                   zeros          = TRUE,
                                   use_letters    = TRUE,
                                   relative       = FALSE,
                                   digits         = 2L,
                                   col_width      = "100px",
                                   highlight_col  = NULL,
                                   background     = "#f0f8ff",
                                   wrap_length    = 10L,
                                   save           = FALSE,
                                   save_extension = ".html",
                                   save_path      = NULL,
                                   return_table   = TRUE) {
  
  # --- 1. Input validation ---------------------------------------------------
  if (!is.list(partitions) || length(partitions) == 0L)
    stop("`partitions` must be a non-empty list of data frames.",
         call. = FALSE)
  
  if (!is.numeric(site_index) || length(site_index) != 1L || site_index < 1L)
    stop("`site_index` must be a single positive integer.",
         call. = FALSE)
  
  if (!is.numeric(alphabet_size) || alphabet_size < 1L)
    stop("`alphabet_size` must be a positive integer.",
         call. = FALSE)
  
  n_part = length(partitions)
  
  # --- 2. Column labels ------------------------------------------------------
  if (is.null(labels)) {
    p_names = names(partitions)
    labels  = if (!is.null(p_names)) p_names else paste0("P", seq_len(n_part))
  }
  
  # --- 3. Pre-compute row names ----------------------------------------------
  # Built once and reused; avoids the duplicate rowname assignment that
  # the original code performed inside the relative-frequency branch.
  row_labels = if (isTRUE(use_letters)) {
    as.vector(decode_aa_sequence(matrix(seq_len(alphabet_size), ncol = 1L)))
  } else {
    as.character(seq_len(alphabet_size))
  }
  
  # --- 4. Counts via get_site_counts -----------------------------------------
  counts_matrix = get_site_counts(partitions, site_index, alphabet_size)
  
  # --- 5. Optional conversion to relative frequencies ------------------------
  if (isTRUE(relative)) {
    col_sums = colSums(counts_matrix)
    col_sums[col_sums == 0] = 1L              # guard against empty partitions
    prot_df = as.data.frame(
      round(sweep(counts_matrix, 2L, col_sums, "/"), digits)
    )
  } else {
    prot_df = as.data.frame(counts_matrix)
  }
  
  rownames(prot_df) = row_labels
  
  # --- 6. Optional zero -> empty-string conversion ---------------------------
  # Note: this forces the data frame to character storage. See @details.
  if (!isTRUE(zeros)) {
    prot_df[prot_df == 0] = ""
  }
  
  # --- 7. Wrap column labels for HTML ----------------------------------------
  wrap_text = function(text, width = wrap_length) {
    paste(strwrap(text, width = width), collapse = "<br>")
  }
  formatted_labels = vapply(labels, wrap_text, character(1L))
  
  # --- 8. Build kable --------------------------------------------------------
  styled = prot_df %>%
    kableExtra::kable("html",
                      row.names = TRUE,
                      col.names = formatted_labels,
                      escape    = FALSE) %>%
    kableExtra::kable_styling("striped", full_width = FALSE) %>%
    kableExtra::column_spec(seq_len(n_part), width = col_width)
  
  # --- 9. Optional column highlight ------------------------------------------
  if (!is.null(highlight_col)) {
    if (highlight_col < 1L || highlight_col > n_part) {
      warning(sprintf(
        "`highlight_col` (%d) is out of range [1, %d]; no highlight applied.",
        highlight_col, n_part), call. = FALSE)
    } else {
      styled = styled %>%
        kableExtra::column_spec(highlight_col + 1L,   # +1: row-name column comes first
                                background = background)
    }
  }
  
  # --- 10. Optional save -----------------------------------------------------
  if (isTRUE(save)) {
    if (is.null(save_path))
      stop("`save_path` must be supplied when `save = TRUE`.", call. = FALSE)
    file_path = file.path(save_path,
                          paste0("Site_", site_index, save_extension))
    styled %>% kableExtra::save_kable(file_path)
  }
  
  # --- 11. Return ------------------------------------------------------------
  if (isTRUE(return_table)) {
    list(table  = prot_df,
         styled = styled)
  } else {
    styled
  }
}

#' @title Get Site Counts Matrix (Internal)
#' @description Extract and tabulate amino acid counts for a specific site
#'   across all partitions.
#'
#' @details Built on \code{\link[base]{tabulate}} for fast integer counting.
#'   Consumed internally by \code{\link{calculate_hellinger_matrix}}.
#'
#' @param partitions List of data frames, one per time window. Each must
#'   have integer-encoded amino acid columns.
#' @param site_index Integer. Index of the site (column) to count.
#' @param alphabet_size Integer. Size of the amino acid alphabet
#'   (typically \code{25} for the package's default encoding).
#' @return Integer matrix with \code{alphabet_size} rows (one per amino
#'   acid code, 1 through \code{alphabet_size}) and one column per
#'   partition. Each cell is the count of that amino acid at
#'   \code{site_index} in that partition.
#' @keywords internal
get_site_counts = function(partitions, site_index, alphabet_size) {
  site_vecs = lapply(partitions, `[[`, site_index)
  vapply(site_vecs, tabulate, integer(alphabet_size), nbins = alphabet_size)
}

#' @title Convert FASTA Object to Character Matrix
#' @description Converts an \code{AAStringSet} object loaded via
#'   \code{\link[Biostrings]{readAAStringSet}} into a character matrix where
#'   rows are sequences and columns are residue positions (sites). Inverse
#'   structural transformation of \code{\link{encode_aa_sequence}}'s expected
#'   input shape.
#'
#' @details
#' \strong{Alignment.} The function expects an aligned \code{AAStringSet}
#' — all sequences of equal width. Unaligned input is accepted and shorter
#' sequences are right-padded with the gap character \code{"-"} to match
#' the longest sequence, but downstream entropy-based analysis assumes
#' positional homology across rows; if sequences in the input are not
#' biologically aligned, results from per-site computations will not be
#' meaningful. For unaligned input, run a multiple-sequence alignment
#' (e.g. \code{msa::msa()} or \code{DECIPHER::AlignSeqs()}) before calling
#' this function.
#'
#' \strong{Performance.} Conversion is fully vectorised: all sequences are
#' coerced to a single character string vector, split simultaneously, and
#' reshaped into a matrix in one operation. No per-row loop, no
#' intermediate list of split sequences kept alive — substantially faster
#' than per-row \code{strsplit} on large inputs (100k+ sequences).
#'
#' @param fsta An \code{AAStringSet} object, typically the output of
#'   \code{\link[Biostrings]{readAAStringSet}}. May be aligned or
#'   unaligned (see Details).
#'
#' @return A character matrix with \code{length(fsta)} rows and
#'   \code{max(nchar(as.character(fsta)))} columns. Each cell contains a
#'   single-character amino acid code from the input sequences (or the
#'   gap character \code{"-"} for padded positions in unaligned input).
#'   The matrix has no row or column names; sequence names from the
#'   \code{AAStringSet} are not carried over. An empty input
#'   (\code{length(fsta) == 0}) returns an empty 0-by-0 character matrix.
#'
#' @seealso \code{\link{encode_aa_sequence}} for converting the resulting
#'   character matrix to an integer-encoded matrix;
#'   \code{\link{filter_ambiguous_sequences}} for removing rows containing
#'   ambiguous residues; \code{\link[Biostrings]{readAAStringSet}} for
#'   loading FASTA files into the input format.
#'
#' @export
#'
#' @examples
#' \donttest{
#' # Convert the bundled sample to a character matrix.
#' path  <- system.file("extdata", "sarscov2_sample.fasta.gz",
#'                       package = "ViralEntropR")
#' fasta <- Biostrings::readAAStringSet(path)
#' mat   <- fasta_to_char_matrix(fasta)
#' dim(mat)
#' mat[1:3, 1:10]
#' }
fasta_to_char_matrix <- function(fsta) {
  
  if (!requireNamespace("Biostrings", quietly = TRUE)) {
    stop("Package 'Biostrings' is required. Install it with:\n",
         "  install.packages('BiocManager')\n",
         "  BiocManager::install('Biostrings')",
         call. = FALSE)
  }
  
  n <- length(fsta)
  
  # Empty input: return empty matrix, schema-stable for downstream consumers.
  if (n == 0L) {
    return(matrix(character(0), nrow = 0L, ncol = 0L))
  }
  
  seqs     <- as.character(fsta)          # named character vector, one string per seq
  widths   <- nchar(seqs)
  m        <- max(widths)

  # For aligned inputs all widths are equal and the reshape is exact.
  # For unaligned inputs, pad shorter sequences with "-" so every string
  # reaches width m before splitting — this preserves the original behaviour
  # of filling the pre-allocated matrix (which used 0; here "-" is more
  # meaningful for amino acid data).
  if (any(widths < m)) {
    seqs <- formatC(seqs, width = -m, flag = "-")   # left-align, pad right with spaces
    # replace padding spaces with gap character
    seqs <- gsub(" ", "-", seqs, fixed = TRUE)
  }

  # Split all sequences at once: strsplit returns a list of length n, each
  # element a character vector of length m. unlist + matrix reshapes in one
  # step — no per-row loop, no intermediate list kept alive.
  seq_matrix <- matrix(
    unlist(strsplit(seqs, "", fixed = TRUE), use.names = FALSE),
    nrow  = n,
    ncol  = m,
    byrow = TRUE
  )

  seq_matrix
}

#' @title Plot GMM Entropy Class Trajectory for a Single Site
#'
#' @description Plots the Shannon entropy trajectory for a single sequence site
#'   across time partitions, with geom-label overlays at each partition
#'   recording the site's current GMM class (above the line, in green) and the
#'   partition's highest-entropy class label (below the line, in red).  The
#'   visual contrast between the two labels tracks the moment a site enters the
#'   highest-entropy cluster — the entropy-based analogue of variant emergence
#'   detection.
#'
#' @details
#' The function operates on the \code{$Data_Frame} element returned by
#' \code{\link{plot_entropy_trajectories}}, which contains columns
#' \code{sites}, \code{entropies}, \code{class}, \code{max_class},
#' \code{period}, and \code{coverage}.
#'
#' \strong{Class label interpretation.}  The green label (upper) records the
#' GMM class assigned to the site in that partition; the red label (lower)
#' records \code{max_class} — the label of the highest-entropy component for
#' that partition.  When the two labels are equal the site has entered the
#' highest-entropy class.  Without relabeling, \code{max_class} is the raw
#' Mclust label carrying the highest mean entropy (i.e.
#' \code{max(classification)} at clustering time).  When the user has
#' pre-relabeled the partitions via \code{\link{relabel_entropy_classes}}
#' before calling \code{\link{plot_entropy_trajectories}}, \code{max_class}
#' is \code{1L} throughout (class 1 is the highest-entropy group by
#' definition), and the sentinel \code{999L} is preserved unchanged.
#'
#' \strong{Axis padding.}  Both axes carry generous expansion margins so that
#' geom-label boxes at extreme entropy values or at the first and last
#' partitions do not clip the panel border.  The returned \code{ggplot} object
#' can be further adjusted by appending standard \code{ggplot2} layers with
#' \code{+} before printing or saving.
#'
#' @param data_frame Data frame.  The \code{$Data_Frame} element from
#'   \code{\link{plot_entropy_trajectories}}.  Must contain columns
#'   \code{sites}, \code{entropies}, \code{class}, \code{max_class},
#'   \code{period}, and \code{coverage}.
#' @param site Integer (length 1).  The site index to plot.  Must be present
#'   in \code{data_frame$sites}.
#' @param site_color Character.  Colour of the entropy trajectory line.
#'   Default is \code{"steelblue"}.  Pass
#'   \code{plot_entropy_trajectories()$Colors[as.character(site)]} for
#'   cross-plot colour consistency.
#' @param xbreaks Integer vector.  Partition period indices for x-axis breaks.
#'   Typically \code{plot_entropy_trajectories()$XBreaks}.
#' @param xlabels Character vector.  Partition label strings aligned with
#'   \code{xbreaks}.  Typically \code{plot_entropy_trajectories()$XLabels}.
#' @param col_current Character.  Fill colour of the upper geom-label
#'   (current GMM class at each partition).  Default is \code{"springgreen2"}.
#' @param col_max_class Character.  Fill colour of the lower geom-label
#'   (highest-entropy class label at each partition).  Default is
#'   \code{"red2"}.
#' @param label_size Numeric.  Font size of the class integer text inside the
#'   geom-labels.  Default is \code{3}.
#' @param x_angle Numeric.  Rotation angle of x-axis tick labels in degrees.
#'   Default is \code{45}.
#' @param line_size Numeric.  Width of the entropy trajectory line.  Default
#'   is \code{1.5}.
#' @param plot_title Character or \code{NULL}.  Plot title.  If \code{NULL}
#'   (default), the title is auto-generated as
#'   \code{"GMM Entropy Class Trajectory \u2014 Site <site>"}.
#' @param save Logical.  If \code{TRUE}, the plot is saved to disk via
#'   \code{\link[ggplot2]{ggsave}}.  Default is \code{FALSE}.
#' @param save_path Character or \code{NULL}.  Directory in which to save the
#'   file.  Created recursively if it does not exist.  Must be supplied when
#'   \code{save = TRUE}.  Default is \code{NULL}.
#' @param save_extension Character.  File extension including the leading dot
#'   (e.g. \code{".jpeg"}, \code{".pdf"}, \code{".png"}).  Default is
#'   \code{".jpeg"}.
#' @param width Numeric.  Saved figure width in inches.  Default is \code{20}.
#' @param height Numeric.  Saved figure height in inches.  Default is
#'   \code{15}.
#' @param dpi Numeric.  Resolution of the saved raster output in dots per
#'   inch.  Default is \code{600}.
#'
#' @return A \code{ggplot} object (returned invisibly).  Additional
#'   \code{ggplot2} layers can be appended with \code{+} before printing or
#'   saving, for example:
#'   \preformatted{
#' p <- plot_site_class_trajectory(traj$Data_Frame, site = 681L, ...)
#' p + ggplot2::geom_vline(xintercept = 14, colour = "darkorange",
#'                          linetype = "dashed", linewidth = 1.5)
#'   }
#'
#' @seealso
#' \code{\link{plot_entropy_trajectories}},
#' \code{\link{relabel_entropy_classes}},
#' \code{\link{partition_time_windows}}
#'
#' @importFrom ggplot2 ggplot aes geom_line geom_label scale_x_continuous
#'   scale_y_continuous expansion labs theme_bw theme element_text unit ggsave
#' @importFrom rlang .data
#' @export
#'
#' @examples
#' # Three-period synthetic dataset (independent of plot_entropy_trajectories'
#' # example). Three sites with different entropy trajectories; site 1 features
#' # a monotone-up trajectory and progressively higher rank among sites,
#' # producing class changes across all three partitions.
#' #
#' # Site 1 follows a monotone-up trajectory (0.469 -> 0.881 -> 1.522), 
#' # climbing from lowest rank in P1 to highest in P3 for a clear 
#' # "emerging variant" class progression. Site 2 follows the
#' # opposite trajectory (1.522 -> 1.000 -> 0.881), starting as the dominant
#' # high-entropy site and declining as site 1 rises. Site 3 (featured) 
#' # peaks at P2 (0.881 -> 1.522 -> 1.000).
#' #
#' # Per-partition Shannon entropy (bits):
#' #          P1     P2     P3
#' #   s1   0.469  0.881  1.522   
#' #   s2   1.522  1.000  0.881
#' #   s3   0.881  1.522  1.000   <- featured site
#' df_cls <- data.frame(
#'   s1 = c(
#'     c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L),       # P1: 9:1
#'     c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L),       # P2: 7:3
#'     c(1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 3L, 3L)        # P3: 4:4:2
#'   ),
#'   s2 = c(
#'     c(1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 3L, 3L),       # P1: 4:4:2
#'     c(1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 2L),       # P2: 5:5
#'     c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L)        # P3: 7:3
#'   ),
#'   s3 = c(
#'     c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L),       # P1: 7:3
#'     c(1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 3L, 3L),       # P2: 4:4:2
#'     c(1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 2L)        # P3: 5:5
#'   ),
#'   Date = rep(
#'     seq(as.Date("2020-01-01"), by = "month", length.out = 3L),
#'     each = 10L
#'   )
#' )
#'
#' part_data <- partition_time_windows(
#'   data          = df_cls,
#'   n_sites       = 3L,
#'   window_length = 1L,
#'   window_type   = 3L,
#'   start_date    = "2020-01-01",
#'   end_date      = "2020-04-01",
#'   removez       = FALSE,
#'   removesngl    = FALSE
#' )
#'
#' # Apply class relabeling so class 1 = highest mean entropy.
#' # plot_entropy_trajectories does not relabel internally; the consumer does.
#' for (i in seq_along(part_data$Clusters)) {
#'   part_data$Clusters[[i]]$DataFrame <-
#'     relabel_entropy_classes(part_data$Clusters[[i]]$DataFrame)
#' }
#'
#' traj <- plot_entropy_trajectories(part_data)
#'
#' p <- plot_site_class_trajectory(
#'   data_frame = traj$Data_Frame,
#'   site       = 3L,
#'   site_color = traj$Colors["1"],
#'   xbreaks    = traj$XBreaks,
#'   xlabels    = traj$XLabels
#' )
#' print(p)
plot_site_class_trajectory <- function(data_frame,
                                       site,
                                       site_color     = "steelblue",
                                       xbreaks,
                                       xlabels,
                                       col_current    = "springgreen2",
                                       col_max_class  = "red2",
                                       label_size     = 3,
                                       x_angle        = 45,
                                       line_size      = 1.5,
                                       plot_title     = NULL,
                                       save           = FALSE,
                                       save_path      = NULL,
                                       save_extension = ".png",
                                       width          = 20,
                                       height         = 15,
                                       dpi            = 300) {

  # ---------------------------------------------------------------------------
  # 1. Input validation
  # ---------------------------------------------------------------------------
  required_cols <- c("sites", "entropies", "class", "max_class",
                     "period", "coverage")
  missing_cols  <- setdiff(required_cols, names(data_frame))
  if (length(missing_cols) > 0L)
    stop("`data_frame` is missing required columns: ",
         paste(missing_cols, collapse = ", "),
         ". Ensure `data_frame` is the `$Data_Frame` element returned by ",
         "plot_entropy_trajectories().", call. = FALSE)

  if (length(site) != 1L || !is.numeric(site))
    stop("`site` must be a single numeric site index.", call. = FALSE)
  site <- as.integer(site)

  # Subset to the requested site.
  # data_frame$sites is a factor (from plot_entropy_trajectories): convert
  # factor labels to integer for comparison with site.
  df_site <- data_frame[as.integer(as.character(data_frame$sites)) == site, ,
                        drop = FALSE]

  if (nrow(df_site) == 0L)
    stop("Site ", site, " is not present in `data_frame`. ",
         "Verify the site index and ensure it was included when calling ",
         "plot_entropy_trajectories().", call. = FALSE)

  # Coerce class and max_class to integer for unambiguous label display.
  # data_frame$class is a factor; as.character() recovers the actual label
  # string before as.integer() converts it — avoids factor-code confusion.
  df_site$class     <- as.integer(as.character(df_site$class))
  df_site$max_class <- as.integer(df_site$max_class)
  df_site$period    <- as.integer(df_site$period)

  # ---------------------------------------------------------------------------
  # 2. Validate xbreaks / xlabels
  # ---------------------------------------------------------------------------
  if (missing(xbreaks) || is.null(xbreaks))
    stop("`xbreaks` is required. Pass `plot_entropy_trajectories()$XBreaks`.", 
         call. = FALSE)
  if (missing(xlabels) || is.null(xlabels))
    stop("`xlabels` is required. Pass `plot_entropy_trajectories()$XLabels`.", 
         call. = FALSE)
  if (length(xbreaks) != length(xlabels))
    stop("`xbreaks` and `xlabels` must have equal length.", call. = FALSE)

  # ---------------------------------------------------------------------------
  # 3. Plot title
  # ---------------------------------------------------------------------------
  if (is.null(plot_title))
    plot_title <- paste0("GMM Entropy Class Trajectory \u2014 Site ", site)

  # ---------------------------------------------------------------------------
  # 4. Build ggplot
  # ---------------------------------------------------------------------------
  # The .data pronoun from rlang is used throughout aes() to make column
  # references unambiguous to R CMD check's static analyser.
  p <- ggplot2::ggplot(df_site,
                       ggplot2::aes(x = .data$period,
                                    y = .data$entropies)) +

    # Entropy trajectory line
    ggplot2::geom_line(colour    = site_color,
                       linewidth = line_size) +

    # Upper label: current GMM class at each partition (green, above line)
    ggplot2::geom_label(
      ggplot2::aes(label = .data$class),
      hjust         = 0.5,
      vjust         = -0.5,
      fill          = col_current,
      size          = label_size,
      fontface      = "bold",
      label.padding = ggplot2::unit(1, "lines")
    ) +

    # Lower label: highest-entropy class at each partition (red, below line)
    ggplot2::geom_label(
      ggplot2::aes(label = .data$max_class),
      hjust         = 0.5,
      vjust         = 0.5,
      fill          = col_max_class,
      size          = label_size,
      fontface      = "bold",
      label.padding = ggplot2::unit(1, "lines")
    ) +

    # Axis scales: generous expansion on both axes prevents geom-label boxes
    # from clipping the panel border at extreme entropy values or at the first
    # and last partitions.
    ggplot2::scale_x_continuous(
      breaks = xbreaks,
      labels = xlabels,
      expand = ggplot2::expansion(add = 0.6)
    ) +
    ggplot2::scale_y_continuous(
      name   = "Shannon Entropy",
      expand = ggplot2::expansion(mult = c(0.4, 0.4))
    ) +

    ggplot2::labs(
      title = plot_title,
      x     = "Partition"
    ) +

    ggplot2::theme_bw() +
    ggplot2::theme(
      plot.title  = ggplot2::element_text(hjust = 0.5),
      axis.text.x = ggplot2::element_text(angle = x_angle, hjust = 1)
    )

  # ---------------------------------------------------------------------------
  # 5. Optional save
  # ---------------------------------------------------------------------------
  if (isTRUE(save)) {
    if (is.null(save_path))
      stop("`save_path` must be supplied when `save = TRUE`.", call. = FALSE)
    if (!dir.exists(save_path))
      dir.create(save_path, recursive = TRUE, showWarnings = FALSE)
    file_name <- paste0("Site_", site, "_class_trajectory", save_extension)
    ggplot2::ggsave(
      filename = file.path(save_path, file_name),
      plot     = p,
      width    = width,
      height   = height,
      dpi      = dpi
    )
  }

  # ---------------------------------------------------------------------------
  # 6. Return
  # ---------------------------------------------------------------------------
  invisible(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_site_class_trajectory.R
\name{plot_site_class_trajectory}
\alias{plot_site_class_trajectory}
\title{Plot GMM Entropy Class Trajectory for a Single Site}
\usage{
plot_site_class_trajectory(
  data_frame,
  site,
  site_color = "steelblue",
  xbreaks,
  xlabels,
  col_current = "springgreen2",
  col_max_class = "red2",
  label_size = 3,
  x_angle = 45,
  line_size = 1.5,
  plot_title = NULL,
  save = FALSE,
  save_path = NULL,
  save_extension = ".png",
  width = 20,
  height = 15,
  dpi = 300
)
}
\arguments{
\item{data_frame}{Data frame.  The \code{$Data_Frame} element from
\code{\link{plot_entropy_trajectories}}.  Must contain columns
\code{sites}, \code{entropies}, \code{class}, \code{max_class},
\code{period}, and \code{coverage}.}

\item{site}{Integer (length 1).  The site index to plot.  Must be present
in \code{data_frame$sites}.}

\item{site_color}{Character.  Colour of the entropy trajectory line.
Default is \code{"steelblue"}.  Pass
\code{plot_entropy_trajectories()$Colors[as.character(site)]} for
cross-plot colour consistency.}

\item{xbreaks}{Integer vector.  Partition period indices for x-axis breaks.
Typically \code{plot_entropy_trajectories()$XBreaks}.}

\item{xlabels}{Character vector.  Partition label strings aligned with
\code{xbreaks}.  Typically \code{plot_entropy_trajectories()$XLabels}.}

\item{col_current}{Character.  Fill colour of the upper geom-label
(current GMM class at each partition).  Default is \code{"springgreen2"}.}

\item{col_max_class}{Character.  Fill colour of the lower geom-label
(highest-entropy class label at each partition).  Default is
\code{"red2"}.}

\item{label_size}{Numeric.  Font size of the class integer text inside the
geom-labels.  Default is \code{3}.}

\item{x_angle}{Numeric.  Rotation angle of x-axis tick labels in degrees.
Default is \code{45}.}

\item{line_size}{Numeric.  Width of the entropy trajectory line.  Default
is \code{1.5}.}

\item{plot_title}{Character or \code{NULL}.  Plot title.  If \code{NULL}
(default), the title is auto-generated as
\code{"GMM Entropy Class Trajectory \u2014 Site <site>"}.}

\item{save}{Logical.  If \code{TRUE}, the plot is saved to disk via
\code{\link[ggplot2]{ggsave}}.  Default is \code{FALSE}.}

\item{save_path}{Character or \code{NULL}.  Directory in which to save the
file.  Created recursively if it does not exist.  Must be supplied when
\code{save = TRUE}.  Default is \code{NULL}.}

\item{save_extension}{Character.  File extension including the leading dot
(e.g. \code{".jpeg"}, \code{".pdf"}, \code{".png"}).  Default is
\code{".jpeg"}.}

\item{width}{Numeric.  Saved figure width in inches.  Default is \code{20}.}

\item{height}{Numeric.  Saved figure height in inches.  Default is
\code{15}.}

\item{dpi}{Numeric.  Resolution of the saved raster output in dots per
inch.  Default is \code{600}.}
}
\value{
A \code{ggplot} object (returned invisibly).  Additional
\code{ggplot2} layers can be appended with \code{+} before printing or
saving, for example:
\preformatted{
p <- plot_site_class_trajectory(traj$Data_Frame, site = 681L, ...)
p + ggplot2::geom_vline(xintercept = 14, colour = "darkorange",
                         linetype = "dashed", linewidth = 1.5)
  }
}
\description{
Plots the Shannon entropy trajectory for a single sequence site
across time partitions, with geom-label overlays at each partition
recording the site's current GMM class (above the line, in green) and the
partition's highest-entropy class label (below the line, in red).  The
visual contrast between the two labels tracks the moment a site enters the
highest-entropy cluster — the entropy-based analogue of variant emergence
detection.
}
\details{
The function operates on the \code{$Data_Frame} element returned by
\code{\link{plot_entropy_trajectories}}, which contains columns
\code{sites}, \code{entropies}, \code{class}, \code{max_class},
\code{period}, and \code{coverage}.

\strong{Class label interpretation.}  The green label (upper) records the
GMM class assigned to the site in that partition; the red label (lower)
records \code{max_class} — the label of the highest-entropy component for
that partition.  When the two labels are equal the site has entered the
highest-entropy class.  Without relabeling, \code{max_class} is the raw
Mclust label carrying the highest mean entropy (i.e.
\code{max(classification)} at clustering time).  When the user has
pre-relabeled the partitions via \code{\link{relabel_entropy_classes}}
before calling \code{\link{plot_entropy_trajectories}}, \code{max_class}
is \code{1L} throughout (class 1 is the highest-entropy group by
definition), and the sentinel \code{999L} is preserved unchanged.

\strong{Axis padding.}  Both axes carry generous expansion margins so that
geom-label boxes at extreme entropy values or at the first and last
partitions do not clip the panel border.  The returned \code{ggplot} object
can be further adjusted by appending standard \code{ggplot2} layers with
\code{+} before printing or saving.
}
\examples{
# Three-period synthetic dataset (independent of plot_entropy_trajectories'
# example). Three sites with different entropy trajectories; site 1 features
# a monotone-up trajectory and progressively higher rank among sites,
# producing class changes across all three partitions.
#
# Site 1 follows a monotone-up trajectory (0.469 -> 0.881 -> 1.522), 
# climbing from lowest rank in P1 to highest in P3 for a clear 
# "emerging variant" class progression. Site 2 follows the
# opposite trajectory (1.522 -> 1.000 -> 0.881), starting as the dominant
# high-entropy site and declining as site 1 rises. Site 3 (featured) 
# peaks at P2 (0.881 -> 1.522 -> 1.000).
#
# Per-partition Shannon entropy (bits):
#          P1     P2     P3
#   s1   0.469  0.881  1.522   
#   s2   1.522  1.000  0.881
#   s3   0.881  1.522  1.000   <- featured site
df_cls <- data.frame(
  s1 = c(
    c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L),       # P1: 9:1
    c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L),       # P2: 7:3
    c(1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 3L, 3L)        # P3: 4:4:2
  ),
  s2 = c(
    c(1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 3L, 3L),       # P1: 4:4:2
    c(1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 2L),       # P2: 5:5
    c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L)        # P3: 7:3
  ),
  s3 = c(
    c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L),       # P1: 7:3
    c(1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 3L, 3L),       # P2: 4:4:2
    c(1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 2L)        # P3: 5:5
  ),
  Date = rep(
    seq(as.Date("2020-01-01"), by = "month", length.out = 3L),
    each = 10L
  )
)

part_data <- partition_time_windows(
  data          = df_cls,
  n_sites       = 3L,
  window_length = 1L,
  window_type   = 3L,
  start_date    = "2020-01-01",
  end_date      = "2020-04-01",
  removez       = FALSE,
  removesngl    = FALSE
)

# Apply class relabeling so class 1 = highest mean entropy.
# plot_entropy_trajectories does not relabel internally; the consumer does.
for (i in seq_along(part_data$Clusters)) {
  part_data$Clusters[[i]]$DataFrame <-
    relabel_entropy_classes(part_data$Clusters[[i]]$DataFrame)
}

traj <- plot_entropy_trajectories(part_data)

p <- plot_site_class_trajectory(
  data_frame = traj$Data_Frame,
  site       = 3L,
  site_color = traj$Colors["1"],
  xbreaks    = traj$XBreaks,
  xlabels    = traj$XLabels
)
print(p)
}
\seealso{
\code{\link{plot_entropy_trajectories}},
\code{\link{relabel_entropy_classes}},
\code{\link{partition_time_windows}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tabulate_site_evolution.R
\name{tabulate_site_evolution}
\alias{tabulate_site_evolution}
\title{Tabulate Site Frequency Evolution}
\usage{
tabulate_site_evolution(
  partitions,
  site_index,
  labels = NULL,
  alphabet_size = 25L,
  zeros = TRUE,
  use_letters = TRUE,
  relative = FALSE,
  digits = 2L,
  col_width = "100px",
  highlight_col = NULL,
  background = "#f0f8ff",
  wrap_length = 10L,
  save = FALSE,
  save_extension = ".html",
  save_path = NULL,
  return_table = TRUE
)
}
\arguments{
\item{partitions}{A list of data frames, typically produced by
\code{\link{partition_time_windows}}. Each data frame must contain
integer-encoded amino acid sequences as columns (values 1 to
\code{alphabet_size}).}

\item{site_index}{Integer. The column index (site) to analyse.}

\item{labels}{Character vector. Column labels for each partition.
Defaults to \code{names(partitions)}, or \code{"P1"}, \code{"P2"},
... if unnamed.}

\item{alphabet_size}{Integer. Total number of possible amino acid
codes. Must match the encoding used during integer encoding
(\code{\link{encode_aa_sequence}} produces values 1-25 by default).
Default is \code{25L}.}

\item{zeros}{Logical. If \code{TRUE} (default), fills missing counts
with \code{0}. If \code{FALSE}, replaces zeros with \code{""}; see
Details for the type-coercion implication.}

\item{use_letters}{Logical. If \code{TRUE} (default), uses decoded
single-character codes from the 25-symbol alphabet (A, R, N, D,
C, ..., V, B, Z, X, *, -) as row names. If \code{FALSE}, uses
numeric codes 1 to \code{alphabet_size}.}

\item{relative}{Logical. If \code{TRUE}, converts counts to
proportions (column-wise division by partition size). Default is
\code{FALSE}.}

\item{digits}{Integer. Decimal places for rounding when
\code{relative = TRUE}. Default is \code{2L}.}

\item{col_width}{Character. CSS width string applied to all columns
(e.g. \code{"100px"}). Default is \code{"100px"}.}

\item{highlight_col}{Integer or \code{NULL}. 1-based column index
(relative to the data columns, not counting the row-name column)
of a partition to highlight with \code{background}. Out-of-range
values trigger a warning and no highlight is applied. \code{NULL}
(default) means no highlight.}

\item{background}{Character. CSS background colour for the
highlighted column. Default is \code{"#f0f8ff"} (light blue).}

\item{wrap_length}{Integer. Character width at which to wrap long
column labels using HTML line breaks. Default is \code{10L}.}

\item{save}{Logical. If \code{TRUE}, saves the rendered HTML table
to disk via \code{\link[kableExtra]{save_kable}}. Default is
\code{FALSE}.}

\item{save_extension}{Character. File extension for the saved file
(including leading dot). Default is \code{".html"}.}

\item{save_path}{Character or \code{NULL}.  Directory in which to save the
file.  ...  Must be supplied when \code{save = TRUE}.  Default is
\code{NULL}.}

\item{return_table}{Logical. If \code{TRUE} (default), returns a
named list with both the raw data frame and the styled kable
object. If \code{FALSE}, returns only the styled kable.}
}
\value{
If \code{return_table = TRUE}, a named list:
\item{table}{The raw count (or proportion) data frame, with row
names corresponding to amino acid codes and column names
corresponding to partition labels.}
\item{styled}{The \pkg{kableExtra} HTML kable object.}
If \code{return_table = FALSE}, returns only the styled kable
object.
}
\description{
Generates a frequency or proportion table showing the amino
acid distribution at a specific site across multiple time partitions,
optionally styled with \pkg{kableExtra} and saved as standalone HTML.
}
\details{
Aggregates amino acid counts per partition using
\code{\link{get_site_counts}}, optionally converts to relative
frequencies, applies kableExtra styling (column width, column
highlighting, striped rows), and optionally saves to disk. Row names
are decoded amino acid codes via \code{\link{decode_aa_sequence}} when
\code{use_letters = TRUE}.

\strong{Empty partitions.} Partitions containing no observations
contribute all-zero columns. When \code{relative = TRUE}, division by
zero is avoided by treating empty-column sums as 1, leaving
proportions at zero. Inspect partition sizes via
\code{sapply(partitions, nrow)} before interpreting the table.

\strong{Note on \code{zeros = FALSE}.} Setting \code{zeros = FALSE}
replaces numeric zeros with empty strings (\code{""}) for visual
clarity in the kable. This conversion forces the underlying data
frame to character storage; numeric operations (\code{sum},
\code{mean}, etc.) will not work on the returned \code{table}
element. Use \code{zeros = TRUE} (default) if downstream numerical
use is intended.
}
\examples{
p1 = data.frame(s1 = c(1L, 1L, 1L, 1L, 2L))
p2 = data.frame(s1 = c(1L, 1L, 2L, 2L, 2L))
parts = list(T1 = p1, T2 = p2)

# Default: counts, letters, no save
tbl = tabulate_site_evolution(parts, site_index = 1)
tbl$table

# Relative frequencies, highlight second partition
tbl2 = tabulate_site_evolution(parts, site_index = 1,
                                relative = TRUE, highlight_col = 2)
tbl2$table

# Numeric codes (skip alphabet decoding)
tbl3 = tabulate_site_evolution(parts, site_index = 1, use_letters = FALSE)
rownames(tbl3$table)
}
\seealso{
\code{\link{partition_time_windows}} for producing the
typical input list of partitions; \code{\link{get_site_counts}}
for the count-tabulation primitive; \code{\link{decode_aa_sequence}}
for the alphabet code mapping; \code{\link{calculate_hellinger_matrix}}
for the related cross-partition distance calculation on the same
data shape.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster_sites_by_entropy.R
\name{cluster_sites_by_entropy}
\alias{cluster_sites_by_entropy}
\title{Cluster a Univariate Numeric Vector by Gaussian Mixture Model}
\usage{
cluster_sites_by_entropy(
  entropies,
  nr,
  nsites = length(entropies),
  precision = 6L,
  removez = TRUE,
  removesngl = TRUE,
  transfr = NULL,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{entropies}{Numeric vector to cluster. In the package's primary use
case these are per-site Shannon entropy values, but any univariate
numeric vector is accepted.}

\item{nr}{Integer. Total number of sequences from which the entropies
were computed. Required only when \code{removesngl = TRUE}.}

\item{nsites}{Integer. Expected number of sites. If it mismatches
\code{length(entropies)}, the actual length is used with a warning.
Default is \code{length(entropies)}.}

\item{precision}{Integer. Decimal places for rounding during singleton
threshold comparison and the all-identical uniqueness check. Default
is \code{6}.}

\item{removez}{Logical. If \code{TRUE}, removes sites with entropy = 0
(invariant sites), using a small tolerance \code{1e-9} to absorb
floating-point near-zeros. Default is \code{TRUE}.}

\item{removesngl}{Logical. If \code{TRUE}, removes sites whose entropy
equals the singleton value (one differing sequence out of \code{nr}).
Uses tolerance-based comparison. Default is \code{TRUE}.}

\item{transfr}{A function, or an object of class \code{transform} with a
\code{$transform()} method, applied to entropies before clustering.
Default is \code{NULL} (no transformation).}

\item{verbose}{Logical. If \code{TRUE}, emits diagnostic warnings for
non-fatal events (empty partitions, Mclust failures, etc.). Default is
\code{FALSE}.}

\item{...}{Additional arguments passed to \code{\link[mclust]{Mclust}}.}
}
\value{
A named list with two elements:
\item{FitObject}{The raw \code{Mclust} result, or a minimal
\code{list(classification = integer(0L))} when clustering was bypassed
or failed.}
\item{DataFrame}{A data frame with columns \code{sites} (original site
indices), \code{entropies} (values after any transformation), and
\code{class} (GMM cluster label). The \code{class} column is
\strong{always} present in every return path, including zero-row
DataFrames. Downstream consumers need only guard on \code{nrow(df) > 0}
before accessing class values. Raw Mclust labels are returned as-is;
call \code{\link{relabel_entropy_classes}} to standardise label ordering.}
}
\description{
Wraps \code{\link[mclust]{Mclust}} for unsupervised clustering
of a univariate numeric vector, with preprocessing rules and edge-case
handling tailored to per-site Shannon entropy values from viral sequence
data, which is the package's primary use case, but applicable to any
univariate data the user wishes to cluster by GMM.
}
\details{
In the package's typical use, sites are clustered by their Shannon entropy
to identify groups of residue positions with similar variability across
a sequence collection. Two preprocessing rules apply when clustering
entropies: \code{removez = TRUE} drops invariant sites (entropy = 0), and
\code{removesngl = TRUE} drops singleton sites whose entropy corresponds
to exactly one differing sequence across \code{nr} rows.

Class assignment rules (applied in priority order):
\itemize{
\item \strong{No rows remaining after filtering}: empty DataFrame returned
with a zero-length \code{class} column (consistent schema).
\item \strong{Single row remaining}: class \code{1} assigned directly;
Mclust is not called (undefined on 1 observation).
\item \strong{All entropies identical}: class \code{999} for all sites
(sentinel — one undifferentiated group).
\item \strong{Normal Mclust result}: raw class labels \code{1, 2, ..., G}.
These are Mclust's own integer labels, ordered by increasing component
mean (univariate Mclust orders components by mean) — call
\code{relabel_entropy_classes()} on the returned data frame to obtain
application-friendly class labels (highest-entropy class \code{= 1},
lowest-entropy class \code{= G}).
\code{\link{relabel_entropy_classes}} on the returned DataFrame to
standardise so that class 1 = highest-entropy group.
\item \strong{Mclust failure}: empty DataFrame returned (same schema as
the no-rows case), treating the partition as uninformative.
}
}
\examples{
# Clear bimodal structure: 5 low-entropy + 5 high-entropy sites.
set.seed(42)
entropies <- c(rnorm(5, mean = 0.1, sd = 0.01),
               rnorm(5, mean = 1.5, sd = 0.1))
result <- cluster_sites_by_entropy(entropies, removez = FALSE,
                                   removesngl = FALSE)
print(result$DataFrame)

# Single-row edge case: class = 1 assigned directly.
res1 <- cluster_sites_by_entropy(0.35, removesngl = FALSE)
print(res1$DataFrame)

# All-identical edge case: class = 999 (sentinel, one undifferentiated group).
res2 <- cluster_sites_by_entropy(c(0.35, 0.35, 0.35), removesngl = FALSE)
print(res2$DataFrame)
}
\seealso{
\code{\link{calculate_entropy}} for computing per-site entropy values,
\code{\link{relabel_entropy_classes}} for standardising the returned
class labels, and \code{\link{partition_time_windows}}, which calls
this function on each temporal partition.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_variants.R
\name{get_variants}
\alias{get_variants}
\title{Parse SARS-CoV-2 VOC/VOI Variant Metadata from Excel}
\usage{
get_variants(tibble, check = FALSE)
}
\arguments{
\item{tibble}{A \code{tibble} or \code{data.frame} produced by
\code{\link[readxl]{read_excel}} from \code{SARS_CoV_2_VOC_VOI.xlsx}.
Column 1 must be the \code{Variant} field column; columns 2 onward are
one column per variant in the order listed above.}

\item{check}{Logical. If \code{TRUE}, validates that all internal vectors
have equal length and stops with an informative message on mismatch.
Default \code{FALSE}.}
}
\value{
A named list with 13 elements; see Details for full descriptions.
}
\description{
Reads a structured Excel workbook of SARS-CoV-2 Variants of
Concern (VOC) and Variants of Interest (VOI) and returns a named list
containing mutation profiles, nomenclature, temporal detection metadata,
defining SNPs, and a fully citable reference table with interactive
display support.
}
\details{
Intended to be called once during data preparation via
\code{data-raw/sarscov2_variants.R}:

\preformatted{
  variants_dat  <- readxl::read_excel("SARS_CoV_2_VOC_VOI.xlsx")
  variants_list <- get_variants(variants_dat)
  saveRDS(variants_list,
          file = "inst/extdata/sarscov2_variants.rds")
}

The saved object can then be loaded anywhere in the package or vignettes:

\preformatted{
  voc_data <- readRDS(system.file("extdata", "sarscov2_variants.rds",
                                  package = "ViralEntropR"))
}

\strong{Column order of variants in the Excel workbook:}
Alpha, Beta, Epsilon, Eta, Iota, Kappa, Delta, Lambda, Gamma, Zeta,
Theta, Omicron.

\strong{Returned list elements:}
\describe{
\item{\code{WHO_Label}}{List. WHO variant label strings (e.g. \code{"Alpha"}).}
\item{\code{Pango_Lineage}}{List. Pango lineage designations.}
\item{\code{GISAID_Clade}}{Character vector. GISAID clade strings.}
\item{\code{Nextstrain_Clade}}{Character vector. Nextstrain clade strings.}
\item{\code{Country_First_Detected}}{List. Country of first documented
detection per variant.}
\item{\code{Date_Earliest_Sample}}{Character vector. Month-Year of the
earliest documented sample per variant.}
\item{\code{Date_First_Detected}}{Character vector. Month-Year of
world-level first detection.}
\item{\code{Date_First_Detected_US}}{Character vector. Month-Year of
first US detection.}
\item{\code{Spike_Mutations}}{List. Spike protein mutation strings read
from the Excel workbook.}
\item{\code{Mutation_Sites}}{List. Integer site positions extracted from
\code{Spike_Mutations}.}
\item{\code{Defining_SNPs}}{List. Canonical defining SNP strings per
variant (\code{NA} where not characterised).}
\item{\code{Defining_SNP_Sites}}{List. Integer positions extracted from
\code{Defining_SNPs}.}
\item{\code{References}}{Named list with three elements:
\code{$data} — data frame of 21 verified references;
\code{$display(variant = NULL)} — interactive \code{\link[DT]{datatable}},
optionally filtered by WHO label;
\code{$cite(variant)} — character vector of formatted citation strings
suitable for manuscript use.}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate_variant_evolution.R
\name{simulate_variant_evolution}
\alias{simulate_variant_evolution}
\title{Simulate Viral Variant Evolution}
\usage{
simulate_variant_evolution(
  ref_sequences,
  n_ref_months = 3L,
  start_date,
  end_date,
  variants_config,
  variant_intervals,
  n_new_mutations = 1L,
  mutation_rate = 2,
  mutation_rate_variability = 0.25,
  deleterious_rate = 0,
  n_deleterious_limit = 1L,
  n_sequences_total = 140L,
  ref_variability = FALSE,
  n_ref_sequences = 100L,
  prob_deletion_event = 0,
  n_rows_to_delete = 0L,
  seed = NULL
)
}
\arguments{
\item{ref_sequences}{Either a single character string giving the reference
amino-acid sequence, or a data.frame with a \code{Date} column and site
columns named with integer position labels (\code{"1"}, \code{"2"},
..., \code{as.character(L)}; one column per site). The sequence length
\code{L} must be at least 3, since variant mutations and random
substitution events are sampled from sites \eqn{>2} (sites 1 and 2 are
reserved for reference variability when \code{ref_variability = TRUE}).}

\item{n_ref_months}{Integer. Duration of the initial reference phase
(months). Default \code{3}.}

\item{start_date}{Character or Date. Simulation start.}

\item{end_date}{Character or Date. Simulation end (inclusive month).}

\item{variants_config}{Integer vector. Number of mutations per variant,
e.g., \code{c(2, 6)}.}

\item{variant_intervals}{Integer vector. Months between consecutive
variant emergences (length \code{length(variants_config) - 1}).}

\item{n_new_mutations}{Integer. Number of de-novo sequence \emph{copies}
introduced at a variant's first appearance. Despite the name, this
controls the count of seeded sequences, not the number of mutations
carried by the variant (which is set per-variant by
\code{variants_config}). Default \code{1}.}

\item{mutation_rate}{Numeric scalar or vector. Monthly growth multiplier
per variant. Default \code{2}.}

\item{mutation_rate_variability}{Numeric in \code{[0, 1)}. Fractional
spread around \code{mutation_rate}: growth is drawn from
\code{Uniform(mult*(1-v), mult*(1+v))}. Set to \code{0} for
deterministic growth. Default \code{0.25}.}

\item{deleterious_rate}{Numeric in \code{[0, 1]}. Per-mutation probability
of being flagged as "deleterious". Default \code{0}.}

\item{n_deleterious_limit}{Integer. Per-site cap on the number of variant
rows allowed to retain a flagged ("deleterious") allele in a given
month; overflow is reverted to the reference allele at the flagged
site only. See \strong{Capped "Deleterious" Sites} in Details.
Default \code{1}.}

\item{n_sequences_total}{Integer. Total sequences generated over the full
mutation phase (spread evenly per month). Default \code{140}.}

\item{ref_variability}{Logical. If \code{TRUE}, introduces low-level
variability at sites 1 and 2 of the reference pool. Requires
\code{n_ref_sequences >= 6 * n_ref_months} for the variability rows
to be correctly captured by the multi-variant fill pool; smaller
values trigger a warning. Default \code{FALSE}.}

\item{n_ref_sequences}{Integer. When \code{ref_sequences} is a single
string, the number of reference rows to create. Default \code{100}.}

\item{prob_deletion_event}{Numeric in \code{[0, 1]}. Per-month probability
of triggering a substitution event affecting \code{n_rows_to_delete}
rows. (Parameter name retained for backward compatibility; the
operation is substitution, not literal deletion. See Details.)
Default \code{0}.}

\item{n_rows_to_delete}{Integer. Number of rows affected when a
substitution event fires (see \code{prob_deletion_event}). Default
\code{0}.}

\item{seed}{Integer or \code{NULL}. Random seed for reproducibility.
Default \code{NULL}.}
}
\value{
An object of class \code{"viralSim"}, a named list:
\item{Simulation_Output}{Data frame of all sequences, one column per site
(named \code{"1"}, ..., \code{as.character(L)}) plus \code{Variant}
(per-row strain label), \code{Phase} (\code{"R"} or \code{"RV"} for
reference rows depending on whether the row matches the wild-type
sequence; \code{"M1"}, \code{"M2"}, ... in mutation-phase months —
note that \code{"Mn"} is a \emph{month label} and applies to every
row in that month regardless of strain), \code{Date}, \code{Period}
(1-indexed month), and \code{Delet} (\code{"Yes"} for rows affected
by substitution events, \code{"No"} otherwise).}
\item{Variant_Details}{List of per-variant metadata. Each element has
\code{em} (emergence-month index), \code{pos} (mutated site indices),
\code{flags} (logical vector of "deleterious" flags per mutated site),
\code{vseq} (full mutated reference sequence as character vector),
\code{mult} (growth multiplier), \code{last} (count in the most recent
month), \code{cum} (cumulative count), and \code{repl} (logical flag
set when the variant has been displaced by its successor).}
\item{Simulation_Dates}{Date vector of all simulated months.}
\item{Baseline_Ref_Sequence}{Character. The wild-type reference string.}
\item{Delet_Records}{Named list of substitution events keyed by date
(\code{"YYYY-MM-DD"}). Each entry holds \code{period}, \code{date},
\code{site}, \code{old_aa}, \code{new_aa}, and \code{rows} (absolute
row indices in \code{Simulation_Output} that were modified).}
\item{Pool}{Fill-pool data frame from the final multi-variant competition
period (includes \code{._weight} column); \code{NULL} if at most one
variant was ever active.}
}
\description{
Generates a synthetic, ground-truth-traceable time series of
viral amino-acid sequences in which a reference strain is progressively
challenged by emerging variants under stochastic growth and pairwise
competition. Designed as a controllable signal generator for benchmarking
the entropy, Hellinger, and change-point components of the
\pkg{ViralEntropR} pipeline; the dynamics are deliberately simplified
rather than a faithful biological model of evolutionary fitness.
}
\details{
Simulates the trajectory of a viral population over discrete monthly time
steps. All stochastic components are reproducible under a fixed
\code{seed} and are designed to expose a known emergence pattern that
downstream detection methods can be evaluated against.

\strong{Simulation Phases:}
\enumerate{
\item \strong{Reference Phase:} A stable period in which only the
reference strain exists. Reference months are resampled with
replacement to assemble \code{n_ref_months} monthly batches.
\item \strong{Variant Emergence:} New variants are introduced at the
monthly intervals specified by \code{variant_intervals}.
\item \strong{Pairwise Growth:} Each month the newest variant grows
alongside its immediate predecessor. Both grow stochastically: each
variant's next-month count is its previous-month count times
\code{Uniform(mult*(1-variability), mult*(1+variability))}. The two
are not coupled by a shared frequency constraint — only by the hard
monthly quota
\code{ceiling(n_sequences_total / (n_periods - n_ref_months))} that
trims any overshoot. Older, non-paired variants enter the fill pool
with sampling weights \code{2^(i+1)}.
\item \strong{Capped "Deleterious" Sites:} A Bernoulli-random subset of
each variant's mutated positions is flagged with probability
\code{deleterious_rate}. At each post-emergence month the count of
variant rows still carrying the flagged allele is capped at
\code{n_deleterious_limit} by reverting overflow sequences to the
reference allele \emph{at the flagged site only}. This is a
per-site cap, not a genotype-level fitness penalty.
\item \strong{Random Substitution Events:} Each month independently
fires a Bernoulli coin flip with probability
\code{prob_deletion_event}. If it fires, the last
\code{n_rows_to_delete} rows of that month's batch receive a single
non-reference amino-acid substitution at a randomly chosen site
(sites 1 and 2 excluded), and those rows are flagged with
\code{Delet = "Yes"}. Parameter names retain "deletion" / "delete"
for backward compatibility; the operation itself is a substitution.
}
}
\examples{
ref_seq <- "MKTIIALSYIFCLVFADYKDDDDK"

sim <- simulate_variant_evolution(
  ref_sequences   = ref_seq,
  n_ref_months    = 3,
  start_date      = "2021-01-01",
  end_date        = "2021-12-01",
  variants_config = c(3, 5),
  variant_intervals = c(4),
  n_sequences_total = 50,
  mutation_rate   = 1.5,
  seed            = 123
)
head(sim$Simulation_Output)
}
\seealso{
\code{\link{partition_time_windows}},
\code{\link{calculate_hellinger_matrix}},
\code{\link{detect_changepoints_ecp}},
\code{\link{detect_changepoints_hdcp}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_hellinger_matrix.R
\name{calculate_hellinger_matrix}
\alias{calculate_hellinger_matrix}
\title{Calculate Hellinger Distance Matrix}
\usage{
calculate_hellinger_matrix(
  partitions,
  sites = seq_len(ncol(partitions[[1]])),
  aa_levels = 25L,
  normalized = FALSE,
  labels = paste0("T", seq_along(partitions)),
  include_freq_tables = FALSE
)
}
\arguments{
\item{partitions}{A list of data frames, one per time window. Each data frame
must have numeric-encoded amino acid sequences as columns (integers 1 to
\code{aa_levels}). This is typically the \code{$Partitions} element
returned by \code{\link{partition_time_windows}}.}

\item{sites}{Integer vector. Indices of the sites to analyse. Defaults to all
sites (\code{seq_len(ncol(partitions[[1]]))}).}

\item{aa_levels}{Integer. Alphabet size. Must match the encoding used
when the partitions were created: \code{\link{encode_aa_sequence}}
produces values 1–25 by default (20 standard residues, three
ambiguous codes, \code{*}, \code{-}). Default is \code{25L}.}

\item{normalized}{Logical. If \code{TRUE}, scales distances by
\eqn{1/\sqrt{2}} to bound the result in \eqn{[0, 1]}. Default is
\code{FALSE}.}

\item{labels}{Character vector of partition labels. Length must equal
\code{length(partitions)}. The first label names the reference
partition; subsequent labels become column names of the returned
matrix. Defaults to \code{"T1"}, \code{"T2"}, …}

\item{include_freq_tables}{Logical. If \code{TRUE}, the return value also
includes raw count tables and proportion tables for each site. Default is
\code{FALSE}.}
}
\value{
When \code{include_freq_tables = FALSE} (default): a numeric matrix
with rows corresponding to \code{sites} and columns corresponding to
\code{labels[-1]}, each entry being the Hellinger distance from the
reference partition (\code{labels[1]}) at that site. Row names are
the site indices; column names are taken from \code{labels[-1]}.
With default labels, the reference partition is \code{"T1"} and the
matrix has columns \code{"T2"}, \code{"T3"}, ….

When \code{include_freq_tables = TRUE}: a named list with elements
\code{Sites}, \code{Hellinger_Distances}, \code{Frequency_Tables}, and
\code{Proportions_Tables}.
}
\description{
Computes the Hellinger distance between the amino acid distribution
of a reference time point (first partition) and all subsequent time points,
for each requested site.
}
\details{
The Hellinger distance between two discrete distributions \eqn{P} and \eqn{Q} is:
\deqn{H(P, Q) = \sqrt{\sum_{i=1}^{k} (\sqrt{p_i} - \sqrt{q_i})^2}}
When \code{normalized = TRUE} the result is scaled by \eqn{1/\sqrt{2}}, bounding
the distance to \eqn{[0, 1]}. Otherwise the range is \eqn{[0, \sqrt{2}]}.

Internally, amino acid counts per site per partition are tabulated using
\code{\link{get_site_counts}} (built on \code{\link[base]{tabulate}}).
Per-partition proportions and Hellinger distances are then computed by
fully vectorised matrix operations — no inner loop over partitions.
}
\examples{
# Toy 3-partition data: site 1 starts homogeneous (all Alanine, 1),
# acquires Valine (20) across two later partitions; site 2 stays
# constant (all Valine throughout).
p1 = data.frame(s1 = c(1, 1, 1, 1, 1),  s2 = c(20, 20, 20, 20, 20))
p2 = data.frame(s1 = c(20, 20, 20, 20, 20), s2 = c(20, 20, 20, 20, 20))
p3 = data.frame(s1 = c(1, 1, 20, 20, 20), s2 = c(20, 20, 20, 20, 20))
parts = list(T1 = p1, T2 = p2, T3 = p3)

result = calculate_hellinger_matrix(parts, sites = 1:2)
print(result)
# Site 1 has nonzero distance in T2 and T3 (composition shift).
# Site 2 has zero distance throughout (constant).

# With raw frequency tables for inspection.
result2 = calculate_hellinger_matrix(parts, sites = 1:2,
                                      include_freq_tables = TRUE)
print(result2$Frequency_Tables[[1]])
}
\seealso{
\code{\link{partition_time_windows}} for producing temporally
partitioned data, \code{\link{encode_aa_sequence}} for integer
encoding consistent with \code{aa_levels}, and
\code{\link{detect_changepoints_ecp}} or
\code{\link{detect_changepoints_hdcp}} for downstream change-point
detection on the returned matrix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/partition_time_windows.R
\name{partition_time_windows}
\alias{partition_time_windows}
\title{Partition Data into Time Windows}
\usage{
partition_time_windows(
  data,
  n_sites,
  window_length = 1,
  window_type = 3,
  start_date = NULL,
  end_date = NULL,
  date_format = "\%b-\%Y",
  verbose = FALSE,
  ...
)
}
\arguments{
\item{data}{Data frame. Must contain a column named \code{Date}
coercible to \code{Date}. Columns \code{1} through \code{n_sites}
must be the numeric site columns; any additional columns (such as
\code{Country} or other per-sequence metadata) may follow.}

\item{n_sites}{Integer. Number of site columns. Site columns must
occupy positions \code{1} through \code{n_sites} of \code{data};
any other columns (\code{Date}, optionally \code{Country}, etc.)
must come after.}

\item{window_length}{Integer. Window duration in months.
Default is \code{1}.}

\item{window_type}{Integer (1, 2, or 3). Windowing strategy:
\code{1} = Cumulative, \code{2} = Sliding, \code{3} = Disjoint.
Default is \code{3}.}

\item{start_date}{Date or character. Window start. Defaults to earliest date
in \code{data}.}

\item{end_date}{Date or character. Window end (cutoff). Defaults to latest
date in \code{data}.}

\item{date_format}{Character. Format string for date labels. Default is
\code{"\%b-\%Y"}.}

\item{verbose}{Logical. If \code{TRUE}, prints a progress bar and completion
summary. Default is \code{FALSE}.}

\item{...}{Additional arguments passed to
\code{\link{cluster_sites_by_entropy}}.}
}
\value{
A named list:
\item{Partitions}{List of data frame chunks, one per window.}
\item{Entropies}{List of numeric entropy vectors, one per window.}
\item{Clusters}{List of clustering results from
\code{\link{cluster_sites_by_entropy}}, one per window.}
\item{Max_Entropy}{Numeric vector. Maximum cluster class label per window
(equals number of clusters found; \code{NA} if window was empty).}
\item{Dates_Labels}{Character vector of window label strings.}
\item{N_partitions}{Integer. Total number of windows.}
}
\description{
Splits a data frame of time-stamped sequences into discrete time
windows, computes per-site entropy and Mclust clustering for each window.
}
\details{
Three windowing strategies are supported via \code{window_type}.
Throughout the descriptions below, \code{T} denotes the number of whole
months between \code{start_date} and \code{end_date}, and \code{w}
denotes \code{window_length}:
\itemize{
\item \strong{1 — Cumulative}: Start is fixed; end expands by
\code{w} months each step. Each window includes all prior data
plus one more period. Produces \code{floor(T / w)} chunks.
\item \strong{2 — Sliding}: A window of \code{w} months slides one
month at a time. Produces \code{T - w + 1} chunks.
\item \strong{3 — Disjoint}: Consecutive non-overlapping windows of
\code{w} months. Produces \code{floor(T / w)} chunks. Default.
}

For example, with \code{T = 12} months and \code{w = 2}: cumulative
produces 6 chunks (each progressively larger); sliding produces 11
chunks (overlapping, each 2 months wide); disjoint produces 6 chunks
(each exactly 2 months, non-overlapping).

\strong{Empty windows.} When a window contains no observations, the
corresponding entries in the returned lists carry placeholder values:
\code{Entropies} is a zero-vector of length \code{n_sites};
\code{Clusters} is a schema-consistent empty result matching
\code{cluster_sites_by_entropy}'s empty-input return; \code{Max_Entropy}
is \code{NA_integer_}. Downstream consumers can guard on
\code{nrow(Partitions[[i]]) > 0} or \code{!is.na(Max_Entropy[i])}.

\strong{Column layout requirement.} The function extracts site columns
as \code{data[, seq_len(n_sites), drop = FALSE]}. The site columns
must therefore occupy positions \code{1} through \code{n_sites}.
\code{Date} (and any other metadata columns such as \code{Country})
must come after the site columns. A common error is to place
\code{Date} first; this will produce incorrect entropy values.
}
\examples{
dates <- seq(as.Date("2020-01-01"), as.Date("2020-06-01"), by = "month")
df <- data.frame(
  s1   = 1L,
  s2   = c(rep(1L, 15), rep(2L, 15)),
  Date = rep(dates, each = 5)
)
res <- partition_time_windows(df, n_sites = 2, window_length = 2,
                              window_type = 3, verbose = TRUE)
print(res$Dates_Labels)
print(res$Max_Entropy)

}
\seealso{
\code{\link{calculate_entropy}} for the per-site entropy
computation; \code{\link{cluster_sites_by_entropy}} for the GMM
clustering applied per window; \code{\link{relabel_entropy_classes}}
for standardizing the cluster labels in downstream consumers; and
\code{\link{calculate_hellinger_matrix}} for typical downstream use
of the \code{Partitions} output.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_ambiguous_sequences.R
\name{filter_ambiguous_sequences}
\alias{filter_ambiguous_sequences}
\title{Remove Sequences Containing Ambiguous Residues}
\usage{
filter_ambiguous_sequences(NumMatrix, option = 1)
}
\arguments{
\item{NumMatrix}{A matrix. Rows are sequences, columns are sites. Either
integer-encoded (\code{option = 1}) or character (\code{option = 2}).
Despite the name, character matrices are also accepted under
\code{option = 2}.}

\item{option}{Integer. \code{1} (default) for integer-encoded matrices
produced by \code{\link{encode_aa_sequence}}; \code{2} for character
matrices produced by \code{\link{fasta_to_char_matrix}}.}
}
\value{
A named list:
\item{OriginalDim}{Character string reporting the number of input
sequences.}
\item{NewDim}{Character string reporting the number of sequences
remaining after filtering.}
\item{NumberAmbiguous}{Character string reporting the number of
sequences that contained at least one ambiguous residue.}
\item{RangeAmbiguous}{Character string reporting the min and max count
of ambiguous residues per removed sequence, or
\code{"No ambiguous sequences found"} when none were removed.}
\item{DeletedSeqId}{Integer vector of row indices that were removed.
Empty integer vector if nothing was removed.}
\item{FilteredMatrix}{The filtered matrix with ambiguous rows removed,
preserving the original column structure and storage mode.}
}
\description{
Removes rows (sequences) that contain at least one ambiguous
amino acid residue (B, J, X, or Z) — and, under integer-encoded input,
any unrecognised character — from a sequence matrix. Accepts both
integer-encoded matrices and character matrices.
}
\details{
\strong{What is removed.} Sequences are flagged for removal if any of
their residue positions contain one of the four IUPAC ambiguous codes:
\itemize{
\item \code{B} — Aspartate / Asparagine.
\item \code{J} — Leucine / Isoleucine.
\item \code{X} — any residue.
\item \code{Z} — Glutamate / Glutamine.
}

\strong{What is NOT removed.} Standard alignment gaps (\code{-},
integer code \code{25}) are retained — gaps represent known absences
rather than uncertain identities and are typically positionally
meaningful in aligned data. Sequences containing only canonical 20
amino acids and gaps are kept.

\strong{How input mode is handled.}
\itemize{
\item \code{option = 1} (integer-encoded input from
\code{\link{encode_aa_sequence}}): rows are removed if any cell
equals \code{0} (unrecognised — including J, NA, empty, lowercase
mismatches, byte-order marks, and other characters that fell
outside the encoding alphabet), \code{21} (B), \code{22} (Z), or
\code{23} (X). The \code{0} sentinel acts as a catch-all for
anything not in the 25-symbol alphabet.
\item \code{option = 2} (character input from
\code{\link{fasta_to_char_matrix}}): rows are removed if any cell
is exactly \code{"B"}, \code{"J"}, \code{"X"}, or \code{"Z"}.
Unrecognised characters in character input (e.g. lowercase letters,
\code{NA}, empty strings) are NOT caught at this stage; encode
first if you need that catch-all behaviour.
}

\strong{Performance.} Detection is fully vectorised: a single logical
matrix comparison followed by \code{\link[base]{rowSums}} counts
ambiguous residues per sequence in one C-level call, replacing the
original row-by-row loop for a substantial speed improvement on
large matrices (100k+ rows).
}
\examples{
# Synthetic example: 50 sequences, 10 sites, drawn from canonical residues.
set.seed(1)
m <- matrix(sample(1:20, 500, replace = TRUE), nrow = 50, ncol = 10)
# Inject ambiguous codes into 3 specific rows: 21 (B), 23 (X), 0 (unrecognised).
m[c(3, 17, 42), sample(1:10, 3)] <- c(21, 23, 0)
result <- filter_ambiguous_sequences(m, option = 1)
cat(result$NumberAmbiguous, "\n")
cat(result$RangeAmbiguous, "\n")
dim(result$FilteredMatrix)

# Character-mode example.
chr <- matrix(c("M", "K", "T", "I", "I", "X", "K", "T", "I", "I"),
               nrow = 2, byrow = TRUE)
filter_ambiguous_sequences(chr, option = 2)$DeletedSeqId
}
\seealso{
\code{\link{encode_aa_sequence}} and
\code{\link{fasta_to_char_matrix}} for producing the typical input;
\code{\link{decode_aa_sequence}} for inspecting the surviving
sequences in character form.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/encode_aa_sequence.R
\name{encode_aa_sequence}
\alias{encode_aa_sequence}
\title{Encode Amino Acid Sequences}
\usage{
encode_aa_sequence(matrix_input)
}
\arguments{
\item{matrix_input}{A character matrix of amino acid codes, typically
produced by \code{\link{fasta_to_char_matrix}}. A non-matrix input is
coerced via \code{\link[base]{as.matrix}}. Lowercase input is
accepted; characters are uppercased before lookup.}
}
\value{
A numeric matrix of the same dimensions as \code{matrix_input},
with the same \code{dimnames}. Each cell is either an integer in
\code{1:25} from the alphabet table above, or \code{0} for any input
not in the alphabet (including \code{NA} and the empty string).
}
\description{
Converts a character matrix of amino acid codes to its
integer-encoded representation under the package's 25-symbol alphabet.
Inverse of \code{\link{decode_aa_sequence}}.
}
\details{
Encoding is a single vectorised lookup against a fixed named map. The
input matrix is flattened, normalised to uppercase, indexed against the
alphabet's name vector in one operation, and reshaped — there is no
per-row or per-element loop.

The 25-symbol alphabet covers the twenty standard residues
(\code{A=1, R=2, N=3, D=4, C=5, Q=6, E=7, G=8, H=9, I=10, L=11, K=12,
M=13, F=14, P=15, S=16, T=17, W=18, Y=19, V=20}), the three IUPAC
ambiguous codes \code{B=21, Z=22, X=23}, the stop codon \code{*=24},
and the alignment gap \code{-=25}.

Any character outside this alphabet — including \code{NA}, the empty
string, lowercase letters not in the standard set after uppercasing
(e.g. \code{J} for leucine/isoleucine), and non-letter symbols — maps
to the sentinel \code{0}. This sentinel is recognised by downstream
functions: \code{\link{filter_ambiguous_sequences}} treats \code{0}
as ambiguous (alongside \code{B}, \code{X}, \code{Z}) and removes
affected sequences, and \code{\link{decode_aa_sequence}} round-trips
it back to the string \code{"0"}.

Row and column names of the input matrix are preserved on the output.
}
\examples{
# 1. Encode a simple matrix of sequences.
seq_mat = matrix(c("A", "R", "N", "D"), nrow = 2, byrow = TRUE)
encoded = encode_aa_sequence(seq_mat)
print(encoded)

# 2. Gaps and unknown characters.
# '-' maps to 25; '?' is not in the alphabet and maps to the sentinel 0.
gapped_mat = matrix(c("A", "-", "G", "?"), nrow = 1)
encode_aa_sequence(gapped_mat)

# 3. Lowercase input is accepted (uppercased before lookup).
encode_aa_sequence(matrix(c("a", "r", "n", "d"), nrow = 2))

# 4. NA and empty string both map to 0.
encode_aa_sequence(matrix(c("A", NA, "G", ""), nrow = 2))
}
\seealso{
\code{\link{decode_aa_sequence}} for the inverse operation;
\code{\link{fasta_to_char_matrix}} for the FASTA-to-character-matrix
step that typically precedes encoding;
\code{\link{filter_ambiguous_sequences}} for downstream removal of
sequences containing ambiguous or unrecognised residues.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/relabel_entropy_classes.R
\name{relabel_entropy_classes}
\alias{relabel_entropy_classes}
\title{Relabel Entropy Classes}
\usage{
relabel_entropy_classes(df)
}
\arguments{
\item{df}{A data frame containing clustering results. Must have columns
\code{class} (integer cluster labels from
\code{\link{cluster_sites_by_entropy}}) and \code{entropies}
(numeric Shannon entropy values per row).}
}
\value{
The data frame with a relabeled \code{class} column (integer)
and an added \code{num_classes} column reporting the count of
distinct labels present after relabeling.
}
\description{
Relabels GMM cluster labels so that class 1 is always the
highest-entropy group, class 2 the second highest, and so on.
}
\details{
For univariate input (the package's typical use case), \code{\link[mclust]{Mclust}}
orders cluster components by increasing mean. Applied to per-site Shannon
entropy values, this places the highest-entropy group at the highest
label number (label \code{G} for \code{G} fitted components). This
function flips that convention so that label \code{1} always denotes
the highest-entropy group, which is more natural for downstream
filtering ("class 1 = top") and visual labeling. Cluster identities
and means are unchanged; only the integer labels are remapped.

\strong{Sentinel preservation.} The class label \code{999L} marks
undifferentiated groups (all entropies equal in
\code{\link{cluster_sites_by_entropy}}) and is never relabeled. If a
data frame contains a mixture of real GMM labels and one or more
\code{999} entries, only the real labels are ranked and overwritten;
the \code{999} rows pass through unchanged.

\strong{No-op return paths.} The input is returned unchanged (with at
most a \code{num_classes} column added) in four situations:
\itemize{
\item Missing required columns (\code{class} or \code{entropies}) —
warns and returns input.
\item Zero rows.
\item All rows already carry the sentinel \code{999L}.
\item Only one class label is present (relabeling is a no-op).
}
}
\examples{
# Standard case: three classes, ranked by mean entropy.
df <- data.frame(
  sites     = 1:6,
  entropies = c(0.1, 0.1, 0.5, 0.5, 1.2, 1.3),
  class     = c(1L, 1L, 2L, 2L, 3L, 3L)
)
relabel_entropy_classes(df)
# Class 3 (highest mean entropy 1.25) -> 1; class 2 (0.5) -> 2;
# class 1 (0.1) -> 3.

# Sentinel preservation: 999 rows pass through unchanged even when
# mixed with real classes.
df_mixed <- data.frame(
  sites     = 1:4,
  entropies = c(0.5, 1.2, 0.3, 0.4),
  class     = c(1L, 2L, 999L, 1L)
)
relabel_entropy_classes(df_mixed)
# Class 2 (highest) -> 1; class 1 -> 2; class 999 stays 999.
}
\seealso{
\code{\link{cluster_sites_by_entropy}} for the upstream
clustering step; \code{\link{plot_entropy_trajectories}} and
\code{\link{plot_site_class_trajectory}} for downstream consumers
that rely on the relabeled class convention;
\code{\link{partition_time_windows}} which calls
\code{cluster_sites_by_entropy} per window.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fasta_to_char_matrix.R
\name{fasta_to_char_matrix}
\alias{fasta_to_char_matrix}
\title{Convert FASTA Object to Character Matrix}
\usage{
fasta_to_char_matrix(fsta)
}
\arguments{
\item{fsta}{An \code{AAStringSet} object, typically the output of
\code{\link[Biostrings]{readAAStringSet}}. May be aligned or
unaligned (see Details).}
}
\value{
A character matrix with \code{length(fsta)} rows and
\code{max(nchar(as.character(fsta)))} columns. Each cell contains a
single-character amino acid code from the input sequences (or the
gap character \code{"-"} for padded positions in unaligned input).
The matrix has no row or column names; sequence names from the
\code{AAStringSet} are not carried over. An empty input
(\code{length(fsta) == 0}) returns an empty 0-by-0 character matrix.
}
\description{
Converts an \code{AAStringSet} object loaded via
\code{\link[Biostrings]{readAAStringSet}} into a character matrix where
rows are sequences and columns are residue positions (sites). Inverse
structural transformation of \code{\link{encode_aa_sequence}}'s expected
input shape.
}
\details{
\strong{Alignment.} The function expects an aligned \code{AAStringSet}
— all sequences of equal width. Unaligned input is accepted and shorter
sequences are right-padded with the gap character \code{"-"} to match
the longest sequence, but downstream entropy-based analysis assumes
positional homology across rows; if sequences in the input are not
biologically aligned, results from per-site computations will not be
meaningful. For unaligned input, run a multiple-sequence alignment
(e.g. \code{msa::msa()} or \code{DECIPHER::AlignSeqs()}) before calling
this function.

\strong{Performance.} Conversion is fully vectorised: all sequences are
coerced to a single character string vector, split simultaneously, and
reshaped into a matrix in one operation. No per-row loop, no
intermediate list of split sequences kept alive — substantially faster
than per-row \code{strsplit} on large inputs (100k+ sequences).
}
\examples{
\donttest{
# Convert the bundled sample to a character matrix.
path  <- system.file("extdata", "sarscov2_sample.fasta.gz",
                      package = "ViralEntropR")
fasta <- Biostrings::readAAStringSet(path)
mat   <- fasta_to_char_matrix(fasta)
dim(mat)
mat[1:3, 1:10]
}
}
\seealso{
\code{\link{encode_aa_sequence}} for converting the resulting
character matrix to an integer-encoded matrix;
\code{\link{filter_ambiguous_sequences}} for removing rows containing
ambiguous residues; \code{\link[Biostrings]{readAAStringSet}} for
loading FASTA files into the input format.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_fasta_countries.R
\name{extract_fasta_countries}
\alias{extract_fasta_countries}
\title{Extract Countries from FASTA Sequence Names}
\usage{
extract_fasta_countries(sequence, position, problematic_characters = FALSE)
}
\arguments{
\item{sequence}{An \code{AAStringSet} object.}

\item{position}{Integer (1–4). Location of the country field within the
sequence name string:
\itemize{
\item \strong{1} — text before the first \code{|} (e.g.
\code{SouthKorea|...}).
\item \strong{2} — text between the first and second \code{|}.
\item \strong{3} — text between the first and second \code{/}.
\item \strong{4} — text after the last \code{|}.
}}

\item{problematic_characters}{Logical. If \code{TRUE}, sequence names are
re-encoded to UTF-8, replacing non-representable bytes with their
escaped form. Useful for FASTA files with non-ASCII characters in
headers. Default is \code{FALSE}.}
}
\value{
A named list with three elements:
\item{countries}{Character vector of extracted country strings, one per
sequence. \code{NA} where extraction failed (no match against the
chosen pattern).}
\item{message}{A single character string summarising extraction
success.}
\item{missing_id}{Integer vector of indices where extraction failed, or
\code{NA} if all extractions succeeded.}
}
\description{
Extracts country names from the sequence name strings of an
\code{AAStringSet} object loaded via
\code{\link[Biostrings]{readAAStringSet}}. Handles single-word (e.g.
\code{UK}), hyphenated (e.g. \code{Timor-Leste}), and multi-word (e.g.
\code{United States of America}) country names.
}
\details{
The function selects one of four regex patterns based on \code{position}
and applies it to each sequence name via
\code{\link[stringr]{str_extract}}. \strong{Only the first match per
header is returned.} If a header contains multiple delimited fields,
the country must be in the first such field for the corresponding
\code{position} value to extract it correctly. For example, with a
GISAID-style header
\code{Spike|hCoV-19/USA/OH/.../2021|2021-05-15|EPI_ISL_...|},
\code{position = 3} (between slashes) returns \code{USA}, but
\code{position = 2} (between pipes) returns \code{hCoV-19/USA/OH/...},
not \code{USA}. Inspect representative headers with
\code{names(sequence)[1]} before choosing \code{position}.

\strong{Encoding.} FASTA files with non-ASCII characters in headers
(accented characters, byte-order marks, etc.) can break regex
extraction. Setting \code{problematic_characters = TRUE} re-encodes
headers to UTF-8 with non-representable bytes escaped, allowing the
regex to proceed.
}
\examples{
\donttest{
path_sample  <- system.file("extdata", "sarscov2_sample.fasta.gz",
                             package = "ViralEntropR")
fasta_sample <- Biostrings::readAAStringSet(path_sample)

# Inspect header structure to confirm field positions before extraction.
sample(names(fasta_sample), 1)

# Extract countries (position 2 = between first and second pipe).
result <- extract_fasta_countries(fasta_sample, position = 2)
result$message
sort(table(result$countries), decreasing = TRUE)
}
}
\seealso{
\code{\link{extract_fasta_dates}} for the date-extraction
companion; \code{\link[Biostrings]{readAAStringSet}} for loading the
input \code{AAStringSet}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_fasta_dates.R
\name{extract_fasta_dates}
\alias{extract_fasta_dates}
\title{Extract Dates from FASTA Sequence Names}
\usage{
extract_fasta_dates(
  sequence,
  option = 1,
  date_format = "\%Y-\%m-\%d",
  custom_pattern = NULL
)
}
\arguments{
\item{sequence}{An \code{AAStringSet} object.}

\item{option}{Integer (1, 2, 3, or 4). Selects the built-in pattern when
\code{custom_pattern} is not supplied. See Details.}

\item{date_format}{Character. \code{strptime}-style format string used
to coerce extracted strings to \code{Date}. Default is
\code{"\%Y-\%m-\%d"}. Use \code{"\%Y-\%m"} together with
\code{option = 3}.}

\item{custom_pattern}{Character or \code{NULL}. A custom regex passed
directly to \code{\link[stringr]{str_extract}}. When supplied,
\code{option} is ignored. Default is \code{NULL}.}
}
\value{
A named list of six elements, each aligned with the input
\code{sequence}:
\item{raw_date_strings}{Character vector of extracted date strings
before any correction. \code{NA} where extraction failed.}
\item{corrected_date_strings}{Character vector with \code{-00}
substrings replaced by \code{-01}. \code{NA} where extraction
failed.}
\item{raw_dates}{\code{Date} vector coerced from
\code{raw_date_strings}. \code{NA} for unparseable or missing
strings (including any record where day = \code{00}).}
\item{corrected_dates}{\code{Date} vector coerced from
\code{corrected_date_strings}. \code{NA} for unparseable or missing
strings.}
\item{message}{Character string summarising extraction success.}
\item{missing_id}{Integer vector of indices where extraction failed,
or \code{NA} if all extractions succeeded.}
}
\description{
Extracts date strings from the sequence name strings of an
\code{AAStringSet} object loaded via
\code{\link[Biostrings]{readAAStringSet}}. Several built-in date patterns
are provided for the common date conventions used in NCBI and GISAID
exports; a fully custom regex can also be supplied.
}
\details{
Date strings of the form \code{yyyy-mm-dd} are matched between pipe
characters (\code{|...|}) by default. Day value \code{00} (a common
GISAID convention indicating unknown collection day) is accepted in the
raw string and corrected to \code{01} before coercion to \code{Date}.
Both raw and corrected versions are returned, so the caller can decide
how to treat unknown-day records downstream.

\strong{Choosing a built-in pattern.} The four options correspond to
the four most common date conventions in viral sequence repositories:
\itemize{
\item \code{option = 1}: \code{yyyy-mm-dd} between pipes — GISAID
export format, where the date is followed by additional
pipe-delimited fields.
\item \code{option = 2}: \code{yyyy-dd-mm} between pipes — some
European data sources reverse day and month.
\item \code{option = 3}: \code{yyyy-mm} between pipes — month-level
resolution, useful when the source omits or hides the day. Pair
with \code{date_format = "\%Y-\%m"}.
\item \code{option = 4}: \code{yyyy-mm-dd} at end of header — NCBI
Virus export format, where the collection date is the final
field with no trailing delimiter. This is the format of the
bundled \code{sarscov2_sample}.
}
For datasets where the date does not lie between pipes, supply a
\code{custom_pattern} matching whatever surrounding context the headers
provide.

\strong{Coercion to Date.} When \code{date_format = "\%Y-\%m"} the
function uses \code{\link[zoo]{as.yearmon}} for coercion so that
year-month strings are handled correctly (base \code{as.Date} cannot
parse \code{"2021-05"} alone). For all other formats, base
\code{as.Date} with the supplied \code{date_format} is used.

\strong{Output alignment.} All six elements of the return list are the
same length as the input \code{sequence}. Where extraction fails for a
record, the corresponding entries are \code{NA}; \code{missing_id}
lists the affected indices.
}
\examples{
\donttest{
path_sample  <- system.file("extdata", "sarscov2_sample.fasta.gz",
                             package = "ViralEntropR")
fasta_sample <- Biostrings::readAAStringSet(path_sample)

# Inspect header structure to confirm date field position.
sample(names(fasta_sample), 1)
# The bundled sample uses NCBI Virus format: date is at end of header.

# Default usage on bundled sample: option = 4 for end-of-header dates.
dates <- extract_fasta_dates(fasta_sample, option = 4)
dates$message
head(dates$corrected_dates)
range(dates$corrected_dates, na.rm = TRUE)

# Custom regex for non-standard headers:
dates_custom <- extract_fasta_dates(
  fasta_sample,
  custom_pattern = "[0-9]{4}-(0?[1-9]|1[0-2])-(0?[1-9]|[12][0-9]|3[01]|00)"
)
}
}
\seealso{
\code{\link{extract_fasta_countries}} for the country-extraction
companion; \code{\link[Biostrings]{readAAStringSet}} for loading the
input \code{AAStringSet}; \code{\link[zoo]{as.yearmon}} for the
year-month coercion path.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_site_counts.R
\name{get_site_counts}
\alias{get_site_counts}
\title{Get Site Counts Matrix (Internal)}
\usage{
get_site_counts(partitions, site_index, alphabet_size)
}
\arguments{
\item{partitions}{List of data frames, one per time window. Each must
have integer-encoded amino acid columns.}

\item{site_index}{Integer. Index of the site (column) to count.}

\item{alphabet_size}{Integer. Size of the amino acid alphabet
(typically \code{25} for the package's default encoding).}
}
\value{
Integer matrix with \code{alphabet_size} rows (one per amino
acid code, 1 through \code{alphabet_size}) and one column per
partition. Each cell is the count of that amino acid at
\code{site_index} in that partition.
}
\description{
Extract and tabulate amino acid counts for a specific site
across all partitions.
}
\details{
Built on \code{\link[base]{tabulate}} for fast integer counting.
Consumed internally by \code{\link{calculate_hellinger_matrix}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decode_aa_sequence.R
\name{decode_aa_sequence}
\alias{decode_aa_sequence}
\title{Decode Amino Acid Sequences}
\usage{
decode_aa_sequence(matrix_input)
}
\arguments{
\item{matrix_input}{Numeric matrix of integer-encoded amino acids,
typically the output of \code{\link{encode_aa_sequence}}. Values
outside \code{1:25} (including \code{0} and \code{NA}) decode to
the sentinel string \code{"0"}. A non-matrix input is coerced via
\code{\link[base]{as.matrix}}.}
}
\value{
A character matrix of the same dimensions as \code{matrix_input},
with the same \code{dimnames}. Each cell is either a one-character
amino acid code from the 25-symbol alphabet, or the sentinel
\code{"0"} for out-of-range and missing values.
}
\description{
Converts an integer-encoded matrix of amino acids back to its
character representation under the package's 25-symbol alphabet. Inverse
of \code{\link{encode_aa_sequence}}.
}
\details{
Decoding is a single vectorised lookup against the fixed alphabet
(\code{A, R, N, D, C, Q, E, G, H, I, L, K, M, F, P, S, T, W, Y, V} for
the twenty standard residues, then \code{B, Z, X, *, -} for ambiguous
codes, stop codons, and gaps). The input matrix is flattened, indexed
against the alphabet vector in one operation, and reshaped — there is
no per-row or per-element loop.

Values outside the valid range \code{1:25} (including \code{0}, which
\code{\link{encode_aa_sequence}} produces for unrecognised characters)
are returned as the sentinel string \code{"0"}. This preserves
round-trip consistency with the encoder: encoding then decoding any
character originally outside the alphabet yields \code{"0"} rather
than throwing an error. \code{NA} values are also mapped to
\code{"0"}.

Row and column names of the input matrix are preserved on the output.
}
\examples{
# 1. Decode a numeric matrix.
num_mat = matrix(c(1, 2, 25, 10), nrow = 2, byrow = TRUE)
decoded = decode_aa_sequence(num_mat)
print(decoded)
# 1 -> "A", 2 -> "R", 25 -> "-", 10 -> "I"

# 2. Round-trip consistency check (excluding unknowns).
orig = matrix(c("A", "C", "W", "G"), nrow = 2)
enc  = encode_aa_sequence(orig)
dec  = decode_aa_sequence(enc)
all.equal(orig, dec)

# 3. Out-of-range and NA values both decode to the sentinel "0".
decode_aa_sequence(matrix(c(0, NA, 30, 5), nrow = 2))
}
\seealso{
\code{\link{encode_aa_sequence}} for the inverse operation;
\code{\link{fasta_to_char_matrix}} for the FASTA-to-character-matrix
step that typically precedes encoding.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{sarscov2_sample}
\alias{sarscov2_sample}
\title{SARS-CoV-2 Surface Glycoprotein Sequences -- NCBI Demo Sample}
\format{
A gzip-compressed FASTA file (\code{.fasta.gz}) readable by
\code{\link[Biostrings]{readAAStringSet}}. Each record contains:
\describe{
\item{Header}{NCBI Virus format: accession, country, and collection date
metadata separated by \code{|}.}
\item{Sequence}{Amino acid sequence using the standard IUPAC one-letter
code. Ambiguous residues (B, X, Z) and gaps (-) may be present and
can be removed with \code{\link{filter_ambiguous_sequences}}.}
}
}
\source{
\url{https://www.ncbi.nlm.nih.gov/labs/virus/vssi/}
}
\description{
A compressed FASTA file containing a random sample of size 100 of SARS-CoV-2
surface glycoprotein (Spike protein) amino acid sequences, drawn from a
dataset of 137,132 complete sequences downloaded from the NCBI SARS-CoV-2
Data Hub on October 12, 2021. Intended to demonstrate the full
\code{ViralEntropR} pipeline on real-world surveillance data.
}
\details{
The file is accessed at runtime via:
\preformatted{
path  <- system.file("extdata", "sarscov2_sample.fasta.gz",
                     package = "ViralEntropR")
fasta <- Biostrings::readAAStringSet(path)
}

\strong{Sample file contents:}
\itemize{
\item \strong{Sample size}: 100 random sequences
\item \strong{Protein}: Surface glycoprotein (Spike, S protein)
\item \strong{Organism}: \emph{Severe acute respiratory syndrome
coronavirus 2} (SARS-CoV-2), NCBI Taxonomy ID: 2697049
\item \strong{Completeness}: Complete sequences only
\item \strong{Format}: Gzip-compressed FASTA (\code{.fasta.gz}),
readable directly by \code{\link[Biostrings]{readAAStringSet}}
}

\strong{Full dataset:}
The complete 137,132-sequence dataset (~181.5 MB, uncompressed FASTA) is
archived on Zenodo (DOI: \doi{10.5281/zenodo.19040165});
\code{\link[Biostrings]{readAAStringSet}} reads it directly.

\strong{Header format:}
Sequence headers follow the NCBI Virus export format. Dates and countries
can be extracted directly using the package helper functions:
\preformatted{
dates     <- extract_fasta_dates(fasta, option = 1)
countries <- extract_fasta_countries(fasta, position = 2)
}

\strong{Subsampling:}
To keep the package within CRAN size limits, the bundled sample was
generated by random sampling of 100 sequences. The full provenance and
reproducible sampling script are in \code{data-raw/sarscov2_ncbi.R} in
the package source.
}
\section{Provenance}{

Downloaded from the NCBI SARS-CoV-2 Data Hub
(\url{https://www.ncbi.nlm.nih.gov/labs/virus/vssi/}) on
\strong{October 12, 2021} with the following filters:
\enumerate{
\item Virus: Severe acute respiratory syndrome coronavirus 2
(SARS-CoV-2), taxid: 2697049
\item Nucleotide completeness: complete
\item Protein: surface glycoprotein
}
Result: \strong{n = 137,132 sequences}, file size ~181.5 MB (uncompressed).
}

\section{License}{

NCBI sequence data is a US Government work and is in the public domain
within the United States. Data from international contributors is subject
to the INSDC open-access policy (Karsch-Mizrachi et al., 2025;
\doi{10.1093/nar/gkae1058}). The compiled dataset is released under
CC0 1.0 Universal. Individual GenBank accession numbers in the FASTA
headers provide full traceability to original submissions.
}

\examples{
\donttest{
path  <- system.file("extdata", "sarscov2_sample.fasta.gz",
                     package = "ViralEntropR")
fasta <- Biostrings::readAAStringSet(path)

# Number of sequences in the demo sample
length(fasta)

# Inspect first 3 headers
names(fasta)[1:3]

# Extract collection dates
dates <- extract_fasta_dates(fasta, option = 1)
head(dates$corrected_dates)

# Extract countries
countries <- extract_fasta_countries(fasta, position = 2)
table(countries$countries)

# Convert to character matrix for pipeline entry
char_mat <- fasta_to_char_matrix(fasta)
dim(char_mat)

# The full 137,132-sequence dataset (~181.5 MB) is available on Zenodo:
# https://zenodo.org/records/19040165
# Download sequences.fasta manually and load with:
# fasta_full <- Biostrings::readAAStringSet("path/to/sequences.fasta")
}
}
\references{
Clark K, Karsch-Mizrachi I, Lipman DJ, Ostell J, Sayers EW (2016).
GenBank.
\emph{Nucleic Acids Research}, 44(D1), D67--D72.
\doi{10.1093/nar/gkv1276}

Sayers EW, Bolton EE, Brister JR, et al. (2022).
Database resources of the National Center for Biotechnology Information.
\emph{Nucleic Acids Research}, 50(D1), D20--D26.
\doi{10.1093/nar/gkab1112}

NCBI Virus [Internet]. Bethesda (MD): National Library of Medicine (US),
National Center for Biotechnology Information; [2020] --
[cited 2021 Oct 12]. Available from:
\url{https://www.ncbi.nlm.nih.gov/labs/virus/vssi/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build_distance_matrix.R
\name{build_distance_matrix}
\alias{build_distance_matrix}
\title{Build Distance/Entropy Matrix}
\usage{
build_distance_matrix(data_input)
}
\arguments{
\item{data_input}{Either:
\itemize{
\item A numeric matrix (\code{sites × time_steps}) as returned by
\code{\link{calculate_hellinger_matrix}}.
\item A named list with elements \code{Sites} and one of
\code{Hellinger_Distances} or \code{Entropies}, as returned by the
legacy \code{get_hellinger_dist} function.
}}
}
\value{
A numeric matrix with:
\item{Rows}{Sites (1 to \code{max(sites)}, sparse if sites are
non-contiguous).}
\item{Columns}{Time steps (partitions T2, T3, … or entropy periods).}
Row names are character site indices. Column names are preserved from
the input where available.
}
\description{
Converts the output of \code{\link{calculate_hellinger_matrix}}
or the legacy \code{get_hellinger_dist} / entropy list into a numeric
\code{sites × time_steps} matrix suitable for change point detection.
}
\details{
In the current pipeline \code{\link{calculate_hellinger_matrix}} already
returns a \code{sites × time_steps} matrix, so calling this function is
usually unnecessary — transposing that matrix directly gives \code{dat_t}
for \code{\link[ecp]{e.agglo}} or \code{\link[ecp]{ks.cp3o}}:

\preformatted{
  hell_mat = calculate_hellinger_matrix(partitions, sites = seq_len(n_sites))
  dat_t    = t(hell_mat)   # time_steps × sites — ready for ECP
}

This function is retained for \strong{backward compatibility} with code that
was written against the old \code{get_hellinger_dist} list format, and as a
convenience wrapper that also accepts entropy lists.

\strong{Input formats accepted:}
\enumerate{
\item A numeric matrix (e.g. from \code{calculate_hellinger_matrix}) —
returned as-is with site rownames normalised.
\item The legacy named list from \code{get_hellinger_dist} containing
\code{$Sites} and \code{$Hellinger_Distances} (list of vectors).
\item An entropy list containing \code{$Sites} and \code{$Entropies}
(list of numeric vectors), used when no Hellinger distances are present.
}
}
\seealso{
\code{\link{calculate_hellinger_matrix}},
\code{\link{detect_changepoints_ecp}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_entropy.R
\name{calculate_entropy}
\alias{calculate_entropy}
\title{Calculate Shannon Entropy}
\usage{
calculate_entropy(vctr, base = 2, precision = 6)
}
\arguments{
\item{vctr}{A vector (character, factor, or integer) representing categorical data.}

\item{base}{A numeric scalar. The base of the logarithm. Default is 2.}

\item{precision}{Integer. The number of decimal places to round the result to. Default is 6.}
}
\value{
A numeric scalar representing the entropy. Returns 0 if the vector contains
only one unique value or has length 0.
}
\description{
Computes the Shannon entropy of a categorical vector.
}
\details{
Entropy is calculated as \eqn{H(X) = -\sum p(x) \log_b p(x)}, where \eqn{p(x)}
is the proportion of observations belonging to category \eqn{x}.
}
\examples{
seq_vec = c("A", "A", "T", "G", "C", "A")
calculate_entropy(seq_vec)

# Pure homogeneity
calculate_entropy(rep("A", 10))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/detect_changepoints_hdcp.R
\name{detect_changepoints_hdcp}
\alias{detect_changepoints_hdcp}
\title{Detect Temporal Change Points (HDcpDetect)}
\usage{
detect_changepoints_hdcp(
  data_matrix,
  min_window,
  max_window,
  n_timesteps,
  rolling_window = FALSE,
  wild = FALSE,
  ...
)
}
\arguments{
\item{data_matrix}{A numeric matrix (Time x Features). Rows are time
points, columns are features (e.g. Hellinger distances per site).}

\item{min_window}{Integer. Starting row index of the initial window.}

\item{max_window}{Integer. Ending row index of the initial window. Must
not exceed \code{nrow(data_matrix)}.}

\item{n_timesteps}{Integer >= 0. Number of \emph{additional} detections
to run after the initial window. The function performs
\code{n_timesteps + 1} detections in total. Set to \code{0} for a
single detection on the initial window.}

\item{rolling_window}{Logical. If \code{TRUE}, both window endpoints
advance per step (sliding window of constant length). If \code{FALSE}
(default), only the end advances (expanding window).}

\item{wild}{Logical. If \code{TRUE}, uses
\code{\link[HDcpDetect]{wild.binary.segmentation}}; if \code{FALSE}
(default), uses \code{\link[HDcpDetect]{binary.segmentation}}. See
Details for the methodological distinction.}

\item{...}{Additional arguments passed to the chosen segmentation
function. Note that \code{M} (number of random intervals) is only
accepted by \code{wild.binary.segmentation}; passing it with
\code{wild = FALSE} produces an "unused argument" error.}
}
\value{
A named list:
\item{Points_List}{List of integer vectors \code{c(start, end)} giving
the row index window used at each step.}
\item{HDcp_list}{List of full segmentation result objects, one per
step. Steps where \code{end_idx} exceeds the matrix rows return
\code{NULL} with a warning.}
}
\description{
Runs high-dimensional change point detection on time-series
matrices (typically Hellinger distance matrices) over one or more time
windows, using either binary segmentation or wild binary segmentation.
}
\details{
The function applies \code{\link[HDcpDetect]{binary.segmentation}} or
\code{\link[HDcpDetect]{wild.binary.segmentation}} repeatedly to slices of
\code{data_matrix}, advancing the slice forward by one row at each step.
A total of \code{n_timesteps + 1} detections are performed: one on the
initial window \code{[min_window, max_window]}, then \code{n_timesteps}
additional detections on subsequent windows.

Two window-advancement modes are supported, mirroring
\code{\link{detect_changepoints_ecp}}:
\itemize{
\item \strong{Expanding} (\code{rolling_window = FALSE}, default): the
start index is fixed at \code{min_window}; the end index grows by 1
each step. Each iteration uses one more row than the previous.
Natural for online surveillance, where change-point detection is
re-run as new data accumulates.
\item \strong{Rolling} (\code{rolling_window = TRUE}): both start and
end indices advance by 1 each step, keeping the window length
constant. Natural for retrospective windowed analysis.
}

\strong{Choice of segmentation method.} Binary segmentation
(\code{wild = FALSE}, default) is the classical recursive method: it
finds the most likely change point, splits the series, and recurses.
Wild binary segmentation (\code{wild = TRUE}) is the variant of
Fryzlewicz (2014) that draws random subintervals to improve detection
of multiple closely-spaced change points; it accepts an additional
\code{M} argument controlling the number of random intervals.
}
\section{Warnings and errors}{

\describe{
\item{\code{stop()}}{Triggered if \code{max_window} exceeds
\code{nrow(data_matrix)}: the initial window cannot be
constructed and no detection can run.}
\item{\code{warning()}}{Triggered (and execution continues) if
\code{max_window + n_timesteps} exceeds \code{nrow(data_matrix)},
in which case the offending later iterations are skipped and
\code{NULL} entries appear in \code{HDcp_list}.}
}
}

\examples{
set.seed(42)
baseline = matrix(rnorm(50, mean = 0, sd = 0.1), nrow = 10, ncol = 5)
variant  = matrix(rnorm(50, mean = 3, sd = 0.1), nrow = 10, ncol = 5)
data_mat = rbind(baseline, variant)

# Single-window detection with the default binary segmentation.
res = detect_changepoints_hdcp(
  data_matrix = data_mat,
  min_window  = 1,
  max_window  = 20,
  n_timesteps = 0
)
print(res$HDcp_list[[1]])

# Wild binary segmentation with a custom number of random intervals.
res_wild = detect_changepoints_hdcp(
  data_matrix = data_mat,
  min_window  = 1,
  max_window  = 20,
  n_timesteps = 0,
  wild        = TRUE,
  M           = 100
)
print(res_wild$HDcp_list[[1]])
}
\seealso{
\code{\link{detect_changepoints_ecp}} for the energy-statistic
alternative; \code{\link{calculate_hellinger_matrix}} for the typical
upstream input.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{sarscov2_variants}
\alias{sarscov2_variants}
\title{SARS-CoV-2 VOC/VOI Curated Variant Metadata}
\format{
A named list of 13 elements (12 metadata fields plus a structured
References object). See Details for per-element descriptions.
}
\source{
Compiled from 21 peer-reviewed and surveillance sources; see
\code{sarscov2_variants$References$data} for the full reference table
with DOIs and URLs. Built from \code{SARS_CoV_2_VOC_VOI.xlsx} via
\code{\link{get_variants}}.
}
\usage{
data(sarscov2_variants)
}
\description{
A pre-built named list containing curated biological and epidemiological
metadata for 12 SARS-CoV-2 Variants of Concern (VOC) and Variants of
Interest (VOI): Alpha, Beta, Delta, Epsilon, Eta, Gamma, Iota, Kappa, Lambda,
Omicron, Theta, Zeta. Includes mutation profiles, nomenclature,
temporal detection records, defining SNPs, and a fully citable reference
table with 21 verified references.
}
\details{
The object is available automatically after \code{library(ViralEntropR)}
via lazy loading. It is produced by \code{\link{get_variants}} from the
curated Excel workbook \code{SARS_CoV_2_VOC_VOI.xlsx} (not bundled with
the package; see \code{data-raw/sarscov2_variants.R} for the reproducible
build script).

\strong{List elements:}
\describe{
\item{\code{WHO_Label}}{List of 12. WHO variant label strings
(e.g. \code{"Alpha"}, \code{"Omicron"}).}
\item{\code{Pango_Lineage}}{List of 12. Pango lineage designations
(e.g. \code{"B.1.1.7"}).}
\item{\code{GISAID_Clade}}{Character(12). GISAID clade strings
(e.g. \code{"GRY"}, \code{"GR/484A"}).}
\item{\code{Nextstrain_Clade}}{Character(12). Nextstrain clade strings
(e.g. \code{"20I/501Y.V1"}).}
\item{\code{Country_First_Detected}}{List of 12. Country of first
documented detection per variant.}
\item{\code{Date_Earliest_Sample}}{Character(12). Month-Year of earliest
documented sample (e.g. \code{"Sep-2020"}).}
\item{\code{Date_First_Detected}}{Character(12). Month-Year of
world-level first detection.}
\item{\code{Date_First_Detected_US}}{Character(12). Month-Year of first
US detection.}
\item{\code{Spike_Mutations}}{List of 12. Character vectors of Spike
protein mutations per variant, read from the Excel workbook.}
\item{\code{Mutation_Sites}}{List of 12. Integer vectors of amino acid
positions extracted from \code{Spike_Mutations}.}
\item{\code{Defining_SNPs}}{List of 12. Canonical defining SNP strings
(\code{NA} for variants not characterised in the reference set).}
\item{\code{Defining_SNP_Sites}}{List of 12. Integer positions extracted
from \code{Defining_SNPs}.}
\item{\code{References}}{Named list with three elements:
\code{$data} -- data frame of 21 verified references (columns: ID,
Authors, Year, Title, Journal_Source, Volume_Issue_Pages, DOI, URL,
Type, Variants_Covered, Data_Field, Citation);
\code{$display(variant = NULL)} -- renders an interactive
\code{\link[DT]{datatable}}, optionally filtered by WHO label
(requires the \pkg{DT} package);
\code{$cite(variant)} -- returns formatted citation strings for a
given WHO label, suitable for manuscript use.}
}
}
\section{References}{

The 21 curated references cover peer-reviewed articles, CDC MMWR government
reports, and outbreak.info surveillance database records. Full provenance
is available via \code{sarscov2_variants$References$data} or the
interactive display.
}

\examples{
# Object is available immediately after library(ViralEntropR)

# --- Basic access ---------------------------------------------------------

# All 12 WHO labels
unlist(sarscov2_variants$WHO_Label)

# Pango lineage for Alpha
sarscov2_variants$Pango_Lineage[[
  which(unlist(sarscov2_variants$WHO_Label) == "Alpha")
]]

# World detection dates for all variants
data.frame(
  Variant  = unlist(sarscov2_variants$WHO_Label),
  Detected = sarscov2_variants$Date_First_Detected,
  Country  = unlist(sarscov2_variants$Country_First_Detected)
)

# --- Mutation sites -------------------------------------------------------

# Defining SNPs and sites for Delta
idx <- which(unlist(sarscov2_variants$WHO_Label) == "Delta")
sarscov2_variants$Defining_SNPs[[idx]]
sarscov2_variants$Defining_SNP_Sites[[idx]]

# All Spike mutation sites for Omicron
sarscov2_variants$Mutation_Sites[[
  which(unlist(sarscov2_variants$WHO_Label) == "Omicron")
]]

# --- Nomenclature ---------------------------------------------------------

data.frame(
  WHO        = unlist(sarscov2_variants$WHO_Label),
  PANGO      = unlist(sarscov2_variants$Pango_Lineage),
  GISAID     = sarscov2_variants$GISAID_Clade,
  Nextstrain = sarscov2_variants$Nextstrain_Clade
)

# --- References -----------------------------------------------------------

# Full reference data frame (21 rows)
sarscov2_variants$References$data[,
  c("ID", "Authors", "Year", "Journal_Source")
]

# Formatted citation strings for Gamma
sarscov2_variants$References$cite("Gamma")

if (interactive() && requireNamespace("DT", quietly = TRUE)) {
  # Interactive DT table of all references
  sarscov2_variants$References$display()

  # Filtered to Omicron references only
  sarscov2_variants$References$display(variant = "Omicron")
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/detect_changepoints_ecp.R
\name{detect_changepoints_ecp}
\alias{detect_changepoints_ecp}
\title{Detect Temporal Change Points (ECP)}
\usage{
detect_changepoints_ecp(
  data_matrix,
  min_window,
  max_window,
  n_timesteps,
  rolling_window = FALSE,
  dynamic_k = FALSE,
  ...
)
}
\arguments{
\item{data_matrix}{A numeric matrix (Time x Features). Rows are time
points, columns are features (e.g. Hellinger distances per site).}

\item{min_window}{Integer. Starting row index of the initial window.}

\item{max_window}{Integer. Ending row index of the initial window. Must
not exceed \code{nrow(data_matrix)}.}

\item{n_timesteps}{Integer >= 0. Number of \emph{additional} detections
to run after the initial window. The function performs
\code{n_timesteps + 1} detections in total. Set to \code{0} for a
single detection on the initial window.}

\item{rolling_window}{Logical. If \code{TRUE}, both window endpoints
advance per step (sliding window of constant length). If \code{FALSE}
(default), only the end advances (expanding window).}

\item{dynamic_k}{Logical. If \code{TRUE}, sets \code{K} to
\code{nrow(subset) - 2L} at each step, so the algorithm considers
the maximum number of change points permitted by the window length
(subject to \code{minsize}). Useful for exploratory analysis where
the true number of change points is unknown a priori. Most idiomatic
when combined with \code{n_timesteps > 0}, where the maximum
admissible K legitimately scales with window size, but also valid
for single-window exploratory use. If \code{FALSE} (default), the
value of \code{K} supplied via \code{...} is honoured at every step;
if no \code{K} is supplied, the \code{ks.cp3o} default of 1 is used.}

\item{...}{Additional arguments passed to \code{\link[ecp]{ks.cp3o}}.
Common choices include \code{K} (maximum number of change points,
default 1) and \code{minsize} (minimum segment size, default 30).
Note that \code{K} is overridden when \code{dynamic_k = TRUE}; a
warning is emitted in that case.}
}
\value{
A named list:
\item{Points_List}{List of integer vectors \code{c(start, end)} giving
the row index window used at each step.}
\item{ECP_list}{List of full \code{ks.cp3o} result objects, one per
step. Inspect \code{$cpLoc} for the optimal change-point locations
at each candidate count, and \code{$gofM} for the goodness-of-fit
curve, to apply post-hoc filtering.}
\item{ECP_est_list}{List of change-point estimate vectors (the
algorithm's own selection, equivalent to \code{$estimates}).}
}
\description{
Runs Energy Change Point detection on time-series matrices
(typically Hellinger distance matrices) over one or more time windows.
}
\details{
The function applies \code{\link[ecp]{ks.cp3o}} repeatedly to slices of
\code{data_matrix}, advancing the slice forward by one row at each step.
A total of \code{n_timesteps + 1} detections are performed: one on the
initial window \code{[min_window, max_window]}, then \code{n_timesteps}
additional detections on subsequent windows.

Two window-advancement modes are supported:
\itemize{
\item \strong{Expanding} (\code{rolling_window = FALSE}, default): the
start index is fixed at \code{min_window}; the end index grows by 1
each step. Each iteration uses one more row than the previous.
Natural for online surveillance, where change-point detection is
re-run as new data accumulates.
\item \strong{Rolling} (\code{rolling_window = TRUE}): both start and
end indices advance by 1 each step, keeping the window length
constant. Natural for retrospective windowed analysis, where local
change-point structure is examined in fixed-size segments.
}
}
\section{Warnings and errors}{

\describe{
\item{\code{stop()}}{Triggered if \code{max_window} exceeds
\code{nrow(data_matrix)}: the initial window cannot be
constructed and no detection can run.}
\item{\code{warning()}}{Triggered (and execution continues) if
(a) \code{max_window + n_timesteps} exceeds
\code{nrow(data_matrix)}, in which case the offending later
iterations are skipped and \code{NULL} entries appear in the
result lists; (b) \code{dynamic_k = TRUE} and \code{K} is also
supplied via \code{...}, in which case the user-supplied
\code{K} is silently overridden by \code{nrow(subset) - 2L}.}
}
}

\examples{
set.seed(123)
baseline = matrix(rnorm(50, mean = 0, sd = 0.1), nrow = 10, ncol = 5)
variant  = matrix(rnorm(50, mean = 3, sd = 0.1), nrow = 10, ncol = 5)
data_mat = rbind(baseline, variant)

# Single-window detection: one true change point at row 11.
res = detect_changepoints_ecp(
  data_matrix = data_mat,
  min_window  = 1,
  max_window  = 20,
  n_timesteps = 0,
  minsize     = 5
)
print(res$ECP_est_list[[1]])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_entropy_trajectories.R
\name{plot_entropy_trajectories}
\alias{plot_entropy_trajectories}
\title{Plot Shannon Entropy Trajectories}
\usage{
plot_entropy_trajectories(
  part_data,
  sites = NULL,
  labels = NULL,
  site_colors = NULL,
  by_group = FALSE,
  groups_list = NULL,
  line_type_groups = NULL,
  line_size_groups = NULL,
  transformation = NULL,
  line_size = 1.5,
  legend = TRUE,
  legend_text_size = 12,
  x_angle = 45,
  grayscale = FALSE,
  plot_title = "Shannon Entropy Trajectories"
)
}
\arguments{
\item{part_data}{Named list.  Output of \code{\link{partition_time_windows}},
optionally with per-partition \code{Clusters[[i]]$DataFrame} already
relabeled by the user via \code{\link{relabel_entropy_classes}}.  Must
contain elements \code{Clusters}, \code{Max_Entropy}, \code{Dates_Labels},
and \code{N_partitions}.}

\item{sites}{Integer vector.  Site indices to include.  Defaults to the
union of all sites observed across all partitions (i.e. every site that
has non-zero, non-singleton entropy in at least one partition window).}

\item{labels}{Character vector of length \code{N_partitions}.  Partition
labels used on the x-axis.  Defaults to \code{part_data$Dates_Labels}.}

\item{site_colors}{Named character vector.  Names are site indices as
character strings (e.g. \code{"681"}); values are colour strings
(e.g. \code{"#FB8072"}).  Sites absent from \code{site_colors} receive
automatically assigned colours.  Default is \code{NULL} (all colours
auto-assigned).}

\item{by_group}{Logical.  If \code{TRUE}, maps line type and line width to
site groups defined by \code{groups_list}.  Default is \code{FALSE}.}

\item{groups_list}{List of integer vectors.  Each element specifies the
site indices belonging to one explicit group.  Sites in \code{sites} not
covered by any explicit group are automatically assigned to a remainder
group appended as the final element.  Total group count (explicit plus
remainder) must not exceed 6.  Required when \code{by_group = TRUE}.}

\item{line_type_groups}{Character vector.  One line-type string per group
(in order, including the automatic remainder group).  Must have length
equal to the total number of groups.  Defaults to \code{"solid"} for the
first group and \code{"dashed"} for all remaining groups.}

\item{line_size_groups}{Numeric vector.  One line-width value per group
(in order, including the automatic remainder group).  Must have length
equal to the total number of groups.  Defaults to \code{2} for the first
group and \code{1} for all remaining groups.}

\item{transformation}{Object of class \code{"transform"} or \code{"trans"}
as returned by \code{\link[scales]{trans_new}}, or \code{NULL} (identity,
no transformation).  Applied to the y-axis via
\code{\link[ggplot2]{scale_y_continuous}}.  Default is \code{NULL}.}

\item{line_size}{Numeric.  Line width used when \code{by_group = FALSE}.
Default is \code{1.5}.}

\item{legend}{Logical.  If \code{TRUE} (default), the site colour legend is
displayed.}

\item{legend_text_size}{Numeric.  Font size of legend text in points.
Default is \code{12}.}

\item{x_angle}{Numeric.  Rotation angle of x-axis tick labels in degrees.
Default is \code{45}.}

\item{grayscale}{Logical.  If \code{TRUE}, overrides \code{site_colors} and
renders all trajectories in greyscale.  Default is \code{FALSE}.}

\item{plot_title}{Character.  Plot title string.  Default is
\code{"Shannon Entropy Trajectories"}.}
}
\value{
A named list with five elements:
\item{Data_Frame}{Long-format data frame with columns \code{sites}
(factor), \code{entropies} (numeric), \code{class} (factor),
\code{max_class} (integer), \code{period} (integer), and
\code{coverage} (character, partition label).  Suitable for direct
input to \code{\link{plot_site_class_trajectory}}.}
\item{Plot}{A \code{ggplot} object.  Augment with additional layers
(e.g. \code{\link[ggplot2]{geom_vline}} for VOC emergence events) before
printing or saving with \code{\link[ggplot2]{ggsave}}.}
\item{Colors}{Named character vector mapping each plotted site index
(character) to its assigned colour string.  Pass as \code{site_colors}
to subsequent calls to \code{plot_entropy_trajectories} for a consistent
colour scheme across figures.}
\item{XBreaks}{Integer vector of partition period indices.  Pass as
\code{xbreaks} to \code{\link{plot_site_class_trajectory}}.}
\item{XLabels}{Character vector of partition labels aligned with
\code{XBreaks}.  Pass as \code{xlabels} to
\code{\link{plot_site_class_trajectory}}.}
}
\description{
Plots per-site Shannon entropy as continuous trajectories across
time partitions for a selected set of sequence sites, using the output of
\code{\link{partition_time_windows}}.
}
\details{
For each partition the function extracts the GMM clustering result from
\code{part_data$Clusters} and assembles a long-format data frame spanning
all selected sites across all partitions.  Sites absent from a given
partition (removed by zero-entropy or singleton filtering, or because the
partition window was empty) are silently omitted from that partition's
trajectory and do not interrupt adjacent observations.

\strong{Class relabeling.}  This function does not perform any relabeling of
GMM class labels.  If class 1 must denote the highest-entropy group
throughout the returned \code{$Data_Frame} (e.g. before passing it to
\code{\link{plot_site_class_trajectory}}), the user should call
\code{\link{relabel_entropy_classes}} on each partition's
\code{Clusters[[i]]$DataFrame} and update \code{Max_Entropy[i]} to
\code{1L} prior to calling this function.

\strong{Colour scheme.}  Site colours are specified through
\code{site_colors}, a named character vector whose names are site indices
(as character strings) and whose values are valid R colour strings.  Any
site not listed in \code{site_colors} receives an automatically assigned
colour from the HCL \code{"Dark 2"} qualitative palette.  The final colour
mapping is returned as \code{$Colors} so the same scheme can be passed to
subsequent calls for cross-plot consistency.

\strong{Group-stratified trajectories (\code{by_group = TRUE}).}  When
biological groupings must be distinguished visually (e.g. defining SNP
sites vs. other mutation sites), \code{groups_list} partitions
\code{sites} into explicitly named groups.  Any site not assigned to an
explicit group is automatically collected into a remainder group appended
as the final element of \code{groups_list}.  Line type and line width are
mapped to group membership via \code{line_type_groups} and
\code{line_size_groups}, both of which must have length equal to the
total number of groups (explicit plus the automatic remainder).  At most
six groups are supported.

\strong{\code{max_class} column.}  The returned \code{$Data_Frame} carries
a \code{max_class} column recording the label of the highest-entropy GMM
component for each partition, taken directly from
\code{part_data$Max_Entropy[i]}.  This column is consumed by
\code{\link{plot_site_class_trajectory}} (red labels) and by downstream
class-assignment tables.
}
\examples{
# Three-period synthetic dataset with distinct trajectory shapes:
# Site 1: up-then-down (peak in P2).
# Site 2: monotone up.
# Per-partition Shannon entropy (bits):
#   P1 (Jan)  s1: 0.469  s2: 0.469
#   P2 (Feb)  s1: 1.522  s2: 0.881
#   P3 (Mar)  s1: 0.722  s2: 1.522
df <- data.frame(
  s1 = c(
    c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L),         # P1: 9:1
    c(1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 3L, 3L),         # P2: 4:4:2
    c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L)          # P3: 8:2
  ),
  s2 = c(
    c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L),         # P1: 9:1
    c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L),         # P2: 7:3
    c(1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L, 3L, 3L)          # P3: 4:4:2
  ),
  Date = rep(
    seq(as.Date("2020-01-01"), by = "month", length.out = 3L),
    each = 10L
  )
)

part_data <- partition_time_windows(
  data          = df,
  n_sites       = 2L,
  window_length = 1L,
  window_type   = 3L,
  start_date    = "2020-01-01",
  end_date      = "2020-04-01",
  removez       = FALSE,
  removesngl    = FALSE
)

result <- plot_entropy_trajectories(part_data = part_data)
print(result$Plot)
}
\seealso{
\code{\link{partition_time_windows}},
\code{\link{relabel_entropy_classes}},
\code{\link{plot_site_class_trajectory}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ViralEntropR-package.R
\docType{package}
\name{ViralEntropR-package}
\alias{ViralEntropR}
\alias{ViralEntropR-package}
\title{ViralEntropR: A Computational Pipeline for Entropy-Informed Detection of Emerging Viral Variants}
\description{
A computational pipeline for detecting emerging variants in viral amino
acid sequence data, combining per-site Shannon entropy, Gaussian mixture
model site selection, Gower-distance Partitioning Around Medoids
clustering, Hellinger-distance quantification of distributional shifts,
and multivariate non-parametric change-point detection.
}
\section{Pipeline overview}{

The package supports a four-stage workflow:
\itemize{
\item \strong{Preprocessing.} Parse FASTA headers, filter ambiguous
residues, and convert between integer and character representations
of amino acid sequences under a 25-symbol alphabet. See
\code{\link{extract_fasta_dates}},
\code{\link{extract_fasta_countries}},
\code{\link{fasta_to_char_matrix}},
\code{\link{filter_ambiguous_sequences}},
\code{\link{encode_aa_sequence}}, and
\code{\link{decode_aa_sequence}}.
\item \strong{Site selection.} Compute per-site Shannon entropy across
temporal partitions and cluster sites by entropy via Gaussian
mixture models. See \code{\link{calculate_entropy}},
\code{\link{partition_time_windows}},
\code{\link{cluster_sites_by_entropy}}, and
\code{\link{relabel_entropy_classes}}.
\item \strong{Distributional analysis.} Quantify residue-composition
shifts between time windows using the Hellinger distance. See
\code{\link{calculate_hellinger_matrix}}.
\item \strong{Change-point detection.} Identify temporal change points
non-parametrically using energy statistics or wild binary
segmentation. See \code{\link{detect_changepoints_ecp}} and
\code{\link{detect_changepoints_hdcp}}.
}
}

\section{Visualisation and tabulation}{

\itemize{
\item \code{\link{plot_entropy_trajectories}} — customizable multi-site
Shannon entropy trajectories, summarizing evolutionary dynamics
across time.
\item \code{\link{plot_site_class_trajectory}} — single-site entropy
trajectory with class-change markers for inspecting individual
residues of interest.
\item \code{\link{tabulate_site_evolution}} — per-site amino acid
count and proportion tables across partitions, optionally
rendered as styled HTML.
}
}

\section{Simulation}{

\code{\link{simulate_variant_evolution}} provides a configurable
multi-variant simulation engine with user-specified emergence
schedules, growth rates, mutation-rate variability, and
deleterious-mutation injection. Generates synthetic time-series data
with known ground truth for benchmarking detection pipelines.
}

\section{Bundled data}{

\itemize{
\item \code{\link{sarscov2_variants}} — curated metadata for twelve
SARS-CoV-2 Variants of Concern and Variants of Interest, including
WHO labels, Pango lineages, GISAID and Nextstrain clades, dates
and countries of first detection, defining Spike-protein
mutations and SNP sites, and 21 peer-reviewed references with DOIs.
\item \code{\link{sarscov2_sample}} — a random sample of 100 NCBI Spike
protein sequences for end-to-end testing without external downloads.
}
}

\section{External data}{

The complete preprocessed NCBI Spike protein dataset (137,132
sequences, ~181.5 MB uncompressed FASTA) underlying the package's
real-data pre-processing vignette is archived on Zenodo:
\doi{10.5281/zenodo.19040165}. The dataset can be read directly with
\code{\link[Biostrings]{readAAStringSet}} and processed end-to-end
using the preprocessing toolkit; see
\code{vignette("preprocessing_pipeline", "ViralEntropR")} for the
full workflow.
}

\section{Vignettes}{

Three pre-rendered vignettes walk through the full workflow on real
and simulated data: \code{vignette("preprocessing_pipeline",
"ViralEntropR")}, \code{vignette("detecting_variants_simulation",
"ViralEntropR")}, and \code{vignette("clustering_accuracy",
"ViralEntropR")}.
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/vadimtyuryaev/ViralEntropR}
  \item \doi{10.5281/zenodo.19040165}
  \item \url{https://vadimtyuryaev.github.io/ViralEntropR/}
  \item Report bugs at \url{https://github.com/vadimtyuryaev/ViralEntropR/issues}
}

}
\author{
\strong{Maintainer}: Vadim Tyuryaev \email{vadim.tyuryaev@gmail.com} (\href{https://orcid.org/0009-0008-1361-6265}{ORCID})

Authors:
\itemize{
  \item Jane Heffernan \email{jmheffer@yorku.ca}
  \item Hanna Jankowski \email{hkj@yorku.ca}
}

}
\keyword{internal}
